@@ -2,6 +2,7 @@
 
 /*
  * Copyright (c) 2010, Oracle America, Inc.
+ * Copyright (c) 2012-2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,11 +35,23 @@
 
 /*
  * clnt.h - Client side remote procedure call interface.
+ *
+ * (1) There is only one SVCXPRT (service transport) per connection.
+ * (2) There can be several CLIENT per SVCXPRT.
+ * (3) Each CLIENT rpc_err cl_error is idempotent, initially
+ *     RPC_SUCCESS (0). Once set to anything else, it should never be
+ *     modified, and CLNT_DESTROY() should be invoked.
+ * (4) There can be many asynchronous clnt_req per CLIENT.
+ * (5) Each clnt_req has its own idempotent rpc_err cc_error.
  */
 
 #ifndef _TIRPC_CLNT_H_
 #define _TIRPC_CLNT_H_
 
+#include <misc/rbtree.h>
+#include <misc/wait_queue.h>
+#include <rpc/svc.h>
+#include <rpc/rpc_err.h>
 #include <rpc/clnt_stat.h>
 #include <rpc/auth.h>
 #include "reentrant.h"
@@ -67,33 +80,8 @@
 				 ((s) == RPC_PROGVERSMISMATCH) ||       \
 				 ((s) == RPC_CANTDECODEARGS))
 
-/*
- * Error info.
- */
-struct rpc_err {
-	enum clnt_stat re_status;
-	union {
-		int RE_errno;	/* related system error */
-		uint32_t RE_flags;
-		enum auth_stat RE_why;	/* why the auth error occurred */
-		struct {
-			rpcvers_t low;	/* lowest version supported */
-			rpcvers_t high;	/* highest version supported */
-		} RE_vers;
-		struct {	/* maybe meaningful if RPC_FAILED */
-			int32_t s1;
-			int32_t s2;
-		} RE_lb;	/* life boot & debugging only */
-	} ru;
-#define re_errno ru.RE_errno
-#define re_why  ru.RE_why
-#define re_vers  ru.RE_vers
-#define re_lb  ru.RE_lb
-#define re_flags        ru.RE_flags
-};
-
-#define RPC_ERR_FLAGS_NONE             0x0000
-#define RPC_ERR_FLAGS_ASYNC_REPLYFAIL  0x0001
+struct clnt_req;
+typedef void (*clnt_req_freer)(struct clnt_req *, size_t);
 
 /*
  * Client rpc handle.
@@ -104,24 +92,14 @@ typedef struct rpc_client {
 
 	struct clnt_ops {
 		/* call remote procedure */
-		enum clnt_stat (*cl_call) (struct rpc_client *, AUTH *,
-					   rpcproc_t, xdrproc_t, void *,
-					   xdrproc_t, void *, struct timeval);
+		enum clnt_stat (*cl_call) (struct clnt_req *);
+
 		/* abort a call */
 		void (*cl_abort) (struct rpc_client *);
 
-		/* get specific error code */
-		void (*cl_geterr) (struct rpc_client *, struct rpc_err *);
-
 		/* frees results */
 		 bool(*cl_freeres) (struct rpc_client *, xdrproc_t, void *);
 
-		/* take lifecycle ref */
-		 bool(*cl_ref) (struct rpc_client *, u_int flags);
-
-		/* release */
-		void (*cl_release) (struct rpc_client *, u_int flags);
-
 		/* release and mark destroyed */
 		void (*cl_destroy) (struct rpc_client *);
 
@@ -129,18 +107,51 @@ typedef struct rpc_client {
 		 bool(*cl_control) (struct rpc_client *, u_int, void *);
 	} *cl_ops;
 
-	mutex_t cl_lock;	/* serialize private data */
-	uint32_t cl_refcnt;	/* handle refcnt */
-	uint32_t cl_flags;	/* state flags */
-
-	void *cl_p1;		/* private data */
-	void *cl_p2;
-	void *cl_p3;
 	char *cl_netid;		/* network token */
 	char *cl_tp;		/* device name */
+	void *cl_u1;		/* client user data */
+	void *cl_u2;		/* client user data */
+
+	mutex_t cl_lock;	/* serialize private data */
+	struct rpc_err cl_error; /* specific error code */
+	uint32_t cl_refcnt;	/* handle refcnt */
+	uint16_t cl_flags;	/* state flags */
 
 } CLIENT;
 
+#define CLNT_REQ_FLAG_NONE	0x0000
+#define CLNT_REQ_FLAG_EXPIRING	0x0001
+#define CLNT_REQ_FLAG_BACKSYNC	0x0004
+#define CLNT_REQ_FLAG_ACKSYNC	0x0008
+
+/*
+ * RPC context.  Intended to enable efficient multiplexing of calls
+ * and replies sharing a common channel.
+ */
+struct clnt_req {
+	struct work_pool_entry cc_wpe;
+	struct opr_rbtree_node cc_dplx;
+	struct opr_rbtree_node cc_rqst;
+	struct waitq_entry cc_we;
+	struct opaque_auth cc_verf;
+
+	AUTH *cc_auth;
+	CLIENT *cc_clnt;
+	struct xdrpair cc_call;
+	struct xdrpair cc_reply;
+	void (*cc_process_cb)(struct clnt_req *);
+	clnt_req_freer cc_free_cb;
+	struct timespec cc_timeout;
+	struct rpc_err cc_error;
+	size_t cc_size;
+	int cc_expire_ms;
+	int cc_refreshes;
+	rpcproc_t cc_proc;
+	uint32_t cc_xid;
+	uint32_t cc_refs;
+	uint16_t cc_flags;
+};
+
 /*
  * Timers used for the pseudo-transport protocol when using datagrams
  */
@@ -167,47 +178,41 @@ struct rpc_timers {
  * CLNT flags
  */
 
-#define CLNT_FLAG_NONE               0x0000
-#define CLNT_FLAG_DESTROYED          0x0001
+#define CLNT_FLAG_NONE			SVC_XPRT_FLAG_NONE
+#define CLNT_FLAG_DESTROYING		SVC_XPRT_FLAG_DESTROYING
+#define CLNT_FLAG_RELEASING		SVC_XPRT_FLAG_RELEASING
+#define CLNT_FLAG_DESTROYED		SVC_XPRT_FLAG_DESTROYED
 
 /*
  * CLNT_REF flags
  */
 
-#define CLNT_REF_FLAG_NONE            0x0000
-#define CLNT_REF_FLAG_LOCKED          0x0001
+#define CLNT_REF_FLAG_NONE		SVC_XPRT_FLAG_NONE
+#define CLNT_REF_FLAG_LOCKED		SVC_XPRT_FLAG_LOCKED
 
 /*
  * CLNT_RELEASE flags
  */
 
-#define CLNT_RELEASE_FLAG_NONE            0x0000
-#define CLNT_RELEASE_FLAG_LOCKED          0x0001
+#define CLNT_RELEASE_FLAG_NONE		SVC_XPRT_FLAG_NONE
+#define CLNT_RELEASE_FLAG_LOCKED	SVC_XPRT_FLAG_LOCKED
 
 /*
  * client side rpc interface ops
- *
- * Parameter types are:
- *
  */
+#define CLNT_FAILURE(rh) ((rh)->cl_error.re_status != RPC_SUCCESS)
+#define CLNT_SUCCESS(rh) ((rh)->cl_error.re_status == RPC_SUCCESS)
 
 /*
  * enum clnt_stat
- * CLNT_CALL(rh, proc, xargs, argsp, xres, resp, timeout)
- *  CLIENT *rh;
- * AUTH * auth;
- * rpcproc_t proc;
- * xdrproc_t xargs;
- * void *argsp;
- * xdrproc_t xres;
- * void *resp;
- * struct timeval timeout;
+ * CLNT_CALL_BACK(cc)
+ * CLNT_CALL_ONCE(cc)
+ * CLNT_CALL_WAIT(cc)
+ *  struct clnt_req *cc;
  */
-#define CLNT_CALL(rh, ah, proc, xargs, argsp, xres, resp, secs) \
-	((*(rh)->cl_ops->cl_call)(rh, ah, proc, xargs, argsp, xres, resp, secs))
-
-#define clnt_call(rh, ah, proc, xargs, argsp, xres, resp, secs) \
-	((*(rh)->cl_ops->cl_call)(rh, ah, proc, xargs, argsp, xres, resp, secs))
+#define CLNT_CALL_BACK(cc) clnt_req_callback(cc)
+#define CLNT_CALL_ONCE(cc) ((*(cc)->cc_clnt->cl_ops->cl_call)(cc))
+#define CLNT_CALL_WAIT(cc) clnt_req_wait_reply(cc)
 
 /*
  * void
@@ -217,30 +222,6 @@ struct rpc_timers {
 #define CLNT_ABORT(rh) ((*(rh)->cl_ops->cl_abort)(rh))
 #define clnt_abort(rh) ((*(rh)->cl_ops->cl_abort)(rh))
 
-/*
- * struct rpc_err
- * CLNT_GETERR(rh);
- *  CLIENT *rh;
- */
-#define CLNT_GETERR(rh, errp) ((*(rh)->cl_ops->cl_geterr)(rh, errp))
-#define clnt_geterr(rh, errp) ((*(rh)->cl_ops->cl_geterr)(rh, errp))
-
-/*
- * uint32_t flags
- * CLNT_REF(rh);
- *  CLIENT *rh;
- */
-#define CLNT_REF(rh, flags) ((*(rh)->cl_ops->cl_ref)(rh, flags))
-#define clnt_ref(rh, flags) ((*(rh)->cl_ops->cl_ref)(rh, flags))
-
-/*
- * uint32_t flags
- * CLNT_RELEASE(rh);
- *  CLIENT *rh;
- */
-#define CLNT_RELEASE(rh, flags) ((*(rh)->cl_ops->cl_release)(rh, flags))
-#define clnt_release(rh, flags) ((*(rh)->cl_ops->cl_release)(rh, flags))
-
 /*
  * bool
  * CLNT_FREERES(rh, xres, resp);
@@ -266,11 +247,13 @@ struct rpc_timers {
 	((*(cl)->cl_ops->cl_control)(cl, rq, in))
 
 /*
- * control operations that apply to both udp and tcp transports
+ * control operations that apply to all transports
  */
-#define CLSET_TIMEOUT  1	/* set timeout (timeval) */
-#define CLGET_TIMEOUT  2	/* get timeout (timeval) */
+/* reserved 1 */
+/* reserved 2 */
 #define CLGET_SERVER_ADDR 3	/* get server's address (sockaddr) */
+/* reserved 4 */
+/* reserved 5 */
 #define CLGET_FD  6		/* get connections file descriptor */
 #define CLGET_SVC_ADDR  7	/* get server's address (netbuf) */
 #define CLSET_FD_CLOSE  8	/* close fd while clnt_destroy */
@@ -284,21 +267,80 @@ struct rpc_timers {
 #define CLSET_SVC_ADDR  16	/* get server's address (netbuf) */
 #define CLSET_PUSH_TIMOD 17	/* push timod if not already present */
 #define CLSET_POP_TIMOD  18	/* pop timod */
-/*
- * Connectionless only control operations
- */
-#define CLSET_RETRY_TIMEOUT 4	/* set retry timeout (timeval) */
-#define CLGET_RETRY_TIMEOUT 5	/* get retry timeout (timeval) */
-#define CLSET_ASYNC  19
-#define CLSET_CONNECT  20	/* Use connect() for UDP. (int) */
 
-/*
- * void
- * CLNT_DESTROY(rh);
- *  CLIENT *rh;
+/* Protect a CLIENT with a CLNT_REF for each call or request.
  */
-#define CLNT_DESTROY(rh) ((*(rh)->cl_ops->cl_destroy)(rh))
-#define clnt_destroy(rh) ((*(rh)->cl_ops->cl_destroy)(rh))
+static inline void clnt_ref_it(CLIENT *clnt, uint32_t flags,
+			       const char *tag, const int line)
+{
+	uint32_t refs = atomic_inc_uint32_t(&clnt->cl_refcnt);
+
+	if (flags & CLNT_REF_FLAG_LOCKED)  {
+		/* unlock before warning trace */
+		mutex_unlock(&clnt->cl_lock);
+	}
+	__warnx(TIRPC_DEBUG_FLAG_REFCNT, "%s: %p %" PRIu32 " @%s:%d",
+		__func__, clnt, refs, tag, line);
+}
+#define CLNT_REF(clnt, flags)						\
+	clnt_ref_it(clnt, flags, __func__, __LINE__)
+
+static inline void clnt_release_it(CLIENT *clnt, uint32_t flags,
+				   const char *tag, const int line)
+{
+	uint32_t refs = atomic_dec_uint32_t(&clnt->cl_refcnt);
+	uint16_t cl_flags;
+
+	if (flags & CLNT_RELEASE_FLAG_LOCKED) {
+		/* unlock before warning trace */
+		mutex_unlock(&clnt->cl_lock);
+	}
+	__warnx(TIRPC_DEBUG_FLAG_REFCNT, "%s: %p %" PRIu32 " @%s:%d",
+		__func__, clnt, refs, tag, line);
+
+	if (likely(refs > 0)) {
+		/* normal case */
+		return;
+	}
+
+	/* enforce once-only semantic, trace others */
+	cl_flags = atomic_postset_uint16_t_bits(&clnt->cl_flags,
+						CLNT_FLAG_RELEASING);
+
+	if (cl_flags & CLNT_FLAG_RELEASING) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: %p WARNING! already destroying! @%s:%d",
+			__func__, clnt, tag, line);
+		return;
+	}
+
+	/* Releasing last reference */
+	(*(clnt)->cl_ops->cl_destroy)(clnt);
+}
+#define CLNT_RELEASE(clnt, flags)					\
+	clnt_release_it(clnt, flags, __func__, __LINE__)
+
+/* CLNT_DESTROY is CLNT_RELEASE with once-only semantics.  Also, idempotent
+ * CLNT_FLAG_DESTROYED indicates that more references should not be taken.
+ */
+static inline void clnt_destroy_it(CLIENT *clnt,
+				   const char *tag, const int line)
+{
+	uint16_t flags = atomic_postset_uint16_t_bits(&clnt->cl_flags,
+						      CLNT_FLAG_DESTROYING);
+
+	__warnx(TIRPC_DEBUG_FLAG_REFCNT, "%s: %p %" PRIu32 " @%s:%d",
+		__func__, clnt, clnt->cl_refcnt, tag, line);
+
+	if (flags & CLNT_FLAG_DESTROYING) {
+		/* previously set, do nothing */
+		return;
+	}
+
+	clnt_release_it(clnt, CLNT_RELEASE_FLAG_NONE, tag, line);
+}
+#define CLNT_DESTROY(clnt)						\
+	clnt_destroy_it(clnt, __func__, __LINE__)
 
 /*
  * RPCTEST is a test program which is accessible on every rpc
@@ -319,29 +361,17 @@ struct rpc_timers {
 
 /*
  * Below are the client handle creation routines for the various
- * implementations of client side rpc.  They can return NULL if a
- * creation failure occurs.
+ * implementations of client side rpc.
+ *
+ * Always returns CLIENT. Must check cl_error.re_status,
+ * followed by CLNT_DESTROY() as necessary.
  */
+__BEGIN_DECLS
 
 /*
  * Generic client creation routine. Supported protocols are those that
  * belong to the nettype namespace (/etc/netconfig).
  */
-__BEGIN_DECLS
-extern CLIENT * clnt_ncreate(const char *, const rpcprog_t,
-			     const rpcvers_t, const char *);
-/*
- *
- * const char *hostname;   -- hostname
- * const rpcprog_t prog;   -- program number
- * const rpcvers_t vers;   -- version number
- * const char *nettype;    -- network type
- */
-
-/*
- * Generic client creation routine. Just like clnt_create(), except
- * it takes an additional timeout parameter.
- */
 extern CLIENT *clnt_ncreate_timed(const char *, const rpcprog_t,
 				  const rpcvers_t, const char *,
 				  const struct timeval *);
@@ -350,28 +380,23 @@ extern CLIENT *clnt_ncreate_timed(const char *, const rpcprog_t,
  * const char *hostname;   -- hostname
  * const rpcprog_t prog;   -- program number
  * const rpcvers_t vers;   -- version number
- * const char *nettype;   -- network type
+ * const char *nettype;    -- network type
  * const struct timeval *tp;  -- timeout
  */
 
 /*
- * Generic client creation routine. Supported protocols are which belong
- * to the nettype name space.
- */
-extern CLIENT *clnt_ncreate_vers(const char *, const rpcprog_t, rpcvers_t *,
-				 const rpcvers_t, const rpcvers_t,
-				 const char *);
-/*
- * const char *host;  -- hostname
- * const rpcprog_t prog;  -- program number
- * rpcvers_t *vers_out;  -- servers highest available version
- * const rpcvers_t vers_low; -- low version number
- * const rpcvers_t vers_high; -- high version number
- * const char *nettype;  -- network type
+ * Calls clnt_ncreate_timed() with a NULL value for the timeout
+ * pointer, indicating that the default timeout should be used.
  */
+static inline CLIENT *
+clnt_ncreate(const char *hostname, const rpcprog_t prog, const rpcvers_t vers,
+	     const char *nettype)
+{
+	return (clnt_ncreate_timed(hostname, prog, vers, nettype, NULL));
+}
 
 /*
- * Generic client creation routine. Supported protocols are which belong
+ * Generic client creation routine. Supported protocols belong
  * to the nettype name space.
  */
 extern CLIENT *clnt_ncreate_vers_timed(const char *, const rpcprog_t,
@@ -389,21 +414,21 @@ extern CLIENT *clnt_ncreate_vers_timed(const char *, const rpcprog_t,
  */
 
 /*
- * Generic client creation routine. It takes a netconfig structure
- * instead of nettype
- */
-extern CLIENT *clnt_tp_ncreate(const char *, const rpcprog_t, const rpcvers_t,
-			       const struct netconfig *);
-/*
- * const char *hostname;   -- hostname
- * const rpcprog_t prog;   -- program number
- * const rpcvers_t vers;   -- version number
- * const struct netconfig *netconf;  -- network config structure
+ * Calls clnt_create_vers_timed() with a NULL value for the timeout
+ * pointer, indicating that the default timeout should be used.
  */
+static inline CLIENT *
+clnt_ncreate_vers(const char *hostname, const rpcprog_t prog,
+		  rpcvers_t *vers_out, const rpcvers_t vers_low,
+		  const rpcvers_t vers_high, const char *nettype)
+{
+	return (clnt_ncreate_vers_timed
+		(hostname, prog, vers_out, vers_low, vers_high, nettype, NULL));
+}
 
 /*
- * Generic client creation routine. Just like clnt_tp_create(), except
- * it takes an additional timeout parameter.
+ * Generic client creation routine. It takes a netconfig structure
+ * instead of nettype
  */
 extern CLIENT *clnt_tp_ncreate_timed(const char *, const rpcprog_t,
 				     const rpcvers_t, const struct netconfig *,
@@ -416,6 +441,17 @@ extern CLIENT *clnt_tp_ncreate_timed(const char *, const rpcprog_t,
  * const struct timeval *tp  -- timeout
  */
 
+/*
+ * Calls clnt_tp_create_timed() with a NULL value for the timeout
+ * pointer, indicating that the default timeout should be used.
+ */
+static inline CLIENT *
+clnt_tp_ncreate(const char *hostname, const rpcprog_t prog,
+		const rpcvers_t vers, const struct netconfig *nconf)
+{
+	return (clnt_tp_ncreate_timed(hostname, prog, vers, nconf, NULL));
+}
+
 /*
  * Generic TLI create routine. Only provided for compatibility.
  */
@@ -437,40 +473,48 @@ extern CLIENT *clnt_tli_ncreate(const int, const struct netconfig *,
  * Low level clnt create routines for connectionful transports, e.g. tcp.
  */
 
-#define CLNT_CREATE_FLAG_NONE           0x0000
-#define CLNT_CREATE_FLAG_CONNECT        0x0001
-#define CLNT_CREATE_FLAG_SVCXPRT        0x0002
+/* uint16_t actually used */
+#define CLNT_CREATE_FLAG_NONE		SVC_CREATE_FLAG_NONE
+#define CLNT_CREATE_FLAG_CLOSE		SVC_CREATE_FLAG_CLOSE
 
-extern CLIENT *clnt_vc_ncreate(const int, const struct netbuf *,
-			       const rpcprog_t, const rpcvers_t, u_int, u_int);
+/* uint32_t instructions */
+#define CLNT_CREATE_FLAG_CONNECT	0x10000000
+#define CLNT_CREATE_FLAG_LISTEN		SVC_CREATE_FLAG_LISTEN
+#define CLNT_CREATE_FLAG_SVCXPRT	0x40000000
+#define CLNT_CREATE_FLAG_XPRT_DOREG	SVC_CREATE_FLAG_XPRT_DOREG
+#define CLNT_CREATE_FLAG_XPRT_NOREG	SVC_CREATE_FLAG_XPRT_NOREG
 
-extern CLIENT *clnt_vc_ncreate2(const int, const struct netbuf *,
-				const rpcprog_t, const rpcvers_t, u_int, u_int,
-				u_int);
+extern CLIENT *clnt_vc_ncreatef(const int, const struct netbuf *,
+				const rpcprog_t, const rpcvers_t,
+				const u_int, const u_int, const uint32_t);
+
+static inline CLIENT *
+clnt_vc_ncreate(const int fd, const struct netbuf *raddr,
+		const rpcprog_t prog, const rpcvers_t vers,
+		const u_int sendsz, const u_int recvsz)
+{
+	return (clnt_vc_ncreatef(fd, raddr, prog, vers, sendsz, recvsz,
+				 CLNT_CREATE_FLAG_CONNECT));
+}
 
-#if !defined(_WIN32)
 /*
- * Added for compatibility to old rpc 4.0. Obsoleted by clnt_vc_create().
+ * Create a client handle from an active service transport handle.
  */
-extern CLIENT *clntunix_ncreate(struct sockaddr_un *, u_long, u_long, int *,
-				u_int, u_int);
-#endif
-
+extern CLIENT *clnt_vc_ncreate_svc(const SVCXPRT *, const rpcprog_t,
+				   const rpcvers_t, const uint32_t);
 /*
- * const int fd;    -- open file descriptor
- * const struct netbuf *svcaddr;  -- servers address
- * const rpcprog_t prog;   -- program number
- * const rpcvers_t vers;   -- version number
- * const u_int sendsz;   -- buffer recv size
- * const u_int recvsz;   -- buffer send size
+ *      const SVCXPRT *xprt;                    -- active service xprt
+ *      const rpcprog_t prog;                   -- RPC program number
+ *      const rpcvers_t vers;                   -- RPC program version
+ *      const uint32_t flags;                   -- flags
  */
 
 /*
  * Low level clnt create routine for connectionless transports, e.g. udp.
  */
-extern CLIENT *clnt_dg_ncreate(const int, const struct netbuf *,
-			       const rpcprog_t, const rpcvers_t, const u_int,
-			       const u_int);
+extern CLIENT *clnt_dg_ncreatef(const int, const struct netbuf *,
+				const rpcprog_t, const rpcvers_t,
+				const u_int, const u_int, const uint32_t);
 /*
  * const int fd;    -- open file descriptor
  * const struct netbuf *svcaddr;  -- servers address
@@ -480,6 +524,15 @@ extern CLIENT *clnt_dg_ncreate(const int, const struct netbuf *,
  * const u_int recvsz;   -- buffer send size
  */
 
+static inline CLIENT *
+clnt_dg_ncreate(const int fd, const struct netbuf *raddr,
+		const rpcprog_t prog, const rpcvers_t vers,
+		const u_int sendsz, const u_int recvsz)
+{
+	return (clnt_dg_ncreatef(fd, raddr, prog, vers, sendsz, recvsz,
+				 CLNT_CREATE_FLAG_NONE));
+}
+
 /*
  * Memory based rpc (for speed check and testing)
  * CLIENT *
@@ -489,41 +542,56 @@ extern CLIENT *clnt_dg_ncreate(const int, const struct netbuf *,
  */
 extern CLIENT *clnt_raw_ncreate(rpcprog_t, rpcvers_t);
 
-__END_DECLS
 /*
- * Print why creation failed
+ * Client request processing
  */
-__BEGIN_DECLS
-extern void clnt_pcreateerror(const char *);	/* stderr */
-extern char *clnt_spcreateerror(const char *);	/* string */
+static inline void clnt_req_fill(struct clnt_req *cc, struct rpc_client *clnt,
+				 AUTH *auth, rpcproc_t proc,
+				 xdrproc_t xargs, void *argsp,
+				 xdrproc_t xresults, void *resultsp)
+{
+	cc->cc_clnt = clnt;
+	cc->cc_auth = auth;
+	cc->cc_proc = proc;
+	cc->cc_call.proc = xargs;
+	cc->cc_call.where = argsp;
+	cc->cc_reply.proc = xresults;
+	cc->cc_reply.where = resultsp;
+	cc->cc_verf = _null_auth;
+
+	cc->cc_free_cb = (clnt_req_freer)__ntirpc_pkg_params.free_size_;
+	cc->cc_size = sizeof(*cc);
+	cc->cc_refs = 1;
+
+	/* protects this */
+	pthread_mutex_init(&cc->cc_we.mtx, NULL);
+	pthread_mutex_lock(&cc->cc_we.mtx);
+	pthread_cond_init(&cc->cc_we.cv, 0);
+}
+
+static inline void clnt_req_fini(struct clnt_req *cc)
+{
+	pthread_cond_destroy(&cc->cc_we.cv);
+	pthread_mutex_unlock(&cc->cc_we.mtx);
+	pthread_mutex_destroy(&cc->cc_we.mtx);
+	CLNT_RELEASE(cc->cc_clnt, CLNT_RELEASE_FLAG_NONE);
+}
+
+enum clnt_stat clnt_req_callback(struct clnt_req *);
+enum clnt_stat clnt_req_refresh(struct clnt_req *);
+void clnt_req_reset(struct clnt_req *);
+enum clnt_stat clnt_req_setup(struct clnt_req *, struct timespec);
+enum clnt_stat clnt_req_wait_reply(struct clnt_req *);
+int clnt_req_release(struct clnt_req *);
+
 __END_DECLS
 /*
- * Like clnt_perror(), but is more verbose in its output
+ * Used by rpc_perror() and rpc_sperror()
  */
 __BEGIN_DECLS
 extern void clnt_perrno(enum clnt_stat);	/* stderr */
 extern char *clnt_sperrno(enum clnt_stat);	/* string */
 __END_DECLS
-/*
- * Print an English error message, given the client error code
- */
-__BEGIN_DECLS
-extern void clnt_perror(CLIENT *, const char *);	/* stderr */
-extern char *clnt_sperror(CLIENT *, const char *);	/* string */
-__END_DECLS
-/*
- * If a creation fails, the following allows the user to figure out why.
- */
-struct rpc_createerr {
-	enum clnt_stat cf_stat;
-	struct rpc_err cf_error; /* useful when cf_stat == RPC_PMAPFAILURE */
-};
-
-__BEGIN_DECLS
-extern struct rpc_createerr *__rpc_createerr(void);
-__END_DECLS
-#define get_rpc_createerr() (*(__rpc_createerr()))
-#define rpc_createerr  (*(__rpc_createerr()))
 /*
  * The simplified interface:
  * enum clnt_stat
@@ -604,6 +672,5 @@ extern enum clnt_stat rpc_broadcast_exp(const rpcprog_t, const rpcvers_t,
 					const int, const char *);
 __END_DECLS
 /* For backward compatibility */
-#include <rpc/clnt_soc.h>
 #include <rpc/tirpc_compat.h>
 #endif				/* !_TIRPC_CLNT_H_ */
