@@ -2,6 +2,7 @@
 
 /*
  * Copyright (c) 2009, Sun Microsystems, Inc.
+ * Copyright (c) 2012-2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,24 +43,31 @@
 #define _TIRPC_SVC_H
 
 #include <sys/cdefs.h>
+#include <rpc/rpc_msg.h>
 #include <rpc/types.h>
 #include <rpc/work_pool.h>
 #include <misc/portable.h>
-#include <misc/queue.h>
-#include <misc/rbtree.h>
 #include "reentrant.h"
 #if defined(HAVE_BLKIN)
 #include <blkin/zipkin_c.h>
 #endif
 
+typedef struct svc_xprt SVCXPRT;
+
+enum xprt_stat {
+	XPRT_IDLE = 0,
+	XPRT_MOREREQS,
+	/* always last in this order for comparisons */
+	XPRT_DIED,
+	XPRT_DESTROYED
+};
 
 /*
  * This interface must manage two items concerning remote procedure calling:
  *
  * 1) An arbitrary number of transport connections upon which rpc requests
  * are received.  The two most notable transports are TCP and UDP;  they are
- * created and registered by routines in svc_tcp.c and svc_udp.c, respectively;
- * they in turn call xprt_register and xprt_unregister.
+ * created and registered by routines in svc_vc.c and svc_dg.c, respectively.
  *
  * 2) An arbitrary number of locally registered services.  Services are
  * described by the following four data: program number, version number,
@@ -67,7 +75,7 @@
  * indicates whether or not the exported program should be registered with a
  * local binder service;  if true the program's number and version and the
  * port number from the transport handle are registered with the binder.
- * These data are registered with the rpc svc system via svc_register.
+ * These data are registered with the rpc svc system via svc_reg.
  *
  * A service's dispatch function is called whenever an rpc request comes in
  * on a transport.  The request's program and version numbers must match
@@ -79,7 +87,6 @@
 #define SVC_INIT_DEFAULT        0x0000
 #define SVC_INIT_XPRTS          0x0001
 #define SVC_INIT_EPOLL          0x0002
-#define SVC_INIT_WARNX          0x0004
 #define SVC_INIT_NOREG_XPRTS    0x0008
 #define SVC_INIT_BLKIN          0x0010
 
@@ -92,16 +99,8 @@
 #define SVCSET_VERSQUIET        2
 #define SVCGET_CONNMAXREC       3
 #define SVCSET_CONNMAXREC       4
-#define SVCGET_XP_RECV          5
-#define SVCSET_XP_RECV          6
 #define SVCGET_XP_FLAGS         7
 #define SVCSET_XP_FLAGS         8
-#define SVCGET_XP_GETREQ        9
-#define SVCSET_XP_GETREQ        10
-#define SVCGET_XP_DISPATCH      11
-#define SVCSET_XP_DISPATCH      12
-#define SVCGET_XP_RECV_USER_DATA        13
-#define SVCSET_XP_RECV_USER_DATA        14
 #define SVCGET_XP_FREE_USER_DATA        15
 #define SVCSET_XP_FREE_USER_DATA        16
 
@@ -115,30 +114,24 @@
 #define RPC_SVC_FDSET_GET       4
 #define RPC_SVC_FDSET_SET       5
 
-/*
- * Flags for svc_fd_ncreate2
- */
-
-#define SVC_VCCR_NONE             0x0000
-
-/* Svc event strategy */
-enum svc_event_type {
-	SVC_EVENT_FDSET /* trad. using select and poll (currently unhooked) */ ,
-	SVC_EVENT_EPOLL		/* Linux epoll interface */
-};
+typedef enum xprt_stat (*svc_xprt_fun_t) (SVCXPRT *);
+typedef enum xprt_stat (*svc_xprt_xdr_fun_t) (SVCXPRT *, XDR *);
 
 typedef struct svc_init_params {
+	svc_xprt_fun_t disconnect_cb;
+	svc_xprt_xdr_fun_t request_cb;
+
 	u_long flags;
-	warnx_t warnx;
 	u_int max_connections;	/* xprts */
 	u_int max_events;	/* evchan events */
-	u_int svc_ioq_maxbuf;
-	int32_t idle_timeout;
+	u_int ioq_send_max;
+	u_int ioq_thrd_max;
 	u_int gss_ctx_hash_partitions;
 	u_int gss_max_ctx;
 	u_int gss_max_idle_gen;
 	u_int gss_max_gc;
-	u_int ioq_thrd_max;
+	uint32_t channels;
+	int32_t idle_timeout;
 } svc_init_params;
 
 /* Svc param flags */
@@ -152,12 +145,18 @@ typedef struct svc_init_params {
 #define SVC_XPRT_FLAG_NONE		0x0000
 /* uint16_t actually used */
 #define SVC_XPRT_FLAG_ADDED		0x0001
-#define SVC_XPRT_FLAG_BLOCKED		0x0002
-#define SVC_XPRT_FLAG_DESTROYED		0x0020	/* SVC_DESTROY() was called */
-#define SVC_XPRT_FLAG_DESTROYING	0x0040	/* (*xp_destroy) was called */
+
+#define SVC_XPRT_FLAG_INITIAL		0x0004
+#define SVC_XPRT_FLAG_INITIALIZED	0x0008
+#define SVC_XPRT_FLAG_CLOSE		0x0010
+#define SVC_XPRT_FLAG_DESTROYING	0x0020	/* SVC_DESTROY() was called */
+#define SVC_XPRT_FLAG_RELEASING		0x0040	/* (*xp_destroy) was called */
+#define SVC_XPRT_FLAG_UREG		0x0080
+
+#define SVC_XPRT_FLAG_DESTROYED (SVC_XPRT_FLAG_DESTROYING \
+				| SVC_XPRT_FLAG_RELEASING)
 
 /* uint32_t instructions */
-#define SVC_XPRT_FLAG_LOCK		SVC_XPRT_FLAG_NONE
 #define SVC_XPRT_FLAG_LOCKED		0x00010000
 #define SVC_XPRT_FLAG_UNLOCK		0x00020000
 
@@ -175,93 +174,71 @@ typedef struct svc_init_params {
 #define SVC_RELEASE_FLAG_NONE		SVC_XPRT_FLAG_NONE
 #define SVC_RELEASE_FLAG_LOCKED		SVC_XPRT_FLAG_LOCKED
 
-/* XXX Ganesha
- * Don't confuse with (currently incomplete) transport type, nee
- * socktype. */
+/* Don't confuse with (currently incomplete) transport type, nee socktype.
+ */
 typedef enum xprt_type {
 	XPRT_UNKNOWN = 0,
+	XPRT_NON_RENDEZVOUS,
 	XPRT_UDP,
+	XPRT_UDP_RENDEZVOUS,
 	XPRT_TCP,
 	XPRT_TCP_RENDEZVOUS,
 	XPRT_SCTP,
-	XPRT_RDMA
+	XPRT_SCTP_RENDEZVOUS,
+	XPRT_RDMA,
+	XPRT_RDMA_RENDEZVOUS,
+	XPRT_VSOCK,
+	XPRT_VSOCK_RENDEZVOUS
 } xprt_type_t;
 
-enum xprt_stat {
-	XPRT_DIED,
-	XPRT_MOREREQS,
-	XPRT_IDLE,
-	XPRT_DESTROYED
-};
-
-struct cf_rendezvous {		/* kept in xprt->xp_p1 for rendezvouser */
-	u_int sendsize;
-	u_int recvsize;
-	int maxrec;
-};
-
 struct SVCAUTH;			/* forward decl. */
+struct svc_req;			/* forward decl. */
+
+typedef enum xprt_stat (*svc_req_fun_t) (struct svc_req *);
 
 /*
  * Server side transport handle
  */
-typedef struct rpc_svcxprt {
+struct svc_xprt {
 	struct xp_ops {
 		/* receive incoming requests */
-		bool (*xp_recv) (struct rpc_svcxprt *, struct svc_req *);
+		svc_xprt_fun_t xp_recv;
 
 		/* get transport status */
-		enum xprt_stat (*xp_stat) (struct rpc_svcxprt *);
+		svc_xprt_fun_t xp_stat;
 
-		/* get arguments, thread u_data in arg4 */
-		bool (*xp_getargs) (struct rpc_svcxprt *, struct svc_req *,
-				    xdrproc_t, void *, void *);
+		/* decode incoming message header (called by request_cb) */
+		svc_req_fun_t xp_decode;
 
 		/* send reply */
-		bool (*xp_reply) (struct rpc_svcxprt *, struct svc_req *,
-				  struct rpc_msg *);
+		svc_req_fun_t xp_reply;
 
-		/* free mem allocated for args */
-		bool (*xp_freeargs) (struct rpc_svcxprt *, struct svc_req *,
-				     xdrproc_t, void *);
+		/* optional checksum (after authentication/decryption) */
+		void (*xp_checksum) (struct svc_req *, void *, size_t);
 
 		/* actually destroy after xp_destroy_it and xp_release_it */
-		void (*xp_destroy) (struct rpc_svcxprt *, u_int,
-				    const char *, const int);
+		void (*xp_destroy) (SVCXPRT *, u_int, const char *, const int);
 
 		/* catch-all function */
-		bool (*xp_control) (struct rpc_svcxprt *, const u_int, void *);
-
-		/* transport locking (may be duplex-aware, etc) */
-		void (*xp_lock) (struct rpc_svcxprt *, uint32_t,
-				 const char *, int);
-		void (*xp_unlock) (struct rpc_svcxprt *, uint32_t,
-				   const char *, int);
-
-		/* handle incoming requests (calls xp_recv) */
-		bool (*xp_getreq) (struct rpc_svcxprt *);
-
-		/* call dispatch strategy function */
-		void (*xp_dispatch) (struct rpc_svcxprt *, struct rpc_msg **);
-
-		/* rendezvous (epilogue) */
-		u_int (*xp_recv_user_data) (struct rpc_svcxprt *,
-					    struct rpc_svcxprt *,
-					    const u_int, void *);
+		bool (*xp_control) (SVCXPRT *, const u_int, void *);
 
 		/* free client user data */
-		bool (*xp_free_user_data) (struct rpc_svcxprt *);
+		svc_xprt_fun_t xp_free_user_data;
 	} *xp_ops;
 
+	/* handle incoming connections (per xp_fd) */
+	union {
+		svc_req_fun_t process_cb;
+		svc_xprt_fun_t rendezvous_cb;
+	}  xp_dispatch;
+	SVCXPRT *xp_parent;
+
 	char *xp_tp;		/* transport provider device name */
 	char *xp_netid;		/* network token */
 
-	void *xp_ev;		/* event handle */
 	void *xp_p1;		/* private: for use by svc ops */
 	void *xp_p2;		/* private: for use by svc ops */
 	void *xp_p3;		/* private: for use by svc lib */
-	void *xp_p4;		/* private: for use by svc lib */
-	void *xp_p5;		/* private: for use by svc lib */
 	void *xp_u1;		/* client user data */
 	void *xp_u2;		/* client user data */
 
@@ -275,47 +252,24 @@ typedef struct rpc_svcxprt {
 		struct blkin_endpoint endp;
 	} blkin;
 #endif
-	/* auth */
-	mutex_t xp_auth_lock;	/* lock owned by installed authenticator */
-
 	/* serialize private data */
 	mutex_t xp_lock;
 
-	/* event vector list */
-	TAILQ_ENTRY(rpc_svcxprt) xp_evq;
-
-	/* indexed by fd */
-	struct opr_rbtree_node xp_fd_node;
-
-	uint32_t xp_refs;	/* handle reference count */
-	uint32_t xp_requests;	/* related requests count */
-
 	int xp_fd;
+	int xp_ifindex;		/* interface index */
 	int xp_si_type;		/* si type */
 	int xp_type;		/* xprt type */
 
+	uint32_t xp_refs;	/* handle reference count */
 	uint16_t xp_flags;	/* flags */
-	u_short xp_port;	/* associated port number */
-
-	/*
-	 * union of event processor types
-	 */
-	enum svc_event_type ev_type;
-	union {
-#if defined(TIRPC_EPOLL)
-		struct {
-			struct epoll_event event;
-		} epoll;
-#endif
-	} ev_u;
-} SVCXPRT;
+};
 
 /* Service record used by exported search routines */
 typedef struct svc_record {
 	rpcprog_t sc_prog;
 	rpcvers_t sc_vers;
 	char *sc_netid;
-	void (*sc_dispatch) (struct svc_req *, SVCXPRT *);
+	void (*sc_dispatch) (struct svc_req *);
 } svc_rec_t;
 
 typedef struct svc_vers_range {
@@ -331,25 +285,10 @@ typedef enum svc_lookup_result {
 	SVC_LKP_ERR = 667,
 } svc_lookup_result_t;
 
-/* functions which can be installed using a control function, e.g.,
- * xp_ops->xp_control */
-typedef bool(*xp_recv_t) (struct rpc_svcxprt *, struct svc_req *);
-typedef bool(*xp_getreq_t) (struct rpc_svcxprt *);
-typedef void (*xp_dispatch_t) (struct rpc_svcxprt *, struct rpc_msg **);
-typedef u_int(*xp_recv_user_data_t) (struct rpc_svcxprt *, struct rpc_svcxprt *,
-				     const u_int, void *);
-typedef bool(*xp_free_user_data_t) (struct rpc_svcxprt *);
-
 /*
  * Service request
  */
 struct svc_req {
-	u_int32_t rq_prog;	/* service program number */
-	u_int32_t rq_vers;	/* service protocol version */
-	u_int32_t rq_proc;	/* the desired procedure */
-
-	struct opaque_auth rq_cred;	/* raw creds from the wire */
-	void *rq_clntcred;	/* read only cooked cred */
 	SVCXPRT *rq_xprt;	/* associated transport */
 
 	/* New with TI-RPC */
@@ -357,37 +296,31 @@ struct svc_req {
 	caddr_t rq_svcname;	/* read only cooked service cred */
 
 	/* New with N TI-RPC */
-	struct rpc_msg *rq_msg;	/* decoded rpc_msg */
-	void *rq_context;	/* private context */
+	XDR *rq_xdrs;
 	void *rq_u1;		/* user data */
 	void *rq_u2;		/* user data */
 	uint64_t rq_cksum;
-	u_int32_t rq_xid;	/* xid */
 
 	/* Moved in N TI-RPC */
-	struct opaque_auth rq_verf;	/* raw response verifier */
 	struct SVCAUTH *rq_auth;	/* auth handle */
 	void *rq_ap1;		/* auth private */
 	void *rq_ap2;		/* auth private */
 
-	/* copy of remote transport address */
-	struct sockaddr_storage rq_raddr;
-
-	/* Store dest addr for UDP to send replys from */
-	struct sockaddr_storage rq_daddr;
-
-	size_t rq_raddr_len;
-	size_t rq_daddr_len;
+	/* avoid separate alloc/free */
+	struct rpc_msg rq_msg;
 
 #if defined(HAVE_BLKIN)
 	/* blkin tracing */
 	struct blkin_trace bl_trace;
 #endif
+	uint32_t rq_refs;
 };
 
 /*
  *  Approved way of getting addresses
  */
+#define svc_getcaller_netbuf(x) (&(x)->xp_remote.nb)
+#define svc_getlocal_netbuf(x) (&(x)->xp_local.nb)
 #define svc_getrpccaller(x) (&(x)->xp_remote.ss)
 #define svc_getrpclocal(x) (&(x)->xp_local.ss)
 
@@ -418,39 +351,26 @@ __END_DECLS
  *
  * SVCXPRT *xprt;
  * struct svc_req *req;
- * xdrproc_t xargs;
- * void * argsp;
  */
-#define SVC_RECV(xprt, req) \
-	(*(xprt)->xp_ops->xp_recv)((xprt), (req))
-#define svc_recv(xprt, req)			\
-	(*(xprt)->xp_ops->xp_recv)((xprt), (req))
+#define SVC_RECV(xprt) \
+	(*(xprt)->xp_ops->xp_recv)(xprt)
 
 #define SVC_STAT(xprt) \
 	(*(xprt)->xp_ops->xp_stat)(xprt)
-#define svc_stat(xprt)				\
-	(*(xprt)->xp_ops->xp_stat)(xprt)
 
-#define SVC_GETARGS(xprt, req, xargs, argsp, u_data)			\
-	(*(xprt)->xp_ops->xp_getargs)((xprt), (req), (xargs), (argsp),	\
-				      (u_data))
-#define svc_getargs(xprt, req, xargs, argsp, u_data)		       \
-	(*(xprt)->xp_ops->xp_getargs)((xprt), (req), (xargs), (argsp), \
-				      (u_data))
+#define SVC_DECODE(req) \
+	(*((req)->rq_xprt)->xp_ops->xp_decode)(req)
 
-#define SVC_REPLY(xprt, req, msg)				\
-	(*(xprt)->xp_ops->xp_reply) ((xprt), (req), (msg))
-#define svc_reply(xprt, req, msg) \
-	(*(xprt)->xp_ops->xp_reply) ((xprt), (req), (msg))
+#define SVC_REPLY(req) \
+	(*((req)->rq_xprt)->xp_ops->xp_reply)(req)
 
-#define SVC_FREEARGS(xprt, req, xargs, argsp)			\
-	(*(xprt)->xp_ops->xp_freeargs)((xprt), (req), (xargs), (argsp))
-#define svc_freeargs(xprt, req, xargs, argsp)			\
-	(*(xprt)->xp_ops->xp_freeargs)((xprt), (req), (xargs), (argsp))
+#define SVC_CHECKSUM(req, what, length) \
+	if (((req)->rq_xprt)->xp_ops->xp_checksum) \
+		(*((req)->rq_xprt)->xp_ops->xp_checksum)(req, what, length)
 
 /* Protect a SVCXPRT with a SVC_REF for each call or request.
  */
-static inline void svc_ref_it(struct rpc_svcxprt *xprt, u_int flags,
+static inline void svc_ref_it(SVCXPRT *xprt, u_int flags,
 			      const char *tag, const int line)
 {
 	atomic_inc_uint32_t(&xprt->xp_refs);
@@ -463,14 +383,10 @@ static inline void svc_ref_it(struct rpc_svcxprt *xprt, u_int flags,
 }
 #define SVC_REF2(xprt, flags, tag, line)				\
 	svc_ref_it(xprt, flags, tag, line)
-#define svc_ref2(xprt, flags, tag, line)				\
-	svc_ref_it(xprt, flags, tag, line)
 #define SVC_REF(xprt, flags)						\
 	svc_ref_it(xprt, flags, __func__, __LINE__)
-#define svc_ref(xprt, flags)						\
-	svc_ref_it(xprt, flags, __func__, __LINE__)
 
-static inline void svc_release_it(struct rpc_svcxprt *xprt, u_int flags,
+static inline void svc_release_it(SVCXPRT *xprt, u_int flags,
 				  const char *tag, const int line)
 {
 	uint32_t refs = atomic_dec_uint32_t(&xprt->xp_refs);
@@ -489,10 +405,10 @@ static inline void svc_release_it(struct rpc_svcxprt *xprt, u_int flags,
 
 	/* enforce once-only semantic, trace others */
 	xp_flags = atomic_postset_uint16_t_bits(&xprt->xp_flags,
-						SVC_XPRT_FLAG_DESTROYING);
+						SVC_XPRT_FLAG_RELEASING);
 
-	if (xp_flags & SVC_XPRT_FLAG_DESTROYING) {
-		XPRT_TRACE(xprt, "ERROR! already destroying!", tag, line);
+	if (xp_flags & SVC_XPRT_FLAG_RELEASING) {
+		XPRT_TRACE(xprt, "WARNING! already destroying!", tag, line);
 		return;
 	}
 
@@ -501,25 +417,21 @@ static inline void svc_release_it(struct rpc_svcxprt *xprt, u_int flags,
 }
 #define SVC_RELEASE2(xprt, flags, tag, line)				\
 	svc_release_it(xprt, flags, __func__, __LINE__)
-#define svc_release2(xprt, flags, tag, line)				\
-	svc_release_it(xprt, flags, __func__, __LINE__)
 #define SVC_RELEASE(xprt, flags)					\
 	svc_release_it(xprt, flags, __func__, __LINE__)
-#define svc_release(xprt, flags)					\
-	svc_release_it(xprt, flags, __func__, __LINE__)
 
 /* SVC_DESTROY is SVC_RELEASE with once-only semantics.  Also, idempotent
  * SVC_XPRT_FLAG_DESTROYED indicates that more references should not be taken.
  */
-static inline void svc_destroy_it(struct rpc_svcxprt *xprt,
+static inline void svc_destroy_it(SVCXPRT *xprt,
 				  const char *tag, const int line)
 {
 	uint16_t flags = atomic_postset_uint16_t_bits(&xprt->xp_flags,
-						      SVC_XPRT_FLAG_DESTROYED);
+						      SVC_XPRT_FLAG_DESTROYING);
 
 	XPRT_TRACE(xprt, __func__, tag, line);
 
-	if (flags & SVC_XPRT_FLAG_DESTROYED) {
+	if (flags & SVC_XPRT_FLAG_DESTROYING) {
 		/* previously set, do nothing */
 		return;
 	}
@@ -528,28 +440,10 @@ static inline void svc_destroy_it(struct rpc_svcxprt *xprt,
 }
 #define SVC_DESTROY(xprt)						\
 	svc_destroy_it(xprt, __func__, __LINE__)
-#define svc_destroy(xprt)						\
-	svc_destroy_it(xprt, __func__, __LINE__)
 
-#define SVC_CONTROL(xprt, rq, in)			\
+#define SVC_CONTROL(xprt, rq, in)					\
 	(*(xprt)->xp_ops->xp_control)((xprt), (rq), (in))
 
-#define XP_LOCK_NONE    0x0000
-#define XP_LOCK_SEND    0x0001
-#define XP_LOCK_RECV    0x0002
-
-#define SVC_LOCK(xprt, flags, func, line) \
-	(*(xprt)->xp_ops->xp_lock)((xprt), (flags), (func), (line))
-
-#define svc_lock(xprt, flags, func, line) \
-	(*(xprt)->xp_ops->xp_lock)((xprt), (flags), (func), (line))
-
-#define SVC_UNLOCK(xprt, flags, func, line) \
-	(*(xprt)->xp_ops->xp_unlock)((xprt), (flags), (func), (line))
-
-#define svc_unlock(xprt, flags, func, line) \
-	(*(xprt)->xp_ops->xp_unlock)((xprt), (flags), (func), (line))
-
 /*
  * Service init (optional).
  */
@@ -577,7 +471,7 @@ __END_DECLS
  */
 __BEGIN_DECLS
 extern bool svc_reg(SVCXPRT *, const rpcprog_t, const rpcvers_t,
-		    void (*)(struct svc_req *, SVCXPRT *),
+		    void (*)(struct svc_req *),
 		    const struct netconfig *);
 __END_DECLS
 /*
@@ -590,32 +484,22 @@ __END_DECLS
 __BEGIN_DECLS
 extern void svc_unreg(const rpcprog_t, const rpcvers_t);
 __END_DECLS
-/*
- * Transport registration.
- *
- * xprt_register(xprt)
- * SVCXPRT *xprt;
- */
-__BEGIN_DECLS
-extern void xprt_register(SVCXPRT *);
-__END_DECLS
-/*
- * Transport un-register
- *
- * xprt_unregister(xprt)
- * SVCXPRT *xprt;
- */
-__BEGIN_DECLS
-extern void xprt_unregister(SVCXPRT *);
-__END_DECLS
 /*
  * This is used to set local and remote addresses in a way legacy
  * apps can deal with, at the same time setting up a corresponding
  * netbuf -- with no alloc/free needed.
  */
 __BEGIN_DECLS
-extern void __rpc_set_address(struct rpc_address *,
-			      const struct sockaddr_storage *, socklen_t);
+extern u_int __rpc_address_port(struct rpc_address *);
+extern void __rpc_address_set_length(struct rpc_address *, socklen_t);
+
+static inline void
+__rpc_address_setup(struct rpc_address *rpca)
+{
+	rpca->nb.buf = &rpca->ss;
+	rpca->nb.len =
+	rpca->nb.maxlen = sizeof(struct sockaddr_storage);
+}
 __END_DECLS
 /*
  * When the service routine is called, it must first check to see if it
@@ -643,15 +527,14 @@ __END_DECLS
  * deadlock the caller and server processes!
  */
 __BEGIN_DECLS
-extern bool svc_sendreply(SVCXPRT *, struct svc_req *, xdrproc_t,
-			  void *);
-extern void svcerr_decode(SVCXPRT *, struct svc_req *);
-extern void svcerr_weakauth(SVCXPRT *, struct svc_req *);
-extern void svcerr_noproc(SVCXPRT *, struct svc_req *);
-extern void svcerr_progvers(SVCXPRT *, struct svc_req *, rpcvers_t, rpcvers_t);
-extern void svcerr_auth(SVCXPRT *, struct svc_req *, enum auth_stat);
-extern void svcerr_noprog(SVCXPRT *, struct svc_req *);
-extern void svcerr_systemerr(SVCXPRT *, struct svc_req *);
+extern enum xprt_stat svc_sendreply(struct svc_req *);
+extern enum xprt_stat svcerr_decode(struct svc_req *);
+extern enum xprt_stat svcerr_weakauth(struct svc_req *);
+extern enum xprt_stat svcerr_noproc(struct svc_req *);
+extern enum xprt_stat svcerr_progvers(struct svc_req *, rpcvers_t, rpcvers_t);
+extern enum xprt_stat svcerr_auth(struct svc_req *, enum auth_stat);
+extern enum xprt_stat svcerr_noprog(struct svc_req *);
+extern enum xprt_stat svcerr_systemerr(struct svc_req *);
 extern int rpc_reg(rpcprog_t, rpcvers_t, rpcproc_t, char *(*)(char *),
 		   xdrproc_t, xdrproc_t, char *);
 __END_DECLS
@@ -662,20 +545,16 @@ __END_DECLS
 __BEGIN_DECLS
 extern void rpctest_service(void);
 __END_DECLS
-__BEGIN_DECLS
-extern void svc_getreq(int);
-extern void svc_getreqset(fd_set *);
-extern void svc_getreq_common(int);
-struct pollfd;
-extern void svc_getreq_poll(struct pollfd *, int);
-extern void svc_run(void);
-extern void svc_exit(void);
-__END_DECLS
 /*
  * Socket to use on svcxxx_ncreate call to get default socket
  */
 #define RPC_ANYSOCK -1
 #define RPC_ANYFD RPC_ANYSOCK
+/*
+ * Usual sizes for svcxxx_ncreate
+ */
+#define RPC_MAXDATA_DEFAULT (8192)
+#define RPC_MAXDATA_LEGACY (262144)
 /*
  * These are the existing service side transport implementations
  */
@@ -683,7 +562,7 @@ __BEGIN_DECLS
 /*
  * Transport independent svc_create routine.
  */
-extern int svc_ncreate(void (*)(struct svc_req *, SVCXPRT *), const rpcprog_t,
+extern int svc_ncreate(void (*)(struct svc_req *), const rpcprog_t,
 		       const rpcvers_t, const char *);
 /*
  *      void (*dispatch)();             -- dispatch routine
@@ -697,7 +576,7 @@ extern int svc_ncreate(void (*)(struct svc_req *, SVCXPRT *), const rpcprog_t,
  * instead of a nettype.
  */
 
-extern SVCXPRT *svc_tp_ncreate(void (*)(struct svc_req *, SVCXPRT *),
+extern SVCXPRT *svc_tp_ncreate(void (*)(struct svc_req *),
 			       const rpcprog_t, const rpcvers_t,
 			       const struct netconfig *);
 /*
@@ -720,20 +599,25 @@ extern SVCXPRT *svc_tli_ncreate(const int, const struct netconfig *,
  *      const u_int sendsz;             -- max sendsize
  *      const u_int recvsz;             -- max recvsize
  */
+__END_DECLS
 
 /*
  * Connectionless and connectionful create routines
  */
 
-extern SVCXPRT *svc_vc_ncreate(const int, const u_int, const u_int);
-/*
- *      const int fd;                           -- open connection end point
- *      const u_int sendsize;                   -- max send size
- *      const u_int recvsize;                   -- max recv size
- */
+/* uint16_t actually used */
+#define SVC_CREATE_FLAG_NONE		SVC_XPRT_FLAG_NONE
+#define SVC_CREATE_FLAG_CLOSE		SVC_XPRT_FLAG_CLOSE
 
-extern SVCXPRT *svc_vc_ncreate2(const int, const u_int, const u_int,
-				const u_int);
+/* uint32_t instructions */
+#define SVC_CREATE_FLAG_LISTEN		0x20000000
+#define SVC_CREATE_FLAG_XPRT_DOREG	0x80000000
+#define SVC_CREATE_FLAG_XPRT_NOREG	0x08000000
+
+__BEGIN_DECLS
+
+extern SVCXPRT *svc_vc_ncreatef(const int, const u_int, const u_int,
+				const uint32_t);
 /*
  *      const int fd;                           -- open connection end point
  *      const u_int sendsize;                   -- max send size
@@ -741,68 +625,44 @@ extern SVCXPRT *svc_vc_ncreate2(const int, const u_int, const u_int,
  *      const u_int flags;                      -- flags
  */
 
-__END_DECLS
-#define SVC_VC_CREATE_NONE             0x0000
-#define SVC_VC_CREATE_BOTHWAYS         0x0001
-#define SVC_VC_CREATE_ONEWAY           0x0002	/* !bothways */
-#define SVC_VC_CREATE_DISPOSE          0x0004	/* !bothways */
-#define SVC_VC_CREATE_XPRT_NOREG       0x0008
-#define SVC_VC_CREATE_LISTEN           0x0010
+static inline SVCXPRT *
+svc_vc_ncreate(const int fd, const u_int sendsize, const u_int recvsize)
+{
+	return (svc_vc_ncreatef(fd, sendsize, recvsize, SVC_CREATE_FLAG_CLOSE));
+}
 
-__BEGIN_DECLS
-/*
- * Create a client handle from an active service transport handle.
- */
-extern CLIENT *clnt_vc_ncreate_svc(SVCXPRT *, const rpcprog_t, const rpcvers_t,
-				   const uint32_t);
-/*
- *      SVCXPRT *xprt;                          -- active service xprt
- *      const rpcprog_t prog;                   -- RPC program number
- *      const rpcvers_t vers;                   -- RPC program version
- */
-
-__END_DECLS
-__BEGIN_DECLS
-/*
- * Create an RPC SVCXPRT handle from an active client transport
- * handle, i.e., to service RPC requests
- */
-extern SVCXPRT *svc_vc_ncreate_clnt(CLIENT *, u_int, u_int, const uint32_t);
-/*
- *
- * CLIENT *cl;                                  -- connected client
- * const u_int sendsize;                        -- max send size
- * const u_int recvsize;                        -- max recv size
- * const uint32_t flags;                        -- flags
- */
-
-/*
- * Added for compatibility to old rpc 4.0. Obsoleted by svc_vc_create().
- */
-extern SVCXPRT *svcunix_ncreate(int, u_int, u_int, char *);
-
-extern SVCXPRT *svc_dg_ncreate(const int, const u_int, const u_int);
-/*
- * const int fd;                                -- open connection
- * const u_int sendsize;                        -- max send size
- * const u_int recvsize;                        -- max recv size
- */
-
-/*
- * the routine takes any *open* connection
- * descriptor as its first input and is used for open connections.
- */
-extern SVCXPRT *svc_fd_ncreate(const int, const u_int, const u_int);
+extern SVCXPRT *svc_dg_ncreatef(const int, const u_int, const u_int,
+				const uint32_t);
 /*
  *      const int fd;                           -- open connection end point
  *      const u_int sendsize;                   -- max send size
  *      const u_int recvsize;                   -- max recv size
+ *      const uint32_t flags;                   -- flags
  */
 
+static inline SVCXPRT *
+svc_dg_ncreate(const int fd, const u_int sendsize, const u_int recvsize)
+{
+	return (svc_dg_ncreatef(fd, sendsize, recvsize, SVC_CREATE_FLAG_CLOSE));
+}
+
 /*
- * Added for compatibility to old rpc 4.0. Obsoleted by svc_fd_create().
+ * the routine takes any *open* connection
  */
-extern SVCXPRT *svcunixfd_ncreate(int, u_int, u_int);
+extern SVCXPRT *svc_fd_ncreatef(const int, const u_int, const u_int,
+				const uint32_t);
+/*
+ *      const int fd;                           -- open connection end point
+ *      const u_int sendsize;                   -- max send size
+ *      const u_int recvsize;                   -- max recv size
+ *      const uint32_t flags;                   -- flags
+ */
+
+static inline SVCXPRT *
+svc_fd_ncreate(const int fd, const u_int sendsize, const u_int recvsize)
+{
+	return (svc_fd_ncreatef(fd, sendsize, recvsize, SVC_CREATE_FLAG_NONE));
+}
 
 /*
  * Memory based rpc (for speed check and testing)
@@ -810,21 +670,14 @@ extern SVCXPRT *svcunixfd_ncreate(int, u_int, u_int);
 extern SVCXPRT *svc_raw_ncreate(void);
 
 /*
- * RDMA based rpc
+ * RPC over RDMA
  */
-extern SVCXPRT *svc_rdma_ncreate(void *arg, const u_int sendsize,
-				 const u_int recvsize, const u_int flags);
-
 struct rpc_rdma_attr {
 	char *statistics_prefix;
 	/* silly char * to pass to rdma_getaddrinfo() */
 	char *node;			/**< remote peer's hostname */
 	char *port;			/**< service port (or name) */
 
-	void (*disconnect_cb)(SVCXPRT *);
-	enum xprt_stat (*request_cb)(void *, SVCXPRT *);
-
-	u_long timeout;			/**< ms wait for events */
 	u_int sq_depth;			/**< depth of Send Queue */
 	u_int max_send_sge;		/**< s/g elements per send */
 	u_int rq_depth;			/**< depth of Receive Queue. */
@@ -832,40 +685,35 @@ struct rpc_rdma_attr {
 
 	u_int backlog;			/**< connection backlog */
 	u_int credits;			/**< parallel messages */
-	u_int worker_count;		/**< worker threads */
-	u_int worker_queue_size;	/**< worker data queue */
 
 	bool destroy_on_disconnect;	/**< should perform cleanup */
 	bool use_srq;			/**< server use srq? */
 };
-extern SVCXPRT *rpc_rdma_create(struct rpc_rdma_attr *arg);
 
+extern SVCXPRT *rpc_rdma_ncreatef(const struct rpc_rdma_attr *, const u_int,
+				  const u_int, const uint32_t);
 /*
- * Getreq plug-out prototype
+ *      const struct rpc_rdma_attr;             -- RDMA configuration
+ *      const u_int sendsize;                   -- max send size
+ *      const u_int recvsize;                   -- max recv size
+ *      const uint32_t flags;                   -- flags
  */
-extern bool svc_getreq_default(SVCXPRT *);
 
-/*
- * Dispatch plug-out prototype
- */
-extern void svc_dispatch_default(SVCXPRT *, struct rpc_msg **);
+static inline SVCXPRT *
+svc_rdma_ncreate(const struct rpc_rdma_attr *xa, const u_int sendsize,
+		 const u_int recvsize)
+{
+	return rpc_rdma_ncreatef(xa, sendsize, recvsize, SVC_CREATE_FLAG_CLOSE);
+}
 
 /*
  * Convenience functions for implementing these
  */
 extern bool svc_validate_xprt_list(SVCXPRT *);
-extern struct rpc_msg *alloc_rpc_msg(void);
-extern void free_rpc_msg(struct rpc_msg *);
-
-/*
- * svc_dg_enable_cache() enables the cache on dg transports.
- */
-int svc_dg_enablecache(SVCXPRT *, const u_int);
 
 int __rpc_get_local_uid(SVCXPRT *, uid_t *);
 
 __END_DECLS
 /* for backward compatibility */
-#include <rpc/svc_soc.h>
 #include <rpc/tirpc_compat.h>
 #endif				/* !_TIRPC_SVC_H */
