@@ -2,6 +2,7 @@
 
 /*
  * Copyright (c) 2009, Sun Microsystems, Inc.
+ * Copyright (c) 2012-2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -124,36 +125,40 @@ typedef int32_t rpc_inline_t;
 #define TIRPC_DEBUG_FLAG_NONE           0x0000000
 #define TIRPC_DEBUG_FLAG_ERROR          0x0000001
 #define TIRPC_DEBUG_FLAG_EVENT          0x0000002
-#define TIRPC_DEBUG_FLAG_REFCNT         0x0000004
+#define TIRPC_DEBUG_FLAG_WARN           0x0000004
 #define TIRPC_DEBUG_FLAG_LOCK           0x0000008
-#define TIRPC_DEBUG_FLAG_MEM            0x0000010
+#define TIRPC_DEBUG_FLAG_REFCNT         0x0000010
 #define TIRPC_DEBUG_FLAG_RBTREE         0x0000020
 #define TIRPC_DEBUG_FLAG_RPCSEC_GSS     0x0000040
 #define TIRPC_DEBUG_FLAG_AUTH           0x0000080
-#define TIRPC_DEBUG_FLAG_CLNT_BCAST     0x0000100
-#define TIRPC_DEBUG_FLAG_CLNT_DG        0x0000200
-#define TIRPC_DEBUG_FLAG_CLNT_GEN       0x0000400
-#define TIRPC_DEBUG_FLAG_CLNT_RAW       0x0000800
-#define TIRPC_DEBUG_FLAG_CLNT_RDMA      0x0001000
-#define TIRPC_DEBUG_FLAG_CLNT_SCTP      0x0002000
-#define TIRPC_DEBUG_FLAG_CLNT_VC        0x0004000
-#define TIRPC_DEBUG_FLAG_SVC            0x0008000
+#define TIRPC_DEBUG_FLAG_CLNT_DG        0x0000100
+#define TIRPC_DEBUG_FLAG_CLNT_RDMA      0x0000200
+#define TIRPC_DEBUG_FLAG_CLNT_SCTP      0x0000400
+#define TIRPC_DEBUG_FLAG_CLNT_VC        0x0000800
+#define TIRPC_DEBUG_FLAG_CLNT_BCAST     0x0001000
+#define TIRPC_DEBUG_FLAG_CLNT_RAW       0x0002000
+#define TIRPC_DEBUG_FLAG_CLNT_REQ       0x0004000
+#define TIRPC_DEBUG_FLAG_CLNT           0x0008000
 #define TIRPC_DEBUG_FLAG_SVC_DG         0x0010000
-#define TIRPC_DEBUG_FLAG_SVC_RQST       0x0020000
-#define TIRPC_DEBUG_FLAG_SVC_XPRT       0x0040000
-#define TIRPC_DEBUG_FLAG_SVC_RDMA       0x0080000
-#define TIRPC_DEBUG_FLAG_SVC_SCTP       0x0100000
-#define TIRPC_DEBUG_FLAG_SVC_VC         0x0200000
-#define TIRPC_DEBUG_FLAG_XDRREC         0x0400000
+#define TIRPC_DEBUG_FLAG_SVC_RDMA       0x0020000
+#define TIRPC_DEBUG_FLAG_SVC_SCTP       0x0040000
+#define TIRPC_DEBUG_FLAG_SVC_VC         0x0080000
+#define TIRPC_DEBUG_FLAG_SVC_RQST       0x0100000
+#define TIRPC_DEBUG_FLAG_SVC_XPRT       0x0200000
+#define TIRPC_DEBUG_FLAG_SVC            0x0400000
 #define TIRPC_DEBUG_FLAG_XDR            0x0800000
-#define TIRPC_DEBUG_FLAG_RPC_CTX        0x1000000
-#define TIRPC_DEBUG_FLAG_RPC_CACHE      0x2000000
-#define TIRPC_DEBUG_FLAG_RPC_MSG        0x4000000
-#define TIRPC_DEBUG_FLAG_RPC_RDMA       0x8000000
+#define TIRPC_DEBUG_FLAG_WORKER         0x1000000
+#define TIRPC_DEBUG_FLAG_RPC_MSG        0x2000000
+#define TIRPC_DEBUG_FLAG_RPC_RDMA       0x4000000
+#define TIRPC_DEBUG_FLAG_XDR_RDMA       0x8000000
 
 /* or symbolic names for default */
 #define TIRPC_DEBUG_FLAG_DEFAULT \
-	(TIRPC_DEBUG_FLAG_ERROR | TIRPC_DEBUG_FLAG_EVENT)
+	(TIRPC_DEBUG_FLAG_ERROR | \
+	 TIRPC_DEBUG_FLAG_EVENT | \
+	 TIRPC_DEBUG_FLAG_WARN)
+
+#define TIRPC_DEBUG_FLAG_CLNT_RPCB      (TIRPC_DEBUG_FLAG_CLNT)
 
 typedef void *(*mem_1_size_t) (size_t,
 	     const char *file, int line, const char *function);
@@ -163,6 +168,7 @@ typedef void *(*mem_p_size_t) (void *, size_t,
 	     const char *file, int line, const char *function);
 typedef void (*mem_free_size_t) (void *, size_t);
 typedef void (*mem_format_t) (const char *fmt, ...);
+typedef void (*mem_char_t) (const char *);
 
 /*
  * Package params support
@@ -170,6 +176,7 @@ typedef void (*mem_format_t) (const char *fmt, ...);
 typedef struct tirpc_pkg_params {
 	uint32_t debug_flags;
 	uint32_t other_flags;
+	mem_char_t	thread_name_;
 	mem_format_t	warnx_;
 	mem_free_size_t	free_size_;
 	mem_1_size_t	malloc_;
