@@ -111,19 +111,20 @@ enum xdr_op {
 
 /* XDR buffer vector descriptors */
 typedef struct xdr_vio {
-	void *vio_base;
-	void *vio_head;	/* minimum vio_tail (header offset) */
-	void *vio_tail;
-	void *vio_wrap;	/* maximum vio_tail */
+	uint8_t *vio_base;
+	uint8_t *vio_head;	/* minimum vio_tail (header offset) */
+	uint8_t *vio_tail;
+	uint8_t *vio_wrap;	/* maximum vio_tail */
 } xdr_vio;
 
 /* vio_wrap >= vio_tail >= vio_head >= vio_base */
 
 #define UIO_FLAG_NONE		0x0000
-#define UIO_FLAG_GIFT		0x0001
+#define UIO_FLAG_BUFQ		0x0001
 #define UIO_FLAG_FREE		0x0002
-#define UIO_FLAG_BUFQ		0x0004
-#define UIO_FLAG_REALLOC	0x0008
+#define UIO_FLAG_GIFT		0x0004
+#define UIO_FLAG_MORE		0x0008
+#define UIO_FLAG_REALLOC	0x0010
 
 struct xdr_uio;
 typedef void (*xdr_uio_release)(struct xdr_uio *, u_int);
@@ -147,9 +148,10 @@ typedef struct xdr_uio {
 #define XDR_PUTBUFS_FLAG_NONE    0x0000
 #define XDR_PUTBUFS_FLAG_RDNLY   0x0001
 
-#define XDR_FLAG_NONE    0x0000
-#define XDR_FLAG_CKSUM   0x0001
-#define XDR_FLAG_VIO     0x0002
+#define XDR_FLAG_NONE		0x0000
+#define XDR_FLAG_CKSUM		0x0001
+#define XDR_FLAG_FREE		0x0002
+#define XDR_FLAG_VIO		0x0004
 
 /*
  * The XDR handle.
@@ -159,12 +161,11 @@ typedef struct xdr_uio {
  * XXX: w/64-bit pointers, u_int not enough!
  */
 typedef struct rpc_xdr {
-	enum xdr_op x_op;  /* operation; fast additional param */
 	const struct xdr_ops {
-		/* get a long from underlying stream */
-		bool (*x_getlong)(struct rpc_xdr *, long *);
-		/* put a long to " */
-		bool (*x_putlong)(struct rpc_xdr *, const long *);
+		/* get 4 unsigned bytes from underlying stream */
+		bool (*x_getunit)(struct rpc_xdr *, uint32_t *);
+		/* put 4 unsigned bytes to underlying stream */
+		bool (*x_putunit)(struct rpc_xdr *, const uint32_t);
 		/* get some bytes from " */
 		bool (*x_getbytes)(struct rpc_xdr *, char *, u_int);
 		/* put some bytes to " */
@@ -173,8 +174,6 @@ typedef struct rpc_xdr {
 		u_int (*x_getpostn)(struct rpc_xdr *);
 		/* lets you reposition the stream */
 		bool (*x_setpostn)(struct rpc_xdr *, u_int);
-		/* buf quick ptr to buffered data */
-		int32_t *(*x_inline)(struct rpc_xdr *, u_int);
 		/* free private resources of this xdr_stream */
 		void (*x_destroy)(struct rpc_xdr *);
 		bool (*x_control)(struct rpc_xdr *, int, void *);
@@ -185,11 +184,12 @@ typedef struct rpc_xdr {
 	void *x_public; /* users' data */
 	void *x_private; /* pointer to private data */
 	void *x_lib[2]; /* RPC library private */
-	void *x_data;  /* private used for position inline */
+	uint8_t *x_data;  /* private used for position inline */
 	void *x_base;  /* private used for position info */
 	struct xdr_vio x_v; /* private buffer vector */
 	u_int x_handy; /* extra private word */
 	u_int x_flags; /* shared flags */
+	enum xdr_op x_op;  /* operation; fast additional param */
 } XDR;
 
 #define XDR_VIO(x) ((xdr_vio *)((x)->x_base))
@@ -247,31 +247,35 @@ typedef bool(*xdrproc_t) (XDR *, ...);
 static inline bool
 xdr_getlong(XDR *xdrs, long *lp)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
+	uint32_t u;
+	bool b;
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data)
-				+ sizeof(uint32_t)) > xdrs->x_v.vio_tail)) {
-		return (*xdrs->x_ops->x_getlong)(xdrs, lp);
+	if (future <= xdrs->x_v.vio_tail) {
+		*lp = (long)ntohl(*((uint32_t *) (xdrs->x_data)));
+		xdrs->x_data = future;
+		return (true);
 	}
-	*lp = (long)ntohl(*((uint32_t *) (xdrs->x_data)));
-	xdrs->x_data = future;
-	return (true);
+
+	b = (*xdrs->x_ops->x_getunit)(xdrs, &u);
+	if (b) {
+		*lp = (int32_t)u;	/* sign extends */
+	}
+	return b;
 }
 
 static inline bool
 xdr_putlong(XDR *xdrs, const long *lp)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data) + sizeof(uint32_t))
-			> xdrs->x_v.vio_wrap)) {
-		return (*xdrs->x_ops->x_putlong)(xdrs, lp);
+	if (future <= xdrs->x_v.vio_wrap) {
+		*((int32_t *) (xdrs->x_data)) =
+			(int32_t) htonl((int32_t) (*lp));
+		xdrs->x_data = future;
+		return (true);
 	}
-	*((int32_t *) (xdrs->x_data)) = (int32_t) htonl((int32_t) (*lp));
-	xdrs->x_data = future;
-	return (true);
+	return (*xdrs->x_ops->x_putunit)(xdrs, (uint32_t) (*lp));
 }
 
 #define XDR_GETLONG(xdrs, lp) xdr_getlong(xdrs, lp)
@@ -307,11 +311,6 @@ xdr_putlong(XDR *xdrs, const long *lp)
 #define xdr_setpos(xdrs, pos)			\
 	(*(xdrs)->x_ops->x_setpostn)(xdrs, pos)
 
-#define XDR_INLINE(xdrs, len)			\
-	(*(xdrs)->x_ops->x_inline)(xdrs, len)
-#define xdr_inline(xdrs, len)			\
-	(*(xdrs)->x_ops->x_inline)(xdrs, len)
-
 #define XDR_DESTROY(xdrs)	     \
 	if ((xdrs)->x_ops->x_destroy)			\
 		(*(xdrs)->x_ops->x_destroy)(xdrs)
@@ -354,17 +353,49 @@ struct xdr_discrim {
  * In-line routines for fast encode/decode of primitive data types.
  * Caveat emptor: these use single memory cycles to get the
  * data from the underlying buffer, and will fail to operate
- * properly where the data is not aligned.  The standard way to use these
- * is to say:
- * if ((buf = XDR_INLINE(xdrs, count)) == NULL)
- *  return (false);
- * <<< macro calls >>>
+ * properly where the data is not aligned.  The standard way to use
+ * these is to say:
+ *      if ((buf = xdr_inline_decode(xdrs, count)) == NULL)
+ *              return (FALSE);
+ *      <<< IXDR_GET_* macro calls >>>
+ *      if ((buf = xdr_inline_encode(xdrs, count)) == NULL)
+ *              return (FALSE);
+ *      <<< IXDR_PUT_* macro calls >>>
  * where ``count'' is the number of bytes of data occupied
  * by the primitive data types.
  *
  * N.B. and frozen for all time: each data type here uses 4 bytes
  * of external representation.
  */
+static inline int32_t *
+xdr_inline_decode(XDR *xdrs, size_t count)
+{
+	int32_t *buf = (int32_t *)xdrs->x_data;
+	uint8_t *future = xdrs->x_data + count;
+
+	/* re-consuming bytes in a stream
+	 * (after SETPOS/rewind) */
+	if (future <= xdrs->x_v.vio_tail) {
+		xdrs->x_data = future;
+		return (buf);
+	}
+	return (NULL);
+}
+
+static inline int32_t *
+xdr_inline_encode(XDR *xdrs, size_t count)
+{
+	int32_t *buf = (int32_t *)xdrs->x_data;
+	uint8_t *future = xdrs->x_data + count;
+
+	if (future <= xdrs->x_v.vio_wrap) {
+		xdrs->x_data = future;
+		xdr_tail_update(xdrs);
+		return (buf);
+	}
+	return (NULL);
+}
+
 #define IXDR_GET_INT32(buf)  ((int32_t)ntohl((u_int32_t)*(buf)++))
 #define IXDR_PUT_INT32(buf, v)  (*(buf)++ = (int32_t)htonl((u_int32_t)v))
 #define IXDR_GET_U_INT32(buf)  ((u_int32_t)IXDR_GET_INT32(buf))
@@ -406,42 +437,31 @@ struct xdr_discrim {
 static inline bool
 xdr_getuint32(XDR *xdrs, uint32_t *ip)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data) + sizeof(uint32_t))
-			> xdrs->x_v.vio_tail)) {
-		long l;
-
-		if (!(*xdrs->x_ops->x_getlong)(xdrs, &l))
-			return (false);
-		*ip = (uint32_t) l;
+	if (future <= xdrs->x_v.vio_tail) {
+		*ip = ntohl(*((uint32_t *) (xdrs->x_data)));
+		xdrs->x_data = future;
 		return (true);
 	}
-	*ip = ntohl(*((uint32_t *) (xdrs->x_data)));
-	xdrs->x_data = future;
-	return (true);
+	return (*xdrs->x_ops->x_getunit)(xdrs, ip);
 }
 
 static inline bool
-xdr_putuint32(XDR *xdrs, uint32_t *ip)
+xdr_putuint32(XDR *xdrs, uint32_t v)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data) + sizeof(uint32_t))
-			> xdrs->x_v.vio_wrap)) {
-		long l = (long)*ip;
-
-		return (*xdrs->x_ops->x_putlong)(xdrs, &l);
+	if (future <= xdrs->x_v.vio_wrap) {
+		*((int32_t *) (xdrs->x_data)) = htonl(v);
+		xdrs->x_data = future;
+		return (true);
 	}
-	*((int32_t *) (xdrs->x_data)) = htonl(*ip);
-	xdrs->x_data = future;
-	return (true);
+	return (*xdrs->x_ops->x_putunit)(xdrs, v);
 }
 
 #define XDR_GETUINT32(xdrs, uint32p) xdr_getuint32(xdrs, uint32p)
-#define XDR_PUTUINT32(xdrs, uint32p) xdr_putuint32(xdrs, uint32p)
+#define XDR_PUTUINT32(xdrs, uint32v) xdr_putuint32(xdrs, uint32v)
 
 static inline bool
 xdr_getint32(XDR *xdrs, int32_t *ip)
@@ -450,49 +470,43 @@ xdr_getint32(XDR *xdrs, int32_t *ip)
 }
 
 static inline bool
-xdr_putint32(XDR *xdrs, int32_t *ip)
+xdr_putint32(XDR *xdrs, int32_t v)
 {
-	return xdr_putuint32(xdrs, (uint32_t *)ip);
+	return xdr_putuint32(xdrs, v);
 }
 
 #define XDR_GETINT32(xdrs, int32p) xdr_getint32(xdrs, int32p)
-#define XDR_PUTINT32(xdrs, int32p) xdr_putint32(xdrs, int32p)
+#define XDR_PUTINT32(xdrs, int32v) xdr_putint32(xdrs, int32v)
 
 static inline bool
 xdr_getuint16(XDR *xdrs, uint16_t *ip)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
+	uint32_t u;
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data) + sizeof(uint32_t))
-			> xdrs->x_v.vio_tail)) {
-		long l;
-
-		if (!(*xdrs->x_ops->x_getlong)(xdrs, &l))
-			return (false);
-		*ip = (uint16_t) l;
+	if (future <= xdrs->x_v.vio_tail) {
+		*ip = (uint16_t)ntohl(*((uint32_t *) (xdrs->x_data)));
+		xdrs->x_data = future;
 		return (true);
 	}
-	*ip = (uint16_t)ntohl(*((uint32_t *) (xdrs->x_data)));
-	xdrs->x_data = future;
-	return (true);
+	if ((*xdrs->x_ops->x_getunit)(xdrs, &u)) {
+		*ip = (uint16_t) u;
+		return (true);
+	}
+	return (false);
 }
 
 static inline bool
 xdr_putuint16(XDR *xdrs, uint32_t uint16v)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data) + sizeof(uint32_t))
-			> xdrs->x_v.vio_wrap)) {
-		long l = (long)uint16v;
-
-		return (*xdrs->x_ops->x_putlong)(xdrs, &l);
+	if (future <= xdrs->x_v.vio_wrap) {
+		*((int32_t *) (xdrs->x_data)) = htonl(uint16v);
+		xdrs->x_data = future;
+		return (true);
 	}
-	*((int32_t *) (xdrs->x_data)) = htonl(uint16v);
-	xdrs->x_data = future;
-	return (true);
+	return (*xdrs->x_ops->x_putunit)(xdrs, (uint32_t)uint16v);
 }
 
 #define XDR_GETUINT16(xdrs, uint16p) xdr_getuint16(xdrs, uint16p)
@@ -517,38 +531,32 @@ xdr_putint16(XDR *xdrs, int32_t int16v)
 static inline bool
 xdr_getuint8(XDR *xdrs, uint8_t *ip)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
+	uint32_t u;
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data) + sizeof(uint32_t))
-			> xdrs->x_v.vio_tail)) {
-		long l;
-
-		if (!(*xdrs->x_ops->x_getlong)(xdrs, &l))
-			return (false);
-		*ip = (uint8_t) l;
+	if (future <= xdrs->x_v.vio_tail) {
+		*ip = (uint8_t)ntohl(*((uint32_t *) (xdrs->x_data)));
+		xdrs->x_data = future;
 		return (true);
 	}
-	*ip = (uint8_t)ntohl(*((uint32_t *) (xdrs->x_data)));
-	xdrs->x_data = future;
-	return (true);
+	if ((*xdrs->x_ops->x_getunit)(xdrs, &u)) {
+		*ip = (uint8_t) u;
+		return (true);
+	}
+	return (false);
 }
 
 static inline bool
 xdr_putuint8(XDR *xdrs, uint32_t uint8v)
 {
-	void *future;
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
 
-	if (!(xdrs->x_flags & XDR_FLAG_VIO)
-		|| unlikely((future = char_ptr(xdrs->x_data) + sizeof(uint32_t))
-			> xdrs->x_v.vio_wrap)) {
-		long l = (long)uint8v;
-
-		return (*xdrs->x_ops->x_putlong)(xdrs, &l);
+	if (future <= xdrs->x_v.vio_wrap) {
+		*((int32_t *) (xdrs->x_data)) = htonl(uint8v);
+		xdrs->x_data = future;
+		return (true);
 	}
-	*((int32_t *) (xdrs->x_data)) = htonl(uint8v);
-	xdrs->x_data = future;
-	return (true);
+	return (*xdrs->x_ops->x_putunit)(xdrs, (uint32_t)uint8v);
 }
 
 #define XDR_GETUINT8(xdrs, uint8p) xdr_getuint8(xdrs, uint8p)
@@ -624,9 +632,6 @@ extern bool xdr_uint64_t(XDR *, u_int64_t *);
 extern bool xdr_bool(XDR *, bool_t *);
 extern bool xdr_enum(XDR *, enum_t *);
 extern bool xdr_array(XDR *, char **, u_int *, u_int, u_int, xdrproc_t);
-extern bool xdr_bytes(XDR *, char **, u_int *, u_int);
-extern bool xdr_opaque(XDR *, char *, u_int);
-extern bool xdr_string(XDR *, char **, u_int);
 extern bool xdr_union(XDR *, enum_t *, char *, const struct xdr_discrim *,
 		      xdrproc_t);
 extern bool xdr_char(XDR *, char *);
@@ -681,23 +686,6 @@ extern void xdrmem_ncreate(XDR *, char *, u_int, enum xdr_op);
 /* intrinsic checksum (be careful) */
 extern uint64_t xdrmem_cksum(XDR *, u_int);
 
-/* XDR using stdio library */
-extern void xdrstdio_create(XDR *, FILE *, enum xdr_op);
-
-/* XDR pseudo records for tcp */
-extern void xdrrec_create(XDR *, u_int, u_int, void *,
-			  int (*)(XDR *, void *, void *, int),
-			  int (*)(XDR *, void *, void *, int));
-
-/* make end of xdr record */
-extern bool xdrrec_endofrecord(XDR *, bool);
-
-/* move to beginning of next record */
-extern bool xdrrec_skiprecord(XDR *);
-
-/* true if no more input */
-extern bool xdrrec_eof(XDR *);
-
 __END_DECLS
 /* For backward compatibility */
 #include <rpc/tirpc_compat.h>
