@@ -91,11 +91,11 @@ print_rpc_gss_sec(struct rpc_gss_sec *ptr)
 	int i;
 	char *p;
 
-	log_debug("rpc_gss_sec:");
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "rpc_gss_sec:");
 	if (ptr->mech == NULL)
-		log_debug("NULL gss_OID mech");
+		__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "NULL gss_OID mech");
 	else {
-		__warnx(TIRPC_DEBUG_FLAG_AUTH, "     mechanism_OID: {");
+		__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "     mechanism_OID: {");
 		p = (char *)ptr->mech->elements;
 		for (i = 0; i < ptr->mech->length; i++)
 			/* First byte of OIDs encoded to save a byte */
@@ -115,23 +115,24 @@ print_rpc_gss_sec(struct rpc_gss_sec *ptr)
 					first = -1;
 					second = -1;
 				}
-				__warnx(TIRPC_DEBUG_FLAG_AUTH, " %u %u", first,
-					second);
+				__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, " %u %u",
+					first, second);
 				p++;
 			} else {
-				__warnx(TIRPC_DEBUG_FLAG_AUTH, " %u",
+				__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, " %u",
 					(unsigned char)*p++);
 			}
-		__warnx(" }\n");
+		__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, " }");
 	}
-	__warnx(TIRPC_DEBUG_FLAG_AUTH, "     qop: %d\n", ptr->qop);
-	__warnx(TIRPC_DEBUG_FLAG_AUTH, "     service: %d\n", ptr->svc);
-	__warnx(TIRPC_DEBUG_FLAG_AUTH, "     cred: %p\n", ptr->cred);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "     qop: %d", ptr->qop);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "     service: %d", ptr->svc);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "     cred: %p", ptr->cred);
 }
 #endif /*DEBUG*/
+
 struct rpc_gss_data {
+	AUTH gd_auth;
 	mutex_t lock;
-	bool established;	/* context established */
 	gss_buffer_desc gc_wire_verf;	/* save GSS_S_COMPLETE NULL RPC verfier
 					 * to process at end of context
 					 * negotiation*/
@@ -141,48 +142,44 @@ struct rpc_gss_data {
 	gss_ctx_id_t ctx;	/* context id */
 	struct rpc_gss_cred gc;	/* client credentials */
 	u_int win;		/* sequence window */
+	bool established;	/* context established */
 };
+#define AUTH_PRIVATE(p) (opr_containerof((p), struct rpc_gss_data, gd_auth))
 
-#define AUTH_PRIVATE(auth) ((struct rpc_gss_data *)auth->ah_private)
-
-static struct timeval AUTH_TIMEOUT = { 25, 0 };
+/* retry timeout default to the moon and back */
+static const struct timespec to = { 3, 0 };
 
 AUTH *
 authgss_ncreate(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 {
-	AUTH *auth;
-	struct rpc_gss_data *gd;
+	struct rpc_gss_data *gd = mem_alloc(sizeof(*gd));
+	AUTH *auth = &gd->gd_auth;
 	OM_uint32 maj_stat;
 	OM_uint32 min_stat = 0;
 
-	log_debug("in authgss_ncreate()");
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
-	memset(&rpc_createerr, 0, sizeof(rpc_createerr));
-	auth = mem_alloc(sizeof(*auth));
+	auth->ah_ops = &authgss_ops;
+	auth->ah_private = NULL;
+	auth->ah_error.re_status = RPC_SUCCESS;
 
 	/* XXX move to ctor */
-	gd = mem_alloc(sizeof(*gd));
-
 	mutex_init(&gd->lock, NULL);
-#ifdef DEBUG
-	__warnx(TIRPC_DEBUG_FLAG_AUTH, "authgss_ncreate: name is %p\n", name);
-#endif
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s() name is %p", __func__, name);
+
 	if (name != GSS_C_NO_NAME) {
 		maj_stat = gss_duplicate_name(&min_stat, name, &gd->name);
 		if (maj_stat != GSS_S_COMPLETE) {
-			rpc_createerr.cf_stat = RPC_SYSTEMERROR;
-			rpc_createerr.cf_error.re_errno = EINVAL;
-			mem_free(gd, sizeof(*gd));
-			mem_free(auth, sizeof(*auth));
-			return (NULL);
+			auth->ah_error.re_status = RPC_SYSTEMERROR;
+			auth->ah_error.re_errno = EINVAL;
+			return (auth);
 		}
 	} else
 		gd->name = name;
 
-#ifdef DEBUG
-	__warnx(TIRPC_DEBUG_FLAG_AUTH, "authgss_ncreate: gd->name is %p\n",
-		gd->name);
-#endif
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s() gd->name is %p",
+		__func__, gd->name);
+
 	gd->clnt = clnt;
 	gd->ctx = GSS_C_NO_CONTEXT;
 	gd->sec = *sec;
@@ -191,9 +188,6 @@ authgss_ncreate(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 	gd->gc.gc_proc = RPCSEC_GSS_INIT;
 	gd->gc.gc_svc = gd->sec.svc;
 
-	auth->ah_ops = &authgss_ops;
-	auth->ah_private = (caddr_t) gd;
-
 	if (!authgss_refresh(auth, NULL))
 		auth = NULL;
 	else
@@ -211,7 +205,7 @@ authgss_ncreate_default(CLIENT *clnt, char *service,
 	gss_buffer_desc sname;
 	gss_name_t name = GSS_C_NO_NAME;
 
-	log_debug("in authgss_ncreate_default()");
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	sname.value = service;
 	sname.length = strlen(service);
@@ -221,18 +215,17 @@ authgss_ncreate_default(CLIENT *clnt, char *service,
 			    (gss_OID) GSS_C_NT_HOSTBASED_SERVICE, &name);
 
 	if (maj_stat != GSS_S_COMPLETE) {
-		log_status("gss_import_name", maj_stat, min_stat);
-		rpc_createerr.cf_stat = RPC_AUTHERROR;
-		return (NULL);
+		gss_log_status("gss_import_name", maj_stat, min_stat);
+		auth = authnone_ncreate_dummy();
+		auth->ah_error.re_status = RPC_AUTHERROR;
+		return (auth);
 	}
 
 	auth = authgss_ncreate(clnt, name, sec);
 
 	if (name != GSS_C_NO_NAME) {
-#ifdef DEBUG
-		__warnx(TIRPC_DEBUG_FLAG_AUTH,
-			"authgss_ncreate_default: freeing name %p\n", name);
-#endif
+		__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS,
+			"%s() freeing name %p", __func__, name);
 		gss_release_name(&min_stat, &name);
 	}
 
@@ -244,7 +237,7 @@ authgss_get_private_data(AUTH *auth, struct authgss_private_data *pd)
 {
 	struct rpc_gss_data *gd;
 
-	log_debug("in authgss_get_private_data()");
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	if (!auth || !pd)
 		return (false);
@@ -264,23 +257,21 @@ authgss_get_private_data(AUTH *auth, struct authgss_private_data *pd)
 static void
 authgss_nextverf(AUTH *auth)
 {
-	log_debug("in authgss_nextverf()");
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 	/* no action necessary */
 }
 
 static bool
 authgss_marshal(AUTH *auth, XDR *xdrs)
 {
+	struct rpc_gss_data *gd = AUTH_PRIVATE(auth);
 	XDR tmpxdrs;
 	char tmp[MAX_AUTH_BYTES];
-	struct rpc_gss_data *gd;
 	gss_buffer_desc rpcbuf, checksum;
 	OM_uint32 maj_stat, min_stat;
 	bool xdr_stat;
 
-	log_debug("in authgss_marshal()");
-
-	gd = AUTH_PRIVATE(auth);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	if (gd->established) {
 		/* XXX */
@@ -296,39 +287,48 @@ authgss_marshal(AUTH *auth, XDR *xdrs)
 		return (false);
 	}
 	auth->ah_cred.oa_flavor = RPCSEC_GSS;
-	auth->ah_cred.oa_base = tmp;
 	auth->ah_cred.oa_length = XDR_GETPOS(&tmpxdrs);
+	if (auth->ah_cred.oa_length > MAX_AUTH_BYTES) {
+		XDR_DESTROY(&tmpxdrs);
+		return (false);
+	}
+	memcpy(auth->ah_cred.oa_body, tmp, auth->ah_cred.oa_length);
 
 	XDR_DESTROY(&tmpxdrs);
 
-	if (!inline_xdr_opaque_auth(xdrs, &auth->ah_cred))
+	if (!xdr_opaque_auth_encode(xdrs, &auth->ah_cred))
 		return (false);
 
 	if (gd->gc.gc_proc == RPCSEC_GSS_INIT
 	    || gd->gc.gc_proc == RPCSEC_GSS_CONTINUE_INIT) {
-		return (inline_xdr_opaque_auth(xdrs, &_null_auth));
+		return (xdr_opaque_auth_encode(xdrs, &_null_auth));
 	}
 	/* Checksum serialized RPC header, up to and including credential. */
 	rpcbuf.length = XDR_GETPOS(xdrs);
 	XDR_SETPOS(xdrs, 0);
-	rpcbuf.value = XDR_INLINE(xdrs, rpcbuf.length);
+	rpcbuf.value = xdr_inline_encode(xdrs, rpcbuf.length);
 
 	maj_stat =
 	    gss_get_mic(&min_stat, gd->ctx, gd->sec.qop, &rpcbuf, &checksum);
 
 	if (maj_stat != GSS_S_COMPLETE) {
-		log_status("gss_get_mic", maj_stat, min_stat);
+		gss_log_status("gss_get_mic", maj_stat, min_stat);
 		if (maj_stat == GSS_S_CONTEXT_EXPIRED) {
 			gd->established = false;
 			authgss_destroy_context(auth);
 		}
 		return (false);
 	}
+	if (checksum.length > MAX_AUTH_BYTES) {
+		gss_log_status("checksum.length", maj_stat, min_stat);
+		gss_release_buffer(&min_stat, &checksum);
+		return (false);
+	}
 	auth->ah_verf.oa_flavor = RPCSEC_GSS;
-	auth->ah_verf.oa_base = checksum.value;
 	auth->ah_verf.oa_length = checksum.length;
+	memcpy(auth->ah_verf.oa_body, checksum.value, checksum.length);
 
-	xdr_stat = xdr_opaque_auth(xdrs, &auth->ah_verf);
+	xdr_stat = xdr_opaque_auth_encode(xdrs, &auth->ah_verf);
 	gss_release_buffer(&min_stat, &checksum);
 
 	return (xdr_stat);
@@ -337,14 +337,12 @@ authgss_marshal(AUTH *auth, XDR *xdrs)
 static bool
 authgss_validate(AUTH *auth, struct opaque_auth *verf)
 {
-	struct rpc_gss_data *gd;
+	struct rpc_gss_data *gd = AUTH_PRIVATE(auth);
 	u_int num, qop_state;
 	gss_buffer_desc signbuf, checksum;
 	OM_uint32 maj_stat, min_stat;
 
-	log_debug("in authgss_validate()");
-
-	gd = AUTH_PRIVATE(auth);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	if (gd->established == false) {
 		/* would like to do this only on NULL rpc --
@@ -354,7 +352,7 @@ authgss_validate(AUTH *auth, struct opaque_auth *verf)
 		 * status is GSS_S_COMPLETE
 		 */
 		gd->gc_wire_verf.value = mem_alloc(verf->oa_length);
-		memcpy(gd->gc_wire_verf.value, verf->oa_base, verf->oa_length);
+		memcpy(gd->gc_wire_verf.value, verf->oa_body, verf->oa_length);
 		gd->gc_wire_verf.length = verf->oa_length;
 		return (true);
 	}
@@ -368,13 +366,13 @@ authgss_validate(AUTH *auth, struct opaque_auth *verf)
 	signbuf.value = &num;
 	signbuf.length = sizeof(num);
 
-	checksum.value = verf->oa_base;
+	checksum.value = verf->oa_body;
 	checksum.length = verf->oa_length;
 
 	maj_stat =
 	    gss_verify_mic(&min_stat, gd->ctx, &signbuf, &checksum, &qop_state);
 	if (maj_stat != GSS_S_COMPLETE || qop_state != gd->sec.qop) {
-		log_status("gss_verify_mic", maj_stat, min_stat);
+		gss_log_status("gss_verify_mic", maj_stat, min_stat);
 		if (maj_stat == GSS_S_CONTEXT_EXPIRED) {
 			gd->established = false;
 			authgss_destroy_context(auth);
@@ -394,15 +392,13 @@ authgss_validate(AUTH *auth, struct opaque_auth *verf)
 static bool
 authgss_refresh(AUTH *auth, void *arg)
 {
-	struct rpc_gss_data *gd;
+	struct rpc_gss_data *gd = AUTH_PRIVATE(auth);
 	struct rpc_gss_init_res gr;
 	gss_buffer_desc *recv_tokenp, send_token;
 	OM_uint32 maj_stat, min_stat, call_stat, ret_flags;
 	bool gd_locked = false;
 
-	log_debug("in authgss_refresh()");
-
-	gd = AUTH_PRIVATE(auth);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	/* Serialize context. */
 	mutex_lock(&gd->lock);
@@ -422,8 +418,9 @@ authgss_refresh(AUTH *auth, void *arg)
 #ifdef DEBUG
 		/* print the token we just received */
 		if (recv_tokenp != GSS_C_NO_BUFFER) {
-			log_debug("The token we just received (length %d):",
-				  recv_tokenp->length);
+			__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS,
+				"The token we just received (length %d):",
+				recv_tokenp->length);
 			gss_log_hexdump(recv_tokenp->value, recv_tokenp->length,
 					0);
 		}
@@ -448,25 +445,32 @@ authgss_refresh(AUTH *auth, void *arg)
 		}
 		if (maj_stat != GSS_S_COMPLETE
 		    && maj_stat != GSS_S_CONTINUE_NEEDED) {
-			log_status("gss_init_sec_context", maj_stat, min_stat);
+			gss_log_status("gss_init_sec_context",
+					maj_stat, min_stat);
 			break;
 		}
 		if (send_token.length != 0) {
+			struct clnt_req *cc = mem_alloc(sizeof(*cc));
+
 			memset(&gr, 0, sizeof(gr));
 
 #ifdef DEBUG
 			/* print the token we are about to send */
-			log_debug("The token being sent (length %d):",
-				  send_token.length);
+			__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS,
+				"The token being sent (length %d):",
+				send_token.length);
 			gss_log_hexdump(send_token.value, send_token.length, 0);
 #endif
 
-			call_stat =
-			    clnt_call(gd->clnt, auth, NULLPROC,
+			clnt_req_fill(cc, gd->clnt, auth, NULLPROC,
 				      (xdrproc_t) xdr_rpc_gss_init_args,
 				      &send_token,
-				      (xdrproc_t) xdr_rpc_gss_init_res,
-				      (caddr_t) &gr, AUTH_TIMEOUT);
+				      (xdrproc_t) xdr_rpc_gss_init_res, &gr);
+			call_stat = clnt_req_setup(cc, to);
+			if (call_stat == RPC_SUCCESS) {
+				call_stat = CLNT_CALL_WAIT(cc);
+			}
+			clnt_req_release(cc);
 
 			gss_release_buffer(&min_stat, &send_token);
 
@@ -509,8 +513,8 @@ authgss_refresh(AUTH *auth, void *arg)
 
 			if (maj_stat != GSS_S_COMPLETE
 			    || qop_state != gd->sec.qop) {
-				log_status("gss_verify_mic", maj_stat,
-					   min_stat);
+				gss_log_status("gss_verify_mic",
+						maj_stat, min_stat);
 				if (maj_stat == GSS_S_CONTEXT_EXPIRED) {
 					gd->established = false;
 					authgss_destroy_context(auth);
@@ -529,10 +533,7 @@ authgss_refresh(AUTH *auth, void *arg)
 		if (gr.gr_token.length != 0)
 			gss_release_buffer(&min_stat, &gr.gr_token);
 
-		authgss_destroy(auth);
-		auth = NULL;
-		rpc_createerr.cf_stat = RPC_AUTHERROR;
-
+		auth->ah_error.re_status = RPC_AUTHERROR;
 		authgss_refresh_return(false);
 	}
 	authgss_refresh_return(true);
@@ -543,7 +544,7 @@ authgss_service(AUTH *auth, int svc)
 {
 	struct rpc_gss_data *gd;
 
-	log_debug("in authgss_service()");
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	if (!auth)
 		return (false);
@@ -558,19 +559,24 @@ authgss_service(AUTH *auth, int svc)
 static void
 authgss_destroy_context(AUTH *auth)
 {
-	struct rpc_gss_data *gd;
+	struct rpc_gss_data *gd = AUTH_PRIVATE(auth);
 	OM_uint32 min_stat;
 
-	log_debug("in authgss_destroy_context()");
-
-	gd = AUTH_PRIVATE(auth);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	if (gd->gc.gc_ctx.length != 0) {
 		if (gd->established) {
+			struct clnt_req *cc = mem_alloc(sizeof(*cc));
+
 			gd->gc.gc_proc = RPCSEC_GSS_DESTROY;
-			clnt_call(gd->clnt, auth, NULLPROC,
-				  (xdrproc_t) xdr_void, NULL,
-				  (xdrproc_t) xdr_void, NULL, AUTH_TIMEOUT);
+
+			clnt_req_fill(cc, gd->clnt, auth, NULLPROC,
+				      (xdrproc_t) xdr_void, NULL,
+				      (xdrproc_t) xdr_void, NULL);
+			if (clnt_req_setup(cc, to) == RPC_SUCCESS) {
+				CLNT_CALL_WAIT(cc);
+			}
+			clnt_req_release(cc);
 		}
 		gss_release_buffer(&min_stat, &gd->gc.gc_ctx);
 		/* XXX ANDROS check size of context  - should be 8 */
@@ -592,57 +598,49 @@ authgss_destroy_context(AUTH *auth)
 static void
 authgss_destroy(AUTH *auth)
 {
-	struct rpc_gss_data *gd;
+	struct rpc_gss_data *gd = AUTH_PRIVATE(auth);
 	OM_uint32 min_stat;
 
-	log_debug("in authgss_destroy()");
-
-	gd = AUTH_PRIVATE(auth);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	authgss_destroy_context(auth);
 
 #ifdef DEBUG
-	__warnx(TIRPC_DEBUG_FLAG_AUTH, "authgss_destroy: freeing name %p\n",
-		gd->name);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s() freeing name %p",
+		__func__, gd->name);
 #endif
 	if (gd->name != GSS_C_NO_NAME)
 		gss_release_name(&min_stat, &gd->name);
 
 	mem_free(gd, sizeof(*gd));
-	mem_free(auth, sizeof(*auth));
 }
 
 bool
 authgss_wrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func, caddr_t xdr_ptr)
 {
-	struct rpc_gss_data *gd;
+	struct rpc_gss_data *gd = AUTH_PRIVATE(auth);
 
-	log_debug("in authgss_wrap()");
-
-	gd = AUTH_PRIVATE(auth);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	if (!gd->established || gd->sec.svc == RPCSEC_GSS_SVC_NONE)
 		return ((*xdr_func) (xdrs, xdr_ptr));
 
-	return (xdr_rpc_gss_data
+	return (xdr_rpc_gss_wrap
 		(xdrs, xdr_func, xdr_ptr, gd->ctx, gd->sec.qop, gd->sec.svc,
 		 gd->gc.gc_seq));
 }
 
 bool
-authgss_unwrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func,
-	       caddr_t xdr_ptr)
+authgss_unwrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func, caddr_t xdr_ptr)
 {
-	struct rpc_gss_data *gd;
+	struct rpc_gss_data *gd = AUTH_PRIVATE(auth);
 
-	log_debug("in authgss_unwrap()");
-
-	gd = AUTH_PRIVATE(auth);
+	__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS, "%s()", __func__);
 
 	if (!gd->established || gd->sec.svc == RPCSEC_GSS_SVC_NONE)
 		return ((*xdr_func) (xdrs, xdr_ptr));
 
-	return (xdr_rpc_gss_data
+	return (xdr_rpc_gss_unwrap
 		(xdrs, xdr_func, xdr_ptr, gd->ctx, gd->sec.qop, gd->sec.svc,
 		 gd->gc.gc_seq));
 }
