@@ -53,7 +53,6 @@
 #include <rpc/rpc.h>
 #ifdef PORTMAP
 #include <rpc/pmap_prot.h>
-#include <rpc/pmap_clnt.h>
 #include <rpc/pmap_rmt.h>
 #endif				/* PORTMAP */
 #include <rpc/nettype.h>
@@ -229,18 +228,14 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 		  int waittime,	/* maximum time to wait */
 		  const char *nettype /* transport type */)
 {
-	enum clnt_stat stat = RPC_SUCCESS;	/* Return status */
-	XDR xdr_stream;		/* XDR stream */
-	XDR *xdrs = &xdr_stream;
+	XDR xdrs[1];
 	struct rpc_msg msg;	/* RPC message */
 	struct timespec ts;
 	char *outbuf = NULL;	/* Broadcast msg buffer */
 	char *inbuf = NULL;	/* Reply buf */
-	int inlen;
-	u_int maxbufsize = 0;
 	AUTH *sys_auth = authunix_ncreate_default();
-	int i;
 	void *handle;
+	struct netconfig *nconf;
 	char uaddress[1024];	/* A self imposed limit */
 	char *uaddrp = uaddress;
 	int pmap_reply_flag;	/* reply recvd from PORTMAP */
@@ -260,11 +255,11 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 	struct r_rpcb_rmtcallargs barg;	/* Remote arguments */
 	struct r_rpcb_rmtcallres bres;	/* Remote results */
 	size_t outlen;
-	struct netconfig *nconf;
 	int msec;
 	int pollretval;
 	int fds_found;
-
+	int inlen;
+	int i;
 #ifdef PORTMAP
 	size_t outlen_pmap = 0;
 	u_long port;		/* Remote port number */
@@ -274,9 +269,14 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 	struct rmtcallres bres_pmap;	/* Remote results */
 	u_int udpbufsz = 0;
 #endif				/* PORTMAP */
+	u_int maxbufsize = 0;
+	enum clnt_stat stat = RPC_SUCCESS;	/* Return status */
 
-	if (sys_auth == NULL)
-		return (RPC_SYSTEMERROR);
+	stat = sys_auth->ah_error.re_status;
+	if (stat) {
+		AUTH_DESTROY(sys_auth);
+		return (stat);
+	}
 
 	/*
 	 * initialization: create a fd, a broadcast address, and send the
@@ -360,9 +360,9 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 	msg.rm_xid = __RPC_GETXID(&ts);
 	msg.rm_direction = CALL;
 	msg.rm_call.cb_rpcvers = RPC_MSG_VERSION;
-	msg.rm_call.cb_prog = RPCBPROG;
-	msg.rm_call.cb_vers = RPCBVERS;
-	msg.rm_call.cb_proc = RPCBPROC_CALLIT;
+	msg.cb_prog = RPCBPROG;
+	msg.cb_vers = RPCBVERS;
+	msg.cb_proc = RPCBPROC_CALLIT;
 	barg.prog = prog;
 	barg.vers = vers;
 	barg.proc = proc;
@@ -371,8 +371,8 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 	bres.addr = uaddrp;
 	bres.results.results_val = resultsp;
 	bres.xdr_res = xresults;
-	msg.rm_call.cb_cred = sys_auth->ah_cred;
-	msg.rm_call.cb_verf = sys_auth->ah_verf;
+	msg.cb_cred = sys_auth->ah_cred;
+	msg.cb_verf = sys_auth->ah_verf;
 	xdrmem_create(xdrs, outbuf, maxbufsize, XDR_ENCODE);
 	if ((!xdr_callmsg(xdrs, &msg))
 	    ||
@@ -393,9 +393,9 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 			goto done_broad;
 		}
 		msg.rm_xid++;	/* One way to distinguish */
-		msg.rm_call.cb_prog = PMAPPROG;
-		msg.rm_call.cb_vers = PMAPVERS;
-		msg.rm_call.cb_proc = PMAPPROC_CALLIT;
+		msg.cb_prog = PMAPPROG;
+		msg.cb_vers = PMAPVERS;
+		msg.cb_proc = PMAPPROC_CALLIT;
 		barg_pmap.prog = prog;
 		barg_pmap.vers = vers;
 		barg_pmap.proc = proc;
@@ -562,20 +562,20 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 			if (*((u_int32_t *) (void *)(inbuf)) ==
 			    *((u_int32_t *) (void *)(outbuf))) {
 				pmap_reply_flag = 0;
-				msg.acpted_rply.ar_verf = _null_auth;
-				msg.acpted_rply.ar_results.where =
+				msg.RPCM_ack.ar_verf = _null_auth;
+				msg.RPCM_ack.ar_results.where =
 				    (caddr_t) (void *)&bres;
-				msg.acpted_rply.ar_results.proc =
+				msg.RPCM_ack.ar_results.proc =
 				    (xdrproc_t) xdr_rpcb_rmtcallres;
 #ifdef PORTMAP
 			} else if (pmap_flag
 				   && *((u_int32_t *) (void *)(inbuf)) ==
 				   *((u_int32_t *) (void *)(outbuf_pmap))) {
 				pmap_reply_flag = 1;
-				msg.acpted_rply.ar_verf = _null_auth;
-				msg.acpted_rply.ar_results.where =
+				msg.RPCM_ack.ar_verf = _null_auth;
+				msg.RPCM_ack.ar_results.where =
 				    (caddr_t) (void *)&bres_pmap;
-				msg.acpted_rply.ar_results.proc =
+				msg.RPCM_ack.ar_results.proc =
 				    (xdrproc_t) xdr_rmtcallres;
 #endif				/* PORTMAP */
 			} else
@@ -583,7 +583,7 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 			xdrmem_create(xdrs, inbuf, (u_int) inlen, XDR_DECODE);
 			if (xdr_replymsg(xdrs, &msg)) {
 				if ((msg.rm_reply.rp_stat == MSG_ACCEPTED)
-				    && (msg.acpted_rply.ar_stat == SUCCESS)) {
+				    && (msg.RPCM_ack.ar_stat == SUCCESS)) {
 					struct netbuf *np;
 #ifdef PORTMAP
 					struct netbuf taddr;
@@ -626,7 +626,7 @@ rpc_broadcast_exp(rpcprog_t prog,	/* program number */
 			/* else some kind of deserialization problem ... */
 
 			xdrs->x_op = XDR_FREE;
-			msg.acpted_rply.ar_results.proc = (xdrproc_t) xdr_void;
+			msg.RPCM_ack.ar_results.proc = (xdrproc_t) xdr_void;
 			(void)xdr_replymsg(xdrs, &msg);
 			(void)(*xresults) (xdrs, resultsp);
 			XDR_DESTROY(xdrs);
