@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2009, Sun Microsystems, Inc.
+ * Copyright (c) 2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -59,6 +60,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <rpc/nettype.h>
+#include <assert.h>
 
 #include "rpc_com.h"
 
@@ -68,6 +70,12 @@ thr_keyfree(void *k)
 	mem_free(k, 0);	/* XXX */
 }
 
+static void
+tirpc_thread_name(const char *p)
+{
+	/* do nothing */
+}
+
 static void
 tirpc_free(void *p, size_t unused)
 {
@@ -77,39 +85,51 @@ tirpc_free(void *p, size_t unused)
 static void *
 tirpc_malloc(size_t size, const char *file, int line, const char *function)
 {
-	return malloc(size);
+	void *r = malloc(size);
+
+	assert(r != NULL);
+	return r;
 }
 
 static void *
 tirpc_aligned(size_t alignment, size_t size, const char *file, int line,
 	      const char *function)
 {
+	void *r;
+
 #if defined(_ISOC11_SOURCE)
-	return aligned_alloc(alignment, size);
+	r = aligned_alloc(alignment, size);
 #else
-	void *ptr;
-	(void) posix_memalign(&ptr, alignment, size);
-	return ptr;
+	(void) posix_memalign(&r, alignment, size);
 #endif
+	assert(r != NULL);
+	return r;
 }
 
 static void *
 tirpc_calloc(size_t count, size_t size, const char *file, int line,
 	     const char *function)
 {
-	return calloc(count, size);
+	void *r = calloc(count, size);
+
+	assert(r != NULL);
+	return r;
 }
 
 static void *
 tirpc_realloc(void *p, size_t size, const char *file, int line,
 	      const char *function)
 {
-	return realloc(p, size);
+	void *r = realloc(p, size);
+
+	assert(r != NULL);
+	return r;
 }
 
 tirpc_pkg_params __ntirpc_pkg_params = {
 	TIRPC_FLAG_NONE,
 	TIRPC_DEBUG_FLAG_NONE,
+	tirpc_thread_name,
 	warnx,
 	tirpc_free,
 	tirpc_malloc,
@@ -537,6 +557,8 @@ __rpc_endconf(void *vhandle)
 	mem_free(handle, sizeof(*handle));
 }
 
+static const struct timespec to = { 3, 0 };
+
 /*
  * Used to ping the NULL procedure for clnt handle.
  * Returns NULL if fails, else a non-NULL pointer.
@@ -544,15 +566,25 @@ __rpc_endconf(void *vhandle)
 void *
 rpc_nullproc(CLIENT *clnt)
 {
-	struct timeval TIMEOUT = { 25, 0 };
-	AUTH *auth;
+	struct clnt_req *cc = mem_alloc(sizeof(*cc));
+	enum clnt_stat stat;
 
-	auth = authnone_create();	/* idempotent */
-	if (clnt_call
-	    (clnt, auth, NULLPROC, (xdrproc_t) xdr_void, NULL,
-	     (xdrproc_t) xdr_void, NULL, TIMEOUT) != RPC_SUCCESS) {
+	clnt_req_fill(cc, clnt, authnone_ncreate(), NULLPROC,
+		      (xdrproc_t) xdr_void, NULL,
+		      (xdrproc_t) xdr_void, NULL);
+	stat = clnt_req_setup(cc, to);
+	if (stat == RPC_SUCCESS) {
+		stat = CLNT_CALL_WAIT(cc);
+	}
+	if (stat != RPC_SUCCESS) {
+		char *t = rpc_sperror(&cc->cc_error, __func__);
+
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
+		clnt_req_release(cc);
 		return (NULL);
 	}
+	clnt_req_release(cc);
 
 	return ((void *)clnt);
 }
