@@ -41,11 +41,7 @@
 #endif
 
 #include <stdio.h>	//printf
-#include <stdlib.h>	//malloc
-#include <string.h>	//memcpy
 #include <limits.h>	//INT_MAX
-#include <errno.h>	//ENOMEM
-#include <err.h>	//warnx
 #include <sys/socket.h> //sockaddr
 #include <sys/un.h>     //sockaddr_un
 #include <pthread.h>	//pthread_* (think it's included by another one)
@@ -67,7 +63,6 @@
 #define NSEC_IN_SEC (1000*1000*1000)
 
 #include "misc/portable.h"
-#include <infiniband/arch.h>
 #include <rdma/rdma_cma.h>
 #include <rpc/types.h>
 #include <rpc/xdr.h>
@@ -76,6 +71,7 @@
 
 #include "misc/abstract_atomic.h"
 #include "rpc_rdma.h"
+#include "svc_internal.h"
 
 #ifdef HAVE_VALGRIND_MEMCHECK_H
 #  include <valgrind/memcheck.h>
@@ -221,18 +217,11 @@ rpc_rdma_pd_by_verbs(RDMAXPRT *xprt)
 	}
 
 	pd = mem_zalloc(sizeof(*pd));
-	if (pd) {
-		rc = 0;
-		pd->context = xprt->cm_id->verbs;
-		pd->pd_used = 1;
+	rc = 0;
+	pd->context = xprt->cm_id->verbs;
+	pd->pd_used = 1;
 
-		LIST_INSERT_HEAD(&rpc_rdma_state.pdh, pd, pdl);
-	} else {
-		rc = ENOSPC;
-		__warnx(TIRPC_DEBUG_FLAG_ERROR,
-			"%s() couldn't malloc rdma pd",
-			__func__);
-	}
+	LIST_INSERT_HEAD(&rpc_rdma_state.pdh, pd, pdl);
 	mutex_unlock(&rpc_rdma_state.lock);
 	xprt->pd = pd;
 	return rc;
@@ -422,7 +411,7 @@ rpc_rdma_worker_callback(struct work_pool_entry *wpe)
 		}
 
 		/* wpe->arg referenced before work_pool_submit() */
-		SVC_RELEASE(&xprt->xprt, SVC_REF_FLAG_NONE);
+		SVC_RELEASE(&xprt->sm_dr.xprt, SVC_REF_FLAG_NONE);
 		return;
 	}
 
@@ -448,7 +437,7 @@ rpc_rdma_worker_callback(struct work_pool_entry *wpe)
 	}
 
 	/* wpe->arg referenced before work_pool_submit() */
-	SVC_RELEASE(&xprt->xprt, SVC_REF_FLAG_NONE);
+	SVC_RELEASE(&xprt->sm_dr.xprt, SVC_REF_FLAG_NONE);
 }
 
 /**
@@ -728,7 +717,7 @@ rpc_rdma_cq_event_handler(RDMAXPRT *xprt)
 						break;
 				}
 
-				SVC_REF(&xprt->xprt, SVC_REF_FLAG_NONE);
+				SVC_REF(&xprt->sm_dr.xprt, SVC_REF_FLAG_NONE);
 				work_pool_submit(&svc_work_pool, &cbc->wpe);
 
 				if (xprt->state != RDMAXS_CLOSING
@@ -766,7 +755,7 @@ rpc_rdma_cq_event_handler(RDMAXPRT *xprt)
 						ntohl(wc[i].imm_data));
 				}
 
-				SVC_REF(&xprt->xprt, SVC_REF_FLAG_NONE);
+				SVC_REF(&xprt->sm_dr.xprt, SVC_REF_FLAG_NONE);
 				work_pool_submit(&svc_work_pool, &cbc->wpe);
 				break;
 
@@ -808,7 +797,7 @@ rpc_rdma_cq_event_handler(RDMAXPRT *xprt)
 						__func__, len);
 				}
 
-				SVC_REF(&xprt->xprt, SVC_REF_FLAG_NONE);
+				SVC_REF(&xprt->sm_dr.xprt, SVC_REF_FLAG_NONE);
 				work_pool_submit(&svc_work_pool, &cbc->wpe);
 				break;
 
@@ -1064,8 +1053,8 @@ rpc_rdma_cm_event_handler(RDMAXPRT *ep_xprt, struct rdma_cm_event *event)
 		cond_broadcast(&xprt->cm_cond);
 		mutex_unlock(&xprt->cm_lock);
 
-		if (xprt->xa->disconnect_cb)
-			xprt->xa->disconnect_cb(&xprt->xprt);
+		if (__svc_params->disconnect_cb)
+			__svc_params->disconnect_cb(&xprt->sm_dr.xprt);
 		break;
 
 	case RDMA_CM_EVENT_DEVICE_REMOVAL:
@@ -1174,7 +1163,7 @@ rpc_rdma_cm_thread(void *nullarg)
 
 			if (cm_xprt->state == RDMAXS_CLOSED
 			 && cm_xprt->destroy_on_disconnect)
-				SVC_DESTROY(&cm_xprt->xprt);
+				SVC_DESTROY(&cm_xprt->sm_dr.xprt);
 		}
 	}
 
@@ -1289,7 +1278,7 @@ rpc_rdma_destroy_stuff(RDMAXPRT *xprt)
 void
 rpc_rdma_destroy(SVCXPRT *s_xprt)
 {
-	RDMAXPRT *xprt = (RDMAXPRT *)s_xprt;
+	RDMAXPRT *xprt = RDMA_DR(REC_XPRT(s_xprt));
 
 	/* inhibit repeated destroy */
 	xprt->destroy_on_disconnect = false;
@@ -1350,8 +1339,7 @@ rpc_rdma_destroy(SVCXPRT *s_xprt)
 	 */
 	cond_destroy(&xprt->cm_cond);
 	mutex_destroy(&xprt->cm_lock);
-	mutex_destroy(&xprt->xprt.xp_lock);
-	mutex_destroy(&xprt->xprt.xp_auth_lock);
+	mutex_destroy(&xprt->sm_dr.xprt.xp_lock);
 
 	mem_free(xprt, sizeof(*xprt));
 }
@@ -1364,7 +1352,7 @@ rpc_rdma_destroy(SVCXPRT *s_xprt)
  * @return xprt on success, NULL on failure
  */
 static RDMAXPRT *
-rpc_rdma_allocate(struct rpc_rdma_attr *xa)
+rpc_rdma_allocate(const struct rpc_rdma_attr *xa)
 {
 	RDMAXPRT *xprt;
 	int rc;
@@ -1377,16 +1365,10 @@ rpc_rdma_allocate(struct rpc_rdma_attr *xa)
 	}
 
 	xprt = mem_zalloc(sizeof(RDMAXPRT));
-	if (!xprt) {
-		__warnx(TIRPC_DEBUG_FLAG_ERROR,
-			"%s() Out of memory",
-			__func__);
-		return NULL;
-	}
 
-	xprt->xprt.xp_type = XPRT_RDMA;
-	xprt->xprt.xp_refs = 1;
-	xprt->xprt.xp_ops = &rpc_rdma_ops;
+	xprt->sm_dr.xprt.xp_type = XPRT_RDMA;
+	xprt->sm_dr.xprt.xp_refs = 1;
+	xprt->sm_dr.xprt.xp_ops = &rpc_rdma_ops;
 
 	xprt->xa = xa;
 	xprt->conn_type = RDMA_PS_TCP;
@@ -1394,15 +1376,7 @@ rpc_rdma_allocate(struct rpc_rdma_attr *xa)
 
 	/* initialize locking first, will be destroyed last (above).
 	 */
-	rc = mutex_init(&xprt->xprt.xp_auth_lock, NULL);
-	if (rc) {
-		__warnx(TIRPC_DEBUG_FLAG_ERROR,
-			"%s() mutex_init xp_auth_lock failed: %s (%d)",
-			__func__, strerror(rc), rc);
-		goto xp_auth_lock;
-	}
-
-	rc = mutex_init(&xprt->xprt.xp_lock, NULL);
+	rc = mutex_init(&xprt->sm_dr.xprt.xp_lock, NULL);
 	if (rc) {
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s() mutex_init xp_lock failed: %s (%d)",
@@ -1431,23 +1405,26 @@ rpc_rdma_allocate(struct rpc_rdma_attr *xa)
 cm_cond:
 	mutex_destroy(&xprt->cm_lock);
 cm_lock:
-	mutex_destroy(&xprt->xprt.xp_lock);
+	mutex_destroy(&xprt->sm_dr.xprt.xp_lock);
 xp_lock:
-	mutex_destroy(&xprt->xprt.xp_auth_lock);
-xp_auth_lock:
 	mem_free(xprt, sizeof(*xprt));
 	return NULL;
 }
 
 /**
- * rpc_rdma_create: initialize rdma transport structures
+ * rpc_rdma_ncreatef: initialize rdma transport structures
  *
- * @param[IN] xa	parameters
+ * @param[IN] xa		parameters
+ * @param[IN] sendsize;		max send size
+ * @param[IN] recvsize;		max recv size
+ * @param[IN] flags; 		unused
  *
  * @return xprt on success, NULL on failure
  */
 SVCXPRT *
-rpc_rdma_create(struct rpc_rdma_attr *xa)
+rpc_rdma_ncreatef(const struct rpc_rdma_attr *xa,
+		  const u_int sendsize, const u_int recvsize,
+		  const uint32_t flags)
 {
 	RDMAXPRT *xprt;
 	int rc;
@@ -1459,13 +1436,6 @@ rpc_rdma_create(struct rpc_rdma_attr *xa)
 		return NULL;
 	}
 
-	if (xa->worker_count > 256) {
-		__warnx(TIRPC_DEBUG_FLAG_ERROR,
-			"%s() worker_count (%u) much too large",
-			__func__, xa->worker_count);
-		return NULL;
-	}
-
 	xprt = rpc_rdma_allocate(xa);
 	if (!xprt) {
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
@@ -1506,9 +1476,27 @@ rpc_rdma_create(struct rpc_rdma_attr *xa)
 			__func__);
 		goto failure;
 	}
-	svc_work_pool.params.thrd_max += xa->worker_count;
 	pthread_mutex_unlock(&svc_work_pool.pqh.qmutex);
 
+	/* buffer sizes MUST be page sized */
+	xprt->sm_dr.pagesz = sysconf(_SC_PAGESIZE);
+	if (recvsize) {
+		/* round up */
+		xprt->sm_dr.recvsz = recvsize + (xprt->sm_dr.pagesz - 1);
+		xprt->sm_dr.recvsz &= ~(xprt->sm_dr.pagesz - 1);
+	} else {
+		/* default */
+		xprt->sm_dr.recvsz = xprt->sm_dr.pagesz;
+	}
+	if (sendsize) {
+		/* round up */
+		xprt->sm_dr.sendsz = sendsize + (xprt->sm_dr.pagesz - 1);
+		xprt->sm_dr.sendsz &= ~(xprt->sm_dr.pagesz - 1);
+	} else {
+		/* default */
+		xprt->sm_dr.recvsz = xprt->sm_dr.pagesz;
+	}
+
 	/* round up to the next power of two */
 	rpc_rdma_state.c_r.q_size = 2;
 	while (rpc_rdma_state.c_r.q_size < xa->backlog) {
@@ -1516,12 +1504,6 @@ rpc_rdma_create(struct rpc_rdma_attr *xa)
 	}
 	rpc_rdma_state.c_r.id_queue = mem_alloc(rpc_rdma_state.c_r.q_size
 						* sizeof(struct rdma_cm_id *));
-	if (!rpc_rdma_state.c_r.id_queue) {
-		__warnx(TIRPC_DEBUG_FLAG_ERROR,
-			"%s() Could not allocate connection_requests queue",
-			__func__);
-		goto failure;
-	}
 	sem_init(&rpc_rdma_state.c_r.u_sem, 0, rpc_rdma_state.c_r.q_size);
 
 	rc = rpc_rdma_bind_server(xprt);
@@ -1530,16 +1512,15 @@ rpc_rdma_create(struct rpc_rdma_attr *xa)
 			"%s() NFS/RDMA dispatcher could not bind engine",
 			__func__);
 		goto failure;
-	} else {
-		__warnx(TIRPC_DEBUG_FLAG_RPC_RDMA,
-			"%s() NFS/RDMA engine bound",
-			__func__);
 	}
+	__warnx(TIRPC_DEBUG_FLAG_RPC_RDMA,
+		"%s() NFS/RDMA engine bound",
+		__func__);
 
-	return (&xprt->xprt);
+	return (&xprt->sm_dr.xprt);
 
 failure:
-	rpc_rdma_destroy(&xprt->xprt);
+	rpc_rdma_destroy(&xprt->sm_dr.xprt);
 	return NULL;
 }
 
@@ -1685,12 +1666,7 @@ rpc_rdma_setup_cbq(struct poolq_head *ioqh, u_int depth, u_int sge)
 	 */
 	while (depth--) {
 		cbc = mem_zalloc(ioqh->qsize);
-		if (!cbc) {
-			__warnx(TIRPC_DEBUG_FLAG_ERROR,
-				"%s() couldn't mem_zalloc callback context",
-				__func__);
-			return ENOMEM;
-		}
+
 		xdr_ioq_setup(&cbc->workq);
 		xdr_ioq_setup(&cbc->holdq);
 
@@ -1788,7 +1764,7 @@ rpc_rdma_bind_server(RDMAXPRT *xprt)
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s() %p[%u] rpc_rdma_thread_create_epoll failed: %s (%d)",
 			__func__, xprt, xprt->state, strerror(rc), rc);
-		atomic_dec_int32_t(&rpc_rdma_state.run_count);
+		(void) atomic_dec_int32_t(&rpc_rdma_state.run_count);
 		return rc;
 	}
 
@@ -1899,7 +1875,7 @@ rpc_rdma_clone(RDMAXPRT *l_xprt, struct rdma_cm_id *cm_id)
 	return xprt;
 
 failure:
-	rpc_rdma_destroy(&xprt->xprt);
+	rpc_rdma_destroy(&xprt->sm_dr.xprt);
 	return (NULL);
 }
 
@@ -2080,7 +2056,8 @@ rpc_rdma_bind_client(RDMAXPRT *xprt)
 		}
 
 		rc = rdma_resolve_addr(xprt->cm_id, res->ai_src_addr,
-					res->ai_dst_addr, xprt->xa->timeout);
+					res->ai_dst_addr,
+					__svc_params->idle_timeout);
 		if (rc) {
 			rc = errno;
 			__warnx(TIRPC_DEBUG_FLAG_ERROR,
@@ -2105,7 +2082,8 @@ rpc_rdma_bind_client(RDMAXPRT *xprt)
 			break;
 		}
 
-		rc = rdma_resolve_route(xprt->cm_id, xprt->xa->timeout);
+		rc = rdma_resolve_route(xprt->cm_id,
+					__svc_params->idle_timeout);
 		if (rc) {
 			xprt->state = RDMAXS_ERROR;
 			__warnx(TIRPC_DEBUG_FLAG_ERROR,
@@ -2230,12 +2208,46 @@ rpc_rdma_connect(RDMAXPRT *xprt)
 				rpc_rdma_state.cm_epollfd);
 }
 
+static void
+rpc_rdma_destroy_it(SVCXPRT *xprt, u_int flags, const char *tag, const int line)
+{
+	if (xprt->xp_ops->xp_free_user_data) {
+		/* call free hook */
+		xprt->xp_ops->xp_free_user_data(xprt);
+	}
+	rpc_rdma_destroy(xprt);
+}
+
+extern mutex_t ops_lock;
+
+static bool
+/*ARGSUSED*/
+rpc_rdma_control(SVCXPRT *xprt, const u_int rq, void *in)
+{
+	switch (rq) {
+	case SVCGET_XP_FREE_USER_DATA:
+	    mutex_lock(&ops_lock);
+	    *(svc_xprt_fun_t *)in = xprt->xp_ops->xp_free_user_data;
+	    mutex_unlock(&ops_lock);
+	    break;
+	case SVCSET_XP_FREE_USER_DATA:
+	    mutex_lock(&ops_lock);
+	    xprt->xp_ops->xp_free_user_data = *(svc_xprt_fun_t)in;
+	    mutex_unlock(&ops_lock);
+	    break;
+	default:
+	    return (FALSE);
+	}
+	return (TRUE);
+}
+
 static struct xp_ops rpc_rdma_ops = {
-	/* XXX wow */
-	.xp_getargs = (bool(*)(SVCXPRT *, struct svc_req *, xdrproc_t,
-				void *, void *))abort,
-	.xp_reply = (bool(*)(SVCXPRT *, struct svc_req *,
-				struct rpc_msg *))abort,
-	.xp_freeargs = (bool(*)(SVCXPRT *, struct svc_req *, xdrproc_t,
-				void *))abort,
+	.xp_recv = svc_rdma_rendezvous,
+	.xp_stat = svc_rendezvous_stat,
+	.xp_decode = (svc_req_fun_t)abort,
+	.xp_reply = (svc_req_fun_t)abort,
+	.xp_checksum = NULL,		/* not used */
+	.xp_destroy = rpc_rdma_destroy_it,
+	.xp_control = rpc_rdma_control,
+	.xp_free_user_data = NULL,	/* no default */
 };
