@@ -43,6 +43,8 @@
 #include <rpc/svc.h>
 #include <rpc/xdr_ioq.h>
 
+#include "rpc_dplx_internal.h"
+
 typedef union sockaddr_union {
 	struct sockaddr sa;
 	struct sockaddr_in sa_in;
@@ -76,6 +78,7 @@ struct rpc_rdma_cbc {
 	struct xdr_ioq holdq;
 
 	struct xdr_ioq_uv *call_uv;
+	void *call_head;
 	void *read_chunk;	/* current in indexed list of arrays */
 	void *write_chunk;	/* current in list of arrays */
 	void *reply_chunk;	/* current in array */
@@ -117,9 +120,9 @@ struct rpc_rdma_pd {
  * RDMA transport instance
  */
 struct rpc_rdma_xprt {
-	struct rpc_svcxprt xprt;	/**< Transport Independent handle */
+	struct rpc_dplx_rec sm_dr;
 
-	struct rpc_rdma_attr *xa;	/**< (shared) configured attributes */
+	const struct rpc_rdma_attr *xa;	/**< (shared) configured attributes */
 
 	struct rdma_event_channel *event_channel;
 	struct rdma_cm_id *cm_id;	/**< RDMA CM ID */
@@ -137,9 +140,6 @@ struct rpc_rdma_xprt {
 	u_int8_t *buffer_aligned;
 	size_t buffer_total;
 
-	/* waiting contexts posted */
-	struct xdr_ioq waitq;
-
 	struct xdr_ioq_uv_head inbufs;	/* recvsize */
 	struct xdr_ioq_uv_head outbufs;	/* sendsz */
 
@@ -151,9 +151,6 @@ struct rpc_rdma_xprt {
 	struct msk_stats stats;
 	int stats_sock;
 
-	u_int recvsize;
-	u_int sendsize;
-
 	u_int conn_type;		/**< RDMA Port space (RDMA_PS_TCP) */
 	int server;			/**< connection backlog on server,
 					 * 0 (RDMAX_CLIENT):
@@ -177,6 +174,7 @@ struct rpc_rdma_xprt {
 	/* FIXME why configurable??? */
 	bool destroy_on_disconnect;	/**< should perform cleanup */
 };
+#define RDMA_DR(p) (opr_containerof((p), struct rpc_rdma_xprt, sm_dr))
 
 typedef struct rec_rdma_strm {
 	RDMAXPRT *xprt;
@@ -239,14 +237,15 @@ int rpc_rdma_accept_finalize(RDMAXPRT *xprt);
 RDMAXPRT *rpc_rdma_accept_wait(RDMAXPRT *l_xprt, int msleep);
 void rpc_rdma_destroy(SVCXPRT *s_xprt);
 
+enum xprt_stat svc_rdma_rendezvous(SVCXPRT *l_xprt);
+
 /* client */
 int rpc_rdma_connect(RDMAXPRT *xprt);
 int rpc_rdma_connect_finalize(RDMAXPRT *xprt);
 
 /* XDR functions */
-int xdr_rdma_create(XDR *, RDMAXPRT *, const u_int sendsize,
-			const u_int recvsize, const u_int flags);
-void xdr_rdma_destroy(XDR *xdrs);
+int xdr_rdma_create(XDR *, RDMAXPRT *);
+void xdr_rdma_destroy(XDR *);
 
 bool xdr_rdma_clnt_call(XDR *, u_int32_t);
 bool xdr_rdma_clnt_reply(XDR *, u_int32_t);
