@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2010, Oracle America, Inc.
+ * Copyright (c) 2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -57,9 +58,8 @@
 
 #include "rpc_com.h"
 
-static struct timeval tottimeout = { 60, 0 };
-static const struct timeval rmttimeout = { 3, 0 };
-static struct timeval rpcbrmttime = { 15, 0 };
+/* retry timeout default to the moon and back */
+static struct timespec to = { 3, 0 };
 
 extern bool xdr_wrapstring(XDR *, char **);
 
@@ -89,7 +89,7 @@ static struct address_cache *check_cache(const char *, const char *);
 static void delete_cache(struct netbuf *);
 static void add_cache(const char *, const char *, struct netbuf *, char *);
 static CLIENT *getclnthandle(const char *, const struct netconfig *, char **);
-static CLIENT *local_rpcb(void);
+static CLIENT *local_rpcb(const char *);
 #ifdef NOTUSED
 static struct netbuf *got_entry(rpcb_entry_list_ptr, const struct netconfig *);
 #endif
@@ -104,10 +104,10 @@ bool __rpc_control(int request, void *info)
 {
 	switch (request) {
 	case CLCR_GET_RPCB_TIMEOUT:
-		*(struct timeval *)info = tottimeout;
+		*(struct timespec *)info = to;
 		break;
 	case CLCR_SET_RPCB_TIMEOUT:
-		tottimeout = *(struct timeval *)info;
+		to = *(struct timespec *)info;
 		break;
 	case CLCR_SET_LOWVERS:
 		__rpc_lowvers = *(int *)info;
@@ -168,15 +168,15 @@ delete_cache(struct netbuf *addr)
 		if (!memcmp(cptr->ac_taddr->buf, addr->buf, addr->len)) {
 			mem_free(cptr->ac_host, 0);	/* XXX */
 			mem_free(cptr->ac_netid, 0);
-			mem_free(cptr->ac_taddr->buf, 0);
-			mem_free(cptr->ac_taddr, 0);
+			mem_free(cptr->ac_taddr->buf, cptr->ac_taddr->len);
+			mem_free(cptr->ac_taddr, sizeof(struct netbuf));
 			if (cptr->ac_uaddr)
 				mem_free(cptr->ac_uaddr, 0);
 			if (prevptr)
 				prevptr->ac_next = cptr->ac_next;
 			else
 				front = cptr->ac_next;
-			mem_free(cptr, 0);
+			mem_free(cptr, sizeof(struct address_cache));
 			cachesize--;
 			break;
 		}
@@ -190,21 +190,18 @@ add_cache(const char *host, const char *netid, struct netbuf *taddr,
 {
 	struct address_cache *ad_cache, *cptr, *prevptr;
 
-	ad_cache = (struct address_cache *)
-	    mem_zalloc(sizeof(struct address_cache));
-	if (!ad_cache)
+	if (!host) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: missing host", __func__);
 		return;
-	ad_cache->ac_host = (host) ? rpc_strdup(host) : NULL;
-	ad_cache->ac_netid = rpc_strdup(netid);
-	ad_cache->ac_uaddr = uaddr ? rpc_strdup(uaddr) : NULL;
+	}
+	ad_cache = (struct address_cache *)mem_zalloc(sizeof(*ad_cache));
+
+	ad_cache->ac_host = mem_strdup(host);
+	ad_cache->ac_netid = mem_strdup(netid);
+	ad_cache->ac_uaddr = uaddr ? mem_strdup(uaddr) : NULL;
 	ad_cache->ac_taddr = (struct netbuf *)mem_zalloc(sizeof(struct netbuf));
-	if (!ad_cache->ac_host || !ad_cache->ac_netid || !ad_cache->ac_taddr
-	    || (uaddr && !ad_cache->ac_uaddr))
-		goto out_free;
 	ad_cache->ac_taddr->len = ad_cache->ac_taddr->maxlen = taddr->len;
 	ad_cache->ac_taddr->buf = (char *)mem_zalloc(taddr->len);
-	if (ad_cache->ac_taddr->buf == NULL)
-		goto out_free;
 	memcpy(ad_cache->ac_taddr->buf, taddr->buf, taddr->len);
 #ifdef ND_DEBUG
 	fprintf(stderr, "Added to cache: %s : %s\n", host, netid);
@@ -231,8 +228,8 @@ add_cache(const char *host, const char *netid, struct netbuf *taddr,
 #endif
 		mem_free(cptr->ac_host, 0);	/* XXX */
 		mem_free(cptr->ac_netid, 0);
-		mem_free(cptr->ac_taddr->buf, 0);
-		mem_free(cptr->ac_taddr, 0);
+		mem_free(cptr->ac_taddr->buf, cptr->ac_taddr->len);
+		mem_free(cptr->ac_taddr, sizeof(struct netbuf));
 		if (cptr->ac_uaddr)
 			mem_free(cptr->ac_uaddr, 0);
 
@@ -244,17 +241,10 @@ add_cache(const char *host, const char *netid, struct netbuf *taddr,
 			front = ad_cache;
 			ad_cache->ac_next = NULL;
 		}
-		mem_free(cptr, 0);
+		mem_free(cptr, sizeof(struct address_cache));
 	}
 	rwlock_unlock(&rpcbaddr_cache_lock);
 	return;
-
- out_free:
-	mem_free(ad_cache->ac_host, 0);	/* XXX */
-	mem_free(ad_cache->ac_netid, 0);
-	mem_free(ad_cache->ac_uaddr, 0);
-	mem_free(ad_cache->ac_taddr, 0);
-	mem_free(ad_cache, 0);
 }
 
 /*
@@ -262,7 +252,7 @@ add_cache(const char *host, const char *netid, struct netbuf *taddr,
  * rpcbind. If targaddr is non-NULL, the "universal address" of the
  * host will be stored in *targaddr; the caller is responsible for
  * freeing this string.
- * On error, returns NULL and free's everything.
+ * On error, a CLIENT with cl_error.re_status set
  */
 static CLIENT *getclnthandle(const char *host, const struct netconfig *nconf,
 			     char **targaddr)
@@ -274,6 +264,7 @@ static CLIENT *getclnthandle(const char *host, const struct netconfig *nconf,
 	struct addrinfo hints, *res, *tres;
 	struct address_cache *ad_cache;
 	char *tmpaddr;
+	char *t;
 
 	memset(&addr_to_delete, '\0', sizeof(addr_to_delete));
 
@@ -294,18 +285,21 @@ static CLIENT *getclnthandle(const char *host, const struct netconfig *nconf,
 		    clnt_tli_ncreate(RPC_ANYFD, nconf, addr,
 				     (rpcprog_t) RPCBPROG,
 				     (rpcvers_t) RPCBVERS4, 0, 0);
-		if (client != NULL) {
+		if (CLNT_SUCCESS(client)) {
 			if (targaddr)
-				*targaddr = rpc_strdup(ad_cache->ac_uaddr);
+				*targaddr = mem_strdup(ad_cache->ac_uaddr);
 			rwlock_unlock(&rpcbaddr_cache_lock);
 			return (client);
 		}
+
+		t = rpc_sperror(&client->cl_error, __func__);
+		__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
+		goto out_err;
+
 		addr_to_delete.len = addr->len;
 		addr_to_delete.buf = (char *)mem_zalloc(addr->len);
-		if (addr_to_delete.buf == NULL)
-			addr_to_delete.len = 0;
-		else
-			memcpy(addr_to_delete.buf, addr->buf, addr->len);
+		memcpy(addr_to_delete.buf, addr->buf, addr->len);
 	}
 	rwlock_unlock(&rpcbaddr_cache_lock);
 	if (addr_to_delete.len != 0) {
@@ -316,11 +310,14 @@ static CLIENT *getclnthandle(const char *host, const struct netconfig *nconf,
 		rwlock_wrlock(&rpcbaddr_cache_lock);
 		delete_cache(&addr_to_delete);
 		rwlock_unlock(&rpcbaddr_cache_lock);
-		mem_free(addr_to_delete.buf, 0);	/* XXX */
+		mem_free(addr_to_delete.buf, addr_to_delete.len);
 	}
 	if (!__rpc_nconf2sockinfo(nconf, &si)) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
 		assert(client == NULL);
+		__warnx(TIRPC_DEBUG_FLAG_WARN, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNPROTO));
+		client = clnt_raw_ncreate(1, 1);
+		client->cl_error.re_status = RPC_UNKNOWNPROTO;
 		goto out_err;
 	}
 
@@ -329,20 +326,14 @@ static CLIENT *getclnthandle(const char *host, const struct netconfig *nconf,
 	hints.ai_socktype = si.si_socktype;
 	hints.ai_protocol = si.si_proto;
 
-#ifdef CLNT_DEBUG
-	__warnx("%s: trying netid %s family %d proto %d socktype %d\n",
+	__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB,
+		"%s: trying netid %s family %d proto %d socktype %d",
 		__func__, nconf->nc_netid, si.si_af, si.si_proto,
 		si.si_socktype);
-#endif
 
 	if (strcmp(nconf->nc_protofmly, NC_LOOPBACK) == 0) {
-		client = local_rpcb();
-		if (!client) {
-#ifdef ND_DEBUG
-			clnt_pcreateerror("rpcbind clnt interface");
-#endif
-			goto out_err;
-		} else {
+		client = local_rpcb(__func__);
+		if (CLNT_SUCCESS(client)) {
 			struct sockaddr_un sun;
 
 			if (targaddr) {
@@ -352,10 +343,14 @@ static CLIENT *getclnthandle(const char *host, const struct netconfig *nconf,
 			}
 			return (client);
 		}
+		goto out_err;
 	} else {
 		if (getaddrinfo(host, "sunrpc", &hints, &res) != 0) {
-			rpc_createerr.cf_stat = RPC_UNKNOWNHOST;
 			assert(client == NULL);
+			__warnx(TIRPC_DEBUG_FLAG_WARN, "%s: %s",
+				__func__, clnt_sperrno(RPC_UNKNOWNHOST));
+			client = clnt_raw_ncreate(1, 1);
+			client->cl_error.re_status = RPC_UNKNOWNHOST;
 			goto out_err;
 		}
 	}
@@ -391,23 +386,22 @@ static CLIENT *getclnthandle(const char *host, const struct netconfig *nconf,
 		    clnt_tli_ncreate(RPC_ANYFD, nconf, &taddr,
 				     (rpcprog_t) RPCBPROG,
 				     (rpcvers_t) RPCBVERS4, 0, 0);
-#ifdef ND_DEBUG
-		if (!client)
-			clnt_pcreateerror("rpcbind clnt interface");
-#endif
-
-		if (client) {
+		if (CLNT_SUCCESS(client)) {
 			tmpaddr = targaddr ? taddr2uaddr(nconf, &taddr) : NULL;
 			add_cache(host, nconf->nc_netid, &taddr, tmpaddr);
 			if (targaddr)
 				*targaddr = tmpaddr;
 			break;
 		}
+
+		t = rpc_sperror(&client->cl_error, __func__);
+		__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
 	}
 	if (res)
 		freeaddrinfo(res);
  out_err:
-	if (!client && targaddr)
+	if (CLNT_FAILURE(client) && targaddr)
 		mem_free(*targaddr, 0);
 	return (client);
 }
@@ -421,26 +415,29 @@ rpcb_set(rpcprog_t program, rpcvers_t version, const struct netconfig *nconf,
 	 /* Network structure of transport */
 	 const struct netbuf *address /* Services netconfig address */)
 {
+	CLIENT *client;
+	struct clnt_req *cc;
 	RPCB parms;
 	char uidbuf[32];
+	enum clnt_stat stat;
 	bool_t rslt = false;	/* yes, bool_t */
-	CLIENT *client;
-	AUTH *auth;
 
 	/* parameter checking */
 	if (nconf == NULL) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNPROTO));
 		return (false);
 	}
 	if (address == NULL) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNADDR));
 		return (false);
 	}
-	client = local_rpcb();
-	if (!client)
+	client = local_rpcb(__func__);
+	if (CLNT_FAILURE(client)) {
+		CLNT_DESTROY(client);
 		return (false);
-
-	auth = authnone_ncreate();	/* idempotent */
+	}
 
 	/* convert to universal */
 	/*LINTED const castaway */
@@ -448,7 +445,8 @@ rpcb_set(rpcprog_t program, rpcvers_t version, const struct netconfig *nconf,
 	    taddr2uaddr((struct netconfig *)nconf, (struct netbuf *)address);
 	if (!parms.r_addr) {
 		CLNT_DESTROY(client);
-		rpc_createerr.cf_stat = RPC_N2AXLATEFAILURE;
+		__warnx(TIRPC_DEBUG_FLAG_WARN, "%s: %s",
+			__func__, clnt_sperrno(RPC_N2AXLATEFAILURE));
 		return (false);	/* no universal address */
 	}
 	parms.r_prog = program;
@@ -462,10 +460,22 @@ rpcb_set(rpcprog_t program, rpcvers_t version, const struct netconfig *nconf,
 	(void)snprintf(uidbuf, sizeof(uidbuf), "%d", geteuid());
 	parms.r_owner = uidbuf;
 
-	CLNT_CALL(client, auth, RPCBPROC_SET, (xdrproc_t) xdr_rpcb,
-		  (char *)&parms, (xdrproc_t) xdr_bool, (char *)&rslt,
-		  tottimeout);
+	cc = mem_alloc(sizeof(*cc));
+	clnt_req_fill(cc, client, authnone_ncreate(), RPCBPROC_SET,
+		      (xdrproc_t) xdr_rpcb, &parms,
+		      (xdrproc_t) xdr_bool, &rslt);
+	stat = clnt_req_setup(cc, to);
+	if (stat == RPC_SUCCESS) {
+		stat = CLNT_CALL_WAIT(cc);
+	}
+	if (stat != RPC_SUCCESS) {
+		char *t = rpc_sperror(&cc->cc_error, __func__);
 
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
+	}
+
+	clnt_req_release(cc);
 	CLNT_DESTROY(client);
 	mem_free(parms.r_addr, 0);
 	return (rslt);
@@ -481,16 +491,19 @@ bool
 rpcb_unset(rpcprog_t program, rpcvers_t version,
 	   const struct netconfig *nconf)
 {
+	CLIENT *client;
+	struct clnt_req *cc;
 	RPCB parms;
 	char uidbuf[32];
+	enum clnt_stat stat;
 	bool_t rslt = false;	/* yes, bool_t */
-	CLIENT *client;
-	AUTH *auth;
 
-	client = local_rpcb();
-	if (!client)
+	client = local_rpcb(__func__);
+	if (CLNT_FAILURE(client)) {
+		CLNT_DESTROY(client);
 		return (false);
-	auth = authnone_ncreate();	/* idempotent */
+	}
+
 	parms.r_prog = program;
 	parms.r_vers = version;
 	if (nconf)
@@ -504,10 +517,22 @@ rpcb_unset(rpcprog_t program, rpcvers_t version,
 	(void)snprintf(uidbuf, sizeof(uidbuf), "%d", geteuid());
 	parms.r_owner = uidbuf;
 
-	CLNT_CALL(client, auth, RPCBPROC_UNSET, (xdrproc_t) xdr_rpcb,
-		  (char *)(void *)&parms, (xdrproc_t) xdr_bool,
-		  (char *)(void *)&rslt, tottimeout);
+	cc = mem_alloc(sizeof(*cc));
+	clnt_req_fill(cc, client, authnone_ncreate(), RPCBPROC_UNSET,
+		      (xdrproc_t) xdr_rpcb, &parms,
+		      (xdrproc_t) xdr_bool, &rslt);
+	stat = clnt_req_setup(cc, to);
+	if (stat == RPC_SUCCESS) {
+		stat = CLNT_CALL_WAIT(cc);
+	}
+	if (stat != RPC_SUCCESS) {
+		char *t = rpc_sperror(&cc->cc_error, __func__);
 
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
+	}
+
+	clnt_req_release(cc);
 	CLNT_DESTROY(client);
 	return (rslt);
 }
@@ -531,13 +556,10 @@ got_entry(rpcb_entry_list_ptr relp,
 		    && (nconf->nc_semantics == rmap->r_nc_semantics)
 		    && (rmap->r_maddr != NULL) && (rmap->r_maddr[0] != 0)) {
 			na = uaddr2taddr(nconf, rmap->r_maddr);
-#ifdef ND_DEBUG
-			fprintf(stderr, "\tRemote address is [%s].\n",
-				rmap->r_maddr);
 			if (!na)
-				fprintf(stderr,
-					"\tCouldn't resolve remote address!\n");
-#endif
+				__warnx(TIRPC_DEBUG_FLAG_WARN, "%s: %s",
+					__func__,
+					clnt_sperrno(RPC_N2AXLATEFAILURE));
 			break;
 		}
 	}
@@ -611,23 +633,26 @@ __rpcb_findaddr_timed(rpcprog_t program, rpcvers_t version,
 		      const char *host, CLIENT **clpp,
 		      struct timeval *tp)
 {
-#ifdef NOTUSED
-	static bool check_rpcbind = true;
-#endif
-	RPCB parms;
-	enum clnt_stat clnt_st;
 	char *ua = NULL;
-	rpcvers_t vers;
 	struct netbuf *address = NULL;
-	rpcvers_t start_vers = RPCBVERS4;
-	struct netbuf servaddr;
 	CLIENT *client = NULL;
+	struct clnt_req *cc;
+	char *t;
 	AUTH *auth;
+	RPCB parms;
+	struct netbuf servaddr;
+	struct timespec tv;
+	rpcvers_t vers;
+	rpcvers_t start_vers = RPCBVERS4;
+	enum clnt_stat clnt_st;
 
 	/* parameter checking */
 	if (nconf == NULL) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
-		return (NULL);
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNPROTO));
+		client = clnt_raw_ncreate(program, version);
+		client->cl_error.re_status = RPC_UNKNOWNPROTO;
+		goto done;
 	}
 
 	parms.r_addr = NULL;
@@ -639,15 +664,19 @@ __rpcb_findaddr_timed(rpcprog_t program, rpcvers_t version,
 	 * Use default total timeout if no timeout is specified.
 	 */
 	if (tp == NULL)
-		tp = &tottimeout;
+		tv = to;
+	else {
+		tv.tv_sec = tp->tv_sec;
+		tv.tv_nsec = tp->tv_usec * 1000;
+	}
 
 #ifdef PORTMAP
 	/* Try version 2 for TCP or UDP */
 	if (strcmp(nconf->nc_protofmly, NC_INET) == 0) {
-		u_short port = 0;
 		struct netbuf remote;
-		rpcvers_t pmapvers = 2;
 		struct pmap pmapparms;
+		rpcvers_t pmapvers = 2;
+		u_short port = 0;
 
 		/*
 		 * Try UDP only - there are some portmappers out
@@ -658,8 +687,9 @@ __rpcb_findaddr_timed(rpcprog_t program, rpcvers_t version,
 
 			newnconf = getnetconfigent("udp");
 			if (!newnconf) {
-				rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
-				return (NULL);
+				client = clnt_raw_ncreate(program, version);
+				client->cl_error.re_status = RPC_UNKNOWNPROTO;
+				goto error;
 			}
 			client = getclnthandle(host, newnconf, &parms.r_addr);
 			freenetconfigent(newnconf);
@@ -667,13 +697,9 @@ __rpcb_findaddr_timed(rpcprog_t program, rpcvers_t version,
 			client = getclnthandle(host, nconf, &parms.r_addr);
 		else
 			goto try_rpcbind;
-		if (client == NULL)
-			return (NULL);
+		if (CLNT_FAILURE(client))
+			goto error;
 
-		/*
-		 * Set version and retry timeout.
-		 */
-		CLNT_CONTROL(client, CLSET_RETRY_TIMEOUT, (char *)&rpcbrmttime);
 		CLNT_CONTROL(client, CLSET_VERS, (char *)&pmapvers);
 
 		pmapparms.pm_prog = program;
@@ -681,39 +707,38 @@ __rpcb_findaddr_timed(rpcprog_t program, rpcvers_t version,
 		pmapparms.pm_prot =
 		    strcmp(nconf->nc_proto, NC_TCP) ? IPPROTO_UDP : IPPROTO_TCP;
 		pmapparms.pm_port = 0;	/* not needed */
-		clnt_st =
-		    CLNT_CALL(client, auth, PMAPPROC_GETPORT,
-			      (xdrproc_t) xdr_pmap,
-			      (caddr_t) (void *)&pmapparms,
-			      (xdrproc_t) xdr_u_short, (caddr_t) (void *)&port,
-			      *tp);
+
+		cc = mem_alloc(sizeof(*cc));
+		clnt_req_fill(cc, client, auth, PMAPPROC_GETPORT,
+			      (xdrproc_t) xdr_pmap, &pmapparms,
+			      (xdrproc_t) xdr_u_short, &port);
+		clnt_st = clnt_req_setup(cc, tv);
+		if (clnt_st == RPC_SUCCESS) {
+			clnt_st = CLNT_CALL_WAIT(cc);
+		}
 		if (clnt_st != RPC_SUCCESS) {
 			if ((clnt_st == RPC_PROGVERSMISMATCH)
-			    || (clnt_st == RPC_PROGUNAVAIL))
+			    || (clnt_st == RPC_PROGUNAVAIL)) {
+				clnt_req_release(cc);
 				goto try_rpcbind;
-			rpc_createerr.cf_stat = RPC_PMAPFAILURE;
-			clnt_geterr(client, &rpc_createerr.cf_error);
+			}
+			client->cl_error = cc->cc_error;
+			clnt_req_release(cc);
 			goto error;
-		} else if (port == 0) {
+		}
+		clnt_req_release(cc);
+
+		if (port == 0) {
 			address = NULL;
-			rpc_createerr.cf_stat = RPC_PROGNOTREGISTERED;
+			client->cl_error.re_status = RPC_PROGNOTREGISTERED;
 			goto error;
 		}
 		port = htons(port);
 		CLNT_CONTROL(client, CLGET_SVC_ADDR, (char *)&remote);
 
-		address = (struct netbuf *)  mem_zalloc(sizeof(struct netbuf));
-		if (address)
-			address->buf = (char *) mem_zalloc(remote.len);
-		if ((!address) || (!address->buf)) {
-			rpc_createerr.cf_stat = RPC_SYSTEMERROR;
-			clnt_geterr(client, &rpc_createerr.cf_error);
-			if (address) {
-				mem_free(address, 0);
-				address = NULL;
-			}
-			goto error;
-		}
+		address = (struct netbuf *)mem_zalloc(sizeof(struct netbuf));
+		address->buf = (char *)mem_alloc(remote.len);
+
 		memcpy(address->buf, remote.buf, remote.len);
 		memcpy(&((char *)address->buf)[sizeof(short)],
 		       (char *)(void *)&port, sizeof(short));
@@ -747,7 +772,7 @@ __rpcb_findaddr_timed(rpcprog_t program, rpcvers_t version,
 
 	if (client == NULL) {
 		client = getclnthandle(host, nconf, &parms.r_addr);
-		if (client == NULL)
+		if (CLNT_FAILURE(client))
 			goto error;
 	}
 	if (parms.r_addr == NULL) {
@@ -755,73 +780,71 @@ __rpcb_findaddr_timed(rpcprog_t program, rpcvers_t version,
 		parms.r_addr = (char *)&nullstring[0];
 	}
 
+	cc = mem_alloc(sizeof(*cc));
+	clnt_req_fill(cc, client, auth, RPCBPROC_GETADDR,
+		      (xdrproc_t) xdr_rpcb, &parms,
+		      (xdrproc_t) xdr_wrapstring, &ua);
 	/* First try from start_vers(4) and then version 3 (RPCBVERS) */
 
-	CLNT_CONTROL(client, CLSET_RETRY_TIMEOUT, (char *)&rpcbrmttime);
 	for (vers = start_vers; vers >= RPCBVERS; vers--) {
 		/* Set the version */
 		CLNT_CONTROL(client, CLSET_VERS, (char *)(void *)&vers);
-		clnt_st =
-		    CLNT_CALL(client, auth, RPCBPROC_GETADDR,
-			      (xdrproc_t) xdr_rpcb, (char *)(void *)&parms,
-			      (xdrproc_t) xdr_wrapstring, (char *)(void *)&ua,
-			      *tp);
+
+		clnt_st = clnt_req_setup(cc, tv);
 		if (clnt_st == RPC_SUCCESS) {
+			clnt_st = CLNT_CALL_WAIT(cc);
+		}
+		if (clnt_st == RPC_SUCCESS) {
+			clnt_req_release(cc);
 			if ((ua == NULL) || (ua[0] == 0)) {
 				/* address unknown */
-				rpc_createerr.cf_stat = RPC_PROGNOTREGISTERED;
+				client->cl_error.re_status =
+							RPC_PROGNOTREGISTERED;
 				goto error;
 			}
 			address = uaddr2taddr(nconf, ua);
-#ifdef ND_DEBUG
-			fprintf(stderr, "\tRemote address is [%s]\n", ua);
-			if (!address)
-				fprintf(stderr,
-					"\tCouldn't resolve remote address!\n");
-#endif
-			xdr_free((xdrproc_t) xdr_wrapstring,
-				 (char *)(void *)&ua);
-
 			if (!address) {
 				/* We don't know about your universal address */
-				rpc_createerr.cf_stat = RPC_N2AXLATEFAILURE;
-				goto error;
+				__warnx(TIRPC_DEBUG_FLAG_WARN, "%s: %s",
+					__func__,
+					clnt_sperrno(RPC_N2AXLATEFAILURE));
+				client->cl_error.re_status =
+							RPC_N2AXLATEFAILURE;
+				goto done;
 			}
 			CLNT_CONTROL(client, CLGET_SVC_ADDR,
 				     (char *)(void *)&servaddr);
 			__rpc_fixup_addr(address, &servaddr);
 			goto done;
 		} else if (clnt_st == RPC_PROGVERSMISMATCH) {
-			struct rpc_err rpcerr;
-			clnt_geterr(client, &rpcerr);
-			if (rpcerr.re_vers.low > RPCBVERS4)
+			if (cc->cc_error.re_vers.low > RPCBVERS4) {
+				client->cl_error = cc->cc_error;
+				clnt_req_release(cc);
 				goto error; /* a new version, can't handle */
+			}
 		} else if (clnt_st != RPC_PROGUNAVAIL) {
 			/* Cant handle this error */
-			rpc_createerr.cf_stat = clnt_st;
-			clnt_geterr(client, &rpc_createerr.cf_error);
+			client->cl_error = cc->cc_error;
+			clnt_req_release(cc);
 			goto error;
 		}
+		clnt_req_reset(cc);
 	}
 
 	if ((address == NULL) || (address->len == 0)) {
-		rpc_createerr.cf_stat = RPC_PROGNOTREGISTERED;
-		clnt_geterr(client, &rpc_createerr.cf_error);
+		client->cl_error.re_status = RPC_PROGNOTREGISTERED;
 	}
+	clnt_req_release(cc);
 
  error:
-	if (client) {
-		CLNT_DESTROY(client);
-		client = NULL;
-	}
+	t = rpc_sperror(&client->cl_error, __func__);
+
+	__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB, "%s", t);
+	mem_free(t, RPC_SPERROR_BUFLEN);
+
  done:
-	if (nconf->nc_semantics != NC_TPI_CLTS) {
-		/* This client is the connectionless one */
-		if (client) {
-			CLNT_DESTROY(client);
-			client = NULL;
-		}
-	}
+	if (ua)
+		xdr_free((xdrproc_t) xdr_wrapstring, (char *)(void *)&ua);
 	if (clpp)
 		*clpp = client;
 	else if (client)
@@ -872,7 +895,9 @@ rpcb_getaddr(rpcprog_t program, rpcvers_t version,
 		/* Too long address */
 		mem_free(na->buf, 0);
 		mem_free(na, 0);
-		rpc_createerr.cf_stat = RPC_FAILED;
+		__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB,
+			"%s: address too long (%u > %u)",
+			__func__, na->len, address->maxlen);
 		return (false);
 	}
 	memcpy(address->buf, na->buf, (size_t) na->len);
@@ -893,28 +918,32 @@ rpcblist *
 rpcb_getmaps(const struct netconfig *nconf, const char *host)
 {
 	rpcblist_ptr head = NULL;
+	CLIENT *client;
+	struct clnt_req *cc;
+	char *t;
 	enum clnt_stat clnt_st;
 	rpcvers_t vers = 0;
-	CLIENT *client;
-	AUTH *auth;
 
 	client = getclnthandle(host, nconf, NULL);
-	if (client == NULL)
+	if (CLNT_FAILURE(client)) {
+		CLNT_DESTROY(client);
 		return (head);
+	}
 
-	auth = authnone_ncreate();	/* idempotent */
+	cc = mem_alloc(sizeof(*cc));
+	clnt_req_fill(cc, client, authnone_ncreate(), RPCBPROC_DUMP,
+		      (xdrproc_t) xdr_void, NULL,
+		      (xdrproc_t) xdr_rpcblist_ptr, &head);
+	clnt_st = clnt_req_setup(cc, to);
+	if (clnt_st != RPC_SUCCESS)
+		goto error;
 
-	clnt_st =
-	    CLNT_CALL(client, auth, RPCBPROC_DUMP, (xdrproc_t) xdr_void, NULL,
-		      (xdrproc_t) xdr_rpcblist_ptr, (char *)(void *)&head,
-		      tottimeout);
+	clnt_st = CLNT_CALL_WAIT(cc);
 	if (clnt_st == RPC_SUCCESS)
 		goto done;
 
 	if ((clnt_st != RPC_PROGVERSMISMATCH) && (clnt_st != RPC_PROGUNAVAIL)) {
-		rpc_createerr.cf_stat = RPC_RPCBFAILURE;
-		clnt_geterr(client, &rpc_createerr.cf_error);
-		goto done;
+		goto error;
 	}
 
 	/* fall back to earlier version */
@@ -922,16 +951,24 @@ rpcb_getmaps(const struct netconfig *nconf, const char *host)
 	if (vers == RPCBVERS4) {
 		vers = RPCBVERS;
 		CLNT_CONTROL(client, CLSET_VERS, (char *)(void *)&vers);
-		if (CLNT_CALL
-		    (client, auth, RPCBPROC_DUMP, (xdrproc_t) xdr_void, NULL,
-		     (xdrproc_t) xdr_rpcblist_ptr, (char *)(void *)&head,
-		     tottimeout) == RPC_SUCCESS)
+
+		clnt_req_reset(cc);
+		clnt_st = clnt_req_setup(cc, to);
+		if (clnt_st != RPC_SUCCESS)
+			goto error;
+
+		clnt_st = CLNT_CALL_WAIT(cc);
+		if (clnt_st == RPC_SUCCESS)
 			goto done;
 	}
-	rpc_createerr.cf_stat = RPC_RPCBFAILURE;
-	clnt_geterr(client, &rpc_createerr.cf_error);
+
+ error:
+	t = rpc_sperror(&cc->cc_error, __func__);
+	__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s", t);
+	mem_free(t, RPC_SPERROR_BUFLEN);
 
  done:
+	clnt_req_release(cc);
 	CLNT_DESTROY(client);
 	return (head);
 }
@@ -954,22 +991,21 @@ rpcb_rmtcall(const struct netconfig *nconf, /* Netconfig structure */
 	     const struct netbuf *addr_ptr
 	     /* Preallocated netbuf address */)
 {
-	enum clnt_stat stat;
+	CLIENT *client;
+	struct clnt_req *cc;
 	struct r_rpcb_rmtcallargs a;
 	struct r_rpcb_rmtcallres r;
+	struct timespec tv;
 	rpcvers_t rpcb_vers;
-	CLIENT *client;
-	AUTH *auth;
+	enum clnt_stat stat;
 
-	stat = 0;
 	client = getclnthandle(host, nconf, NULL);
-	if (client == NULL)
+	if (CLNT_FAILURE(client)) {
+		CLNT_DESTROY(client);
 		return (RPC_FAILED);
-
-	auth = authnone_ncreate();	/* idempotent */
+	}
 
 	/*LINTED const castaway */
-	CLNT_CONTROL(client, CLSET_RETRY_TIMEOUT, (char *)(void *)&rmttimeout);
 	a.prog = prog;
 	a.vers = vers;
 	a.proc = proc;
@@ -978,20 +1014,30 @@ rpcb_rmtcall(const struct netconfig *nconf, /* Netconfig structure */
 	r.addr = NULL;
 	r.results.results_val = resp;
 	r.xdr_res = xdrres;
+	tv.tv_sec = tout.tv_sec;
+	tv.tv_nsec = tout.tv_usec * 1000;
+
+	cc = mem_alloc(sizeof(*cc));
+	clnt_req_fill(cc, client, authnone_ncreate(), RPCBPROC_CALLIT,
+		      (xdrproc_t) xdr_rpcb_rmtcallargs, &a,
+		      (xdrproc_t) xdr_rpcb_rmtcallres, &r);
 
 	for (rpcb_vers = RPCBVERS4; rpcb_vers >= RPCBVERS; rpcb_vers--) {
 		CLNT_CONTROL(client, CLSET_VERS, (char *)(void *)&rpcb_vers);
-		stat =
-		    CLNT_CALL(client, auth, RPCBPROC_CALLIT,
-			      (xdrproc_t) xdr_rpcb_rmtcallargs,
-			      (char *)(void *)&a,
-			      (xdrproc_t) xdr_rpcb_rmtcallres,
-			      (char *)(void *)&r, tout);
+
+		stat = clnt_req_setup(cc, tv);
+		if (stat == RPC_SUCCESS) {
+			stat = CLNT_CALL_WAIT(cc);
+		}
+
 		if ((stat == RPC_SUCCESS) && (addr_ptr != NULL)) {
 			struct netbuf *na;
 			/*LINTED const castaway */
 			na = uaddr2taddr((struct netconfig *)nconf, r.addr);
 			if (!na) {
+				__warnx(TIRPC_DEBUG_FLAG_WARN, "%s: %s",
+					__func__,
+					clnt_sperrno(RPC_N2AXLATEFAILURE));
 				stat = RPC_N2AXLATEFAILURE;
 				/*LINTED const castaway */
 				((struct netbuf *)addr_ptr)->len = 0;
@@ -999,6 +1045,9 @@ rpcb_rmtcall(const struct netconfig *nconf, /* Netconfig structure */
 			}
 			if (na->len > addr_ptr->maxlen) {
 				/* Too long address */
+				__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB,
+					"%s: address too long (%u > %u)",
+					__func__, na->len, addr_ptr->maxlen);
 				stat = RPC_FAILED; /* XXX A better error no */
 				mem_free(na->buf, 0);	/* XXX */
 				mem_free(na, 0);
@@ -1016,8 +1065,10 @@ rpcb_rmtcall(const struct netconfig *nconf, /* Netconfig structure */
 			   && (stat != RPC_PROGUNAVAIL)) {
 			goto error;
 		}
+		clnt_req_reset(cc);
 	}
  error:
+	clnt_req_release(cc);
 	CLNT_DESTROY(client);
 	if (r.addr)
 		xdr_free((xdrproc_t) xdr_wrapstring, (char *)(void *)&r.addr);
@@ -1031,12 +1082,12 @@ rpcb_rmtcall(const struct netconfig *nconf, /* Netconfig structure */
 int
 boolrpcb_gettime(const char *host, time_t *timep)
 {
+	CLIENT *client = NULL;
+	struct clnt_req *cc;
 	void *handle;
 	struct netconfig *nconf;
 	rpcvers_t vers;
 	enum clnt_stat st;
-	CLIENT *client = NULL;
-	AUTH *auth;
 
 	if ((host == NULL) || (host[0] == 0)) {
 		time(timep);
@@ -1045,30 +1096,38 @@ boolrpcb_gettime(const char *host, time_t *timep)
 
 	handle = __rpc_setconf("netpath");
 	if (!handle) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
+		__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNPROTO));
 		return (false);
 	}
-	rpc_createerr.cf_stat = RPC_SUCCESS;
-	while (client == NULL) {
+
+	do {
+		if (client)
+			CLNT_DESTROY(client);
+
 		nconf = __rpc_getconf(handle);
 		if (!nconf) {
-			if (rpc_createerr.cf_stat == RPC_SUCCESS)
-				rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
+			__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB, "%s: %s",
+				__func__, clnt_sperrno(RPC_UNKNOWNPROTO));
 			break;
 		}
 		client = getclnthandle(host, nconf, NULL);
-		if (client)
-			break;
-	}
+	} while (CLNT_FAILURE(client));
+
 	__rpc_endconf(handle);
-	if (client == (CLIENT *) NULL)
+	if (CLNT_FAILURE(client)) {
+		CLNT_DESTROY(client);
 		return (false);
+	}
 
-	auth = authnone_ncreate();	/* idempotent */
-
-	st = CLNT_CALL(client, auth, RPCBPROC_GETTIME, (xdrproc_t) xdr_void,
-		       NULL, (xdrproc_t) xdr_int, (char *)(void *)timep,
-		       tottimeout);
+	cc = mem_alloc(sizeof(*cc));
+	clnt_req_fill(cc, client, authnone_ncreate(), RPCBPROC_GETTIME,
+		      (xdrproc_t) xdr_void, NULL,
+		      (xdrproc_t) xdr_int, timep);
+	st = clnt_req_setup(cc, to);
+	if (st == RPC_SUCCESS) {
+		st = CLNT_CALL_WAIT(cc);
+	}
 
 	if ((st == RPC_PROGVERSMISMATCH) || (st == RPC_PROGUNAVAIL)) {
 		CLNT_CONTROL(client, CLGET_VERS, (char *)(void *)&vers);
@@ -1076,12 +1135,23 @@ boolrpcb_gettime(const char *host, time_t *timep)
 			/* fall back to earlier version */
 			vers = RPCBVERS;
 			CLNT_CONTROL(client, CLSET_VERS, (char *)(void *)&vers);
-			st = CLNT_CALL(client, auth, RPCBPROC_GETTIME,
-				       (xdrproc_t) xdr_void, NULL,
-				       (xdrproc_t) xdr_int,
-				       (char *)(void *)timep, tottimeout);
+
+			clnt_req_reset(cc);
+			st = clnt_req_setup(cc, to);
+			if (st == RPC_SUCCESS) {
+				st = CLNT_CALL_WAIT(cc);
+			}
 		}
 	}
+
+	if (st != RPC_SUCCESS) {
+		char *t = rpc_sperror(&cc->cc_error, __func__);
+
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
+	}
+
+	clnt_req_release(cc);
 	CLNT_DESTROY(client);
 	return (st == RPC_SUCCESS ? true : false);
 }
@@ -1092,25 +1162,45 @@ boolrpcb_gettime(const char *host, time_t *timep)
  */
 char *rpcb_taddr2uaddr(struct netconfig *nconf, struct netbuf *taddr)
 {
-	CLIENT *client;
-	AUTH *auth = authnone_ncreate();
 	char *uaddr = NULL;
+	CLIENT *client;
+	struct clnt_req *cc;
+	enum clnt_stat st;
 
 	/* parameter checking */
 	if (nconf == NULL) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNPROTO));
 		return (NULL);
 	}
 	if (taddr == NULL) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNADDR));
 		return (NULL);
 	}
-	client = local_rpcb();
-	if (!client)
-		return (NULL);
-	CLNT_CALL(client, auth, RPCBPROC_TADDR2UADDR, (xdrproc_t) xdr_netbuf,
-		  (char *)(void *)taddr, (xdrproc_t) xdr_wrapstring,
-		  (char *)(void *)&uaddr, tottimeout);
+	client = local_rpcb(__func__);
+	if (CLNT_FAILURE(client)) {
+		CLNT_DESTROY(client);
+		return (false);
+	}
+
+	cc = mem_alloc(sizeof(*cc));
+	clnt_req_fill(cc, client, authnone_ncreate(), RPCBPROC_TADDR2UADDR,
+		      (xdrproc_t) xdr_netbuf, taddr,
+		      (xdrproc_t) xdr_wrapstring, &uaddr);
+	st = clnt_req_setup(cc, to);
+	if (st == RPC_SUCCESS) {
+		st = CLNT_CALL_WAIT(cc);
+	}
+
+	if (st != RPC_SUCCESS) {
+		char *t = rpc_sperror(&cc->cc_error, __func__);
+
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
+	}
+
+	clnt_req_release(cc);
 	CLNT_DESTROY(client);
 	return (uaddr);
 }
@@ -1123,35 +1213,46 @@ struct netbuf *rpcb_uaddr2taddr(struct netconfig *nconf, char *uaddr)
 {
 	struct netbuf *taddr;
 	CLIENT *client;
-	AUTH *auth;
+	struct clnt_req *cc;
+	enum clnt_stat st;
 
 	/* parameter checking */
 	if (nconf == NULL) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNPROTO));
 		return (NULL);
 	}
 	if (uaddr == NULL) {
-		rpc_createerr.cf_stat = RPC_UNKNOWNADDR;
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s: %s",
+			__func__, clnt_sperrno(RPC_UNKNOWNADDR));
 		return (NULL);
 	}
-	client = local_rpcb();
-	if (!client)
-		return (NULL);
-
-	auth = authnone_ncreate();	/* idempotent */
-
-	taddr = (struct netbuf *)mem_zalloc(sizeof(struct netbuf));
-	if (taddr == NULL) {
+	client = local_rpcb(__func__);
+	if (CLNT_FAILURE(client)) {
 		CLNT_DESTROY(client);
-		return (NULL);
+		return (false);
 	}
-	if (CLNT_CALL
-	    (client, auth, RPCBPROC_UADDR2TADDR, (xdrproc_t) xdr_wrapstring,
-	     (char *)(void *)&uaddr, (xdrproc_t) xdr_netbuf,
-	     (char *)(void *)taddr, tottimeout) != RPC_SUCCESS) {
-		mem_free(taddr, 0);	/* XXX */
+
+	cc = mem_alloc(sizeof(*cc));
+	taddr = (struct netbuf *)mem_zalloc(sizeof(struct netbuf));
+	clnt_req_fill(cc, client, authnone_ncreate(), RPCBPROC_UADDR2TADDR,
+		      (xdrproc_t) xdr_wrapstring, &uaddr,
+		      (xdrproc_t) xdr_netbuf, taddr);
+	st = clnt_req_setup(cc, to);
+	if (st == RPC_SUCCESS) {
+		st = CLNT_CALL_WAIT(cc);
+	}
+
+	if (st != RPC_SUCCESS) {
+		char *t = rpc_sperror(&cc->cc_error, __func__);
+
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, "%s", t);
+		mem_free(t, RPC_SPERROR_BUFLEN);
+		mem_free(taddr, sizeof(*taddr));
 		taddr = NULL;
 	}
+
+	clnt_req_release(cc);
 	CLNT_DESTROY(client);
 	return (taddr);
 }
@@ -1162,18 +1263,20 @@ struct netbuf *rpcb_uaddr2taddr(struct netconfig *nconf, char *uaddr)
 
 /*
  * This routine will return a client handle that is connected to the local
- * rpcbind. Returns NULL on error and free's everything.
+ * rpcbind.
+ * On error, a CLIENT with cl_error.re_status set
  */
-static CLIENT *local_rpcb(void)
+static CLIENT *local_rpcb(const char *tag)
 {
-	CLIENT *client;
+	CLIENT *client = NULL;
+	char *t;
 	static struct netconfig *loopnconf;
 	static char *hostname;
 	extern mutex_t loopnconf_lock;
-	int sock;
-	size_t tsize;
 	struct netbuf nbuf;
 	struct sockaddr_un sun;
+	size_t tsize;
+	int sock;
 
 	/*
 	 * Try connecting to the local rpcbind through a local socket
@@ -1191,15 +1294,20 @@ static CLIENT *local_rpcb(void)
 	nbuf.buf = &sun;
 
 	tsize = __rpc_get_t_size(AF_LOCAL, 0, 0);
-	client =
-	    clnt_vc_ncreate(sock, &nbuf, (rpcprog_t) RPCBPROG,
-			    (rpcvers_t) RPCBVERS, tsize, tsize);
+	client = clnt_vc_ncreatef(sock, &nbuf, (rpcprog_t) RPCBPROG,
+				  (rpcvers_t) RPCBVERS, tsize, tsize,
+				  CLNT_CREATE_FLAG_CLOSE |
+				  CLNT_CREATE_FLAG_CONNECT);
 
-	if (client != NULL) {
-		/* Mark the socket to be closed in destructor */
-		(void)CLNT_CONTROL(client, CLSET_FD_CLOSE, NULL);
+	if (CLNT_SUCCESS(client)) {
 		return client;
 	}
+	t = rpc_sperror(&client->cl_error, tag);
+
+	__warnx(TIRPC_DEBUG_FLAG_CLNT_RPCB, "%s", t);
+	mem_free(t, RPC_SPERROR_BUFLEN);
+
+	/* Save client for error return */
 
 	/* Nobody needs this socket anymore; free the descriptor. */
 	close(sock);
@@ -1216,9 +1324,12 @@ static CLIENT *local_rpcb(void)
 		if (nc_handle == NULL) {
 			/* fails to open netconfig file */
 			syslog(LOG_ERR, "rpc: failed to open " NETCONFIG);
-			rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
 			mutex_unlock(&loopnconf_lock);
-			return (NULL);
+			__warnx(TIRPC_DEBUG_FLAG_WARN,
+				"%s(%s): failed to open " NETCONFIG " %s",
+				__func__, tag, clnt_sperrno(RPC_UNKNOWNPROTO));
+			client->cl_error.re_status = RPC_UNKNOWNPROTO;
+			return (client);
 		}
 		while ((nconf = getnetconfig(nc_handle)) != NULL) {
 #ifdef INET6
@@ -1245,16 +1356,22 @@ static CLIENT *local_rpcb(void)
 			}
 		}
 		if (tmpnconf == NULL) {
-			rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
 			endnetconfig(nc_handle);
 			mutex_unlock(&loopnconf_lock);
-			return (NULL);
+			__warnx(TIRPC_DEBUG_FLAG_WARN,
+				"%s(%s): failed to find " NETCONFIG " %s",
+				__func__, tag, clnt_sperrno(RPC_UNKNOWNPROTO));
+			client->cl_error.re_status = RPC_UNKNOWNPROTO;
+			return (client);
 		}
 		loopnconf = getnetconfigent(tmpnconf->nc_netid);
 		/* loopnconf is never freed */
 		endnetconfig(nc_handle);
 	}
 	mutex_unlock(&loopnconf_lock);
+	/* Free client used for errors before */
+	CLNT_DESTROY(client);
+
 	client = getclnthandle(hostname, loopnconf, NULL);
 	return (client);
 }
