@@ -39,28 +39,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <rpc/rpc.h>
-#include <rpc/svc.h>
 #include <rpc/svc_auth.h>
-#include "rpc_com.h"
 #include <rpc/gss_internal.h>
 #include <misc/portable.h>
 
-static bool svcauth_gss_wrap(SVCAUTH *, struct svc_req *, XDR *, xdrproc_t,
-			     caddr_t);
-
-static bool svcauth_gss_unwrap(SVCAUTH *, struct svc_req *, XDR *, xdrproc_t,
-			       caddr_t);
-
-static bool svcauth_gss_release(SVCAUTH *, struct svc_req *);
-
-bool svcauth_gss_destroy(SVCAUTH *);
-
-struct svc_auth_ops svc_auth_gss_ops = {
-	svcauth_gss_wrap,
-	svcauth_gss_unwrap,
-	svcauth_gss_release,
-	svcauth_gss_destroy
-};
+static struct svc_auth_ops svc_auth_gss_ops;
 
 #define SVCAUTH_PRIVATE(auth) \
 	((struct svc_rpc_gss_data *)(auth)->svc_ah_private)
@@ -208,16 +191,18 @@ svcauth_gss_accept_sec_context(struct svc_req *req,
 #define INDEF_EXPIRE 60*60*24	/* from mit k5 src/lib/rpc/svc_auth_gssapi.c */
 	OM_uint32 time_rec;
 
-	gc = (struct rpc_gss_cred *)req->rq_clntcred;
+	gc = (struct rpc_gss_cred *)req->rq_msg.rq_cred_body;
 	memset(gr, 0, sizeof(*gr));
 
 	/* Deserialize arguments. */
 	memset(&recv_tok, 0, sizeof(recv_tok));
 
-	if (!svc_getargs
-	    (req->rq_xprt, req, (xdrproc_t) xdr_rpc_gss_init_args,
-	     (caddr_t) &recv_tok, NULL /* u_data */))
+	req->rq_msg.rm_xdr.where = (caddr_t)&recv_tok;
+	req->rq_msg.rm_xdr.proc = (xdrproc_t)xdr_rpc_gss_init_args;
+	if (!SVCAUTH_UNWRAP(req)) {
+		xdr_free((xdrproc_t)xdr_rpc_gss_init_args, (caddr_t)&recv_tok);
 		return (false);
+	}
 
 	gr->gr_major =
 	    gss_accept_sec_context(&gr->gr_minor, &gd->ctx, svcauth_gss_creds,
@@ -225,8 +210,7 @@ svcauth_gss_accept_sec_context(struct svc_req *req,
 				   &gd->client_name, &mech, &gr->gr_token,
 				   &ret_flags, &time_rec, NULL);
 
-	svc_freeargs(req->rq_xprt, req, (xdrproc_t) xdr_rpc_gss_init_args,
-		     (caddr_t) &recv_tok);
+	xdr_free((xdrproc_t)xdr_rpc_gss_init_args, (caddr_t)&recv_tok);
 
 	if ((gr->gr_major != GSS_S_COMPLETE)
 	    && (gr->gr_major != GSS_S_CONTINUE_NEEDED)) {
@@ -305,9 +289,14 @@ svcauth_gss_accept_sec_context(struct svc_req *req,
 		}
 
 		/* XXX ref? (assert gd->locked?) */
-		req->rq_verf.oa_flavor = RPCSEC_GSS;
-		req->rq_verf.oa_base = checksum.value;
-		req->rq_verf.oa_length = checksum.length;
+		if (checksum.length > MAX_AUTH_BYTES){
+			gss_release_buffer(&min_stat, &gr->gr_token);
+			return (false);
+		}
+		req->rq_msg.RPCM_ack.ar_verf.oa_flavor = RPCSEC_GSS;
+		req->rq_msg.RPCM_ack.ar_verf.oa_length = checksum.length;
+		memcpy(req->rq_msg.RPCM_ack.ar_verf.oa_body, checksum.value,
+		       checksum.length);
 	}
 	return (true);
 }
@@ -316,19 +305,18 @@ svcauth_gss_accept_sec_context(struct svc_req *req,
 
 static int
 svcauth_gss_validate(struct svc_req *req,
-		     struct svc_rpc_gss_data *gd,
-		     struct rpc_msg *msg)
+		     struct svc_rpc_gss_data *gd)
 {
 	struct opaque_auth *oa;
+	int32_t *buf;
 	gss_buffer_desc rpcbuf, checksum;
 	OM_uint32 maj_stat, min_stat, qop_state;
 	u_char rpchdr[RPCHDR_LEN];
-	int32_t *buf;
 
 	memset(rpchdr, 0, RPCHDR_LEN);
 
 	/* XXX - Reconstruct RPC header for signing (from xdr_callmsg). */
-	oa = &msg->rm_call.cb_cred;
+	oa = &req->rq_msg.cb_cred;
 	if (oa->oa_length > MAX_AUTH_BYTES)
 		return GSS_S_CALL_BAD_STRUCTURE;
 	/* XXX since MAX_AUTH_BYTES is 400, the following code trivially
@@ -336,23 +324,23 @@ svcauth_gss_validate(struct svc_req *req,
 	 * what is marshalled below). */
 
 	buf = (int32_t *) rpchdr;
-	IXDR_PUT_LONG(buf, msg->rm_xid);
-	IXDR_PUT_ENUM(buf, msg->rm_direction);
-	IXDR_PUT_LONG(buf, msg->rm_call.cb_rpcvers);
-	IXDR_PUT_LONG(buf, msg->rm_call.cb_prog);
-	IXDR_PUT_LONG(buf, msg->rm_call.cb_vers);
-	IXDR_PUT_LONG(buf, msg->rm_call.cb_proc);
+	IXDR_PUT_LONG(buf, req->rq_msg.rm_xid);
+	IXDR_PUT_ENUM(buf, req->rq_msg.rm_direction);
+	IXDR_PUT_LONG(buf, req->rq_msg.rm_call.cb_rpcvers);
+	IXDR_PUT_LONG(buf, req->rq_msg.cb_prog);
+	IXDR_PUT_LONG(buf, req->rq_msg.cb_vers);
+	IXDR_PUT_LONG(buf, req->rq_msg.cb_proc);
 	IXDR_PUT_ENUM(buf, oa->oa_flavor);
 	IXDR_PUT_LONG(buf, oa->oa_length);
 	if (oa->oa_length) {
-		memcpy((caddr_t) buf, oa->oa_base, oa->oa_length);
+		memcpy((caddr_t) buf, oa->oa_body, oa->oa_length);
 		buf += RNDUP(oa->oa_length) / sizeof(int32_t);
 	}
 	rpcbuf.value = rpchdr;
 	rpcbuf.length = (u_char *) buf - rpchdr;
 
-	checksum.value = msg->rm_call.cb_verf.oa_base;
-	checksum.length = msg->rm_call.cb_verf.oa_length;
+	checksum.value = req->rq_msg.cb_verf.oa_body;
+	checksum.length = req->rq_msg.cb_verf.oa_length;
 
 	maj_stat =
 	    gss_verify_mic(&min_stat, gd->ctx, &rpcbuf, &checksum, &qop_state);
@@ -360,9 +348,8 @@ svcauth_gss_validate(struct svc_req *req,
 	if (maj_stat != GSS_S_COMPLETE) {
 		__warnx(TIRPC_DEBUG_FLAG_AUTH, "%s: %d %d", __func__, maj_stat,
 			min_stat);
-		return (maj_stat);
 	}
-	return GSS_S_COMPLETE;
+	return (maj_stat);
 }
 
 bool
@@ -379,12 +366,17 @@ svcauth_gss_nextverf(struct svc_req *req, struct svc_rpc_gss_data *gd,
 	    gss_get_mic(&min_stat, gd->ctx, gd->sec.qop, &signbuf, &checksum);
 
 	if (maj_stat != GSS_S_COMPLETE) {
-		log_status("gss_get_mic", maj_stat, min_stat);
+		gss_log_status("gss_get_mic", maj_stat, min_stat);
 		return (false);
 	}
-	req->rq_verf.oa_flavor = RPCSEC_GSS;
-	req->rq_verf.oa_base = (caddr_t) checksum.value;
-	req->rq_verf.oa_length = (u_int) checksum.length;
+	if (checksum.length > MAX_AUTH_BYTES) {
+		gss_log_status("checksum.length", maj_stat, min_stat);
+		return (false);
+	}
+	req->rq_msg.RPCM_ack.ar_verf.oa_flavor = RPCSEC_GSS;
+	req->rq_msg.RPCM_ack.ar_verf.oa_length = checksum.length;
+	memcpy(req->rq_msg.RPCM_ack.ar_verf.oa_body, checksum.value,
+	       checksum.length);
 
 	return (true);
 }
@@ -399,8 +391,7 @@ svcauth_gss_nextverf(struct svc_req *req, struct svc_rpc_gss_data *gd,
 	} while (0)
 
 enum auth_stat
-_svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
-	     bool *no_dispatch)
+_svcauth_gss(struct svc_req *req, bool *no_dispatch)
 {
 	XDR xdrs[1];
 	SVCAUTH *auth;
@@ -413,17 +404,17 @@ _svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
 	bool gd_hashed = false;
 
 	/* Initialize reply. */
-	req->rq_verf = _null_auth;
+	req->rq_msg.RPCM_ack.ar_verf = _null_auth;
 
 	/* Unserialize client credentials. */
-	if (req->rq_cred.oa_length <= 0)
+	if (req->rq_msg.cb_cred.oa_length <= 0)
 		svcauth_gss_return(AUTH_BADCRED);
 
-	gc = (struct rpc_gss_cred *)req->rq_clntcred;
+	gc = (struct rpc_gss_cred *)req->rq_msg.rq_cred_body;
 	memset(gc, 0, sizeof(struct rpc_gss_cred));
 
-	xdrmem_create(xdrs, req->rq_cred.oa_base, req->rq_cred.oa_length,
-		      XDR_DECODE);
+	xdrmem_create(xdrs, req->rq_msg.cb_cred.oa_body,
+		      req->rq_msg.cb_cred.oa_length, XDR_DECODE);
 
 	if (!xdr_rpc_gss_cred(xdrs, gc)) {
 		XDR_DESTROY(xdrs);
@@ -513,7 +504,7 @@ _svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
 	case RPCSEC_GSS_INIT:
 	case RPCSEC_GSS_CONTINUE_INIT:
 
-		if (req->rq_proc != NULLPROC)
+		if (req->rq_msg.cb_proc != NULLPROC)
 			svcauth_gss_return(AUTH_FAILED); /* XXX ? */
 
 		/* XXX why unconditionally acquire creds? */
@@ -532,17 +523,17 @@ _svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
 
 		*no_dispatch = true;
 
-		call_stat =
-		    svc_sendreply(req->rq_xprt, req,
-				  (xdrproc_t) xdr_rpc_gss_init_res,
-				  (caddr_t) &gr);
+		req->rq_msg.RPCM_ack.ar_results.where = (caddr_t) &gr;
+		req->rq_msg.RPCM_ack.ar_results.proc =
+					(xdrproc_t) xdr_rpc_gss_init_res;
+		call_stat = svc_sendreply(req);
 
 		/* XXX */
 		gss_release_buffer(&min_stat, &gr.gr_token);
 		gss_release_buffer(&min_stat, &gd->checksum);
 		mem_free(gr.gr_ctx.value, 0);
 
-		if (!call_stat)
+		if (call_stat >= XPRT_DIED)
 			svcauth_gss_return(AUTH_FAILED);
 
 		if (gr.gr_major == GSS_S_COMPLETE) {
@@ -588,7 +579,7 @@ _svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
 		 * after a validate or verf failure ? */
 
 	case RPCSEC_GSS_DATA:
-		call_stat = svcauth_gss_validate(req, gd, msg);
+		call_stat = svcauth_gss_validate(req, gd);
 		switch (call_stat) {
 		default:
 			svcauth_gss_return(RPCSEC_GSS_CREDPROBLEM);
@@ -601,10 +592,10 @@ _svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
 		break;
 
 	case RPCSEC_GSS_DESTROY:
-		if (req->rq_proc != NULLPROC)
+		if (req->rq_msg.cb_proc != NULLPROC)
 			svcauth_gss_return(AUTH_FAILED);	/* XXX ? */
 
-		if (svcauth_gss_validate(req, gd, msg))
+		if (svcauth_gss_validate(req, gd))
 			svcauth_gss_return(RPCSEC_GSS_CREDPROBLEM);
 
 		if (!svcauth_gss_nextverf(req, gd, htonl(gc->gc_seq)))
@@ -618,9 +609,13 @@ _svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
 		mutex_unlock(&gd->lock);
 		gd_locked = false;
 
-		call_stat =
-		    svc_sendreply(req->rq_xprt, req, (xdrproc_t) xdr_void,
-				  (caddr_t) NULL);
+		req->rq_msg.RPCM_ack.ar_results.where = NULL;
+		req->rq_msg.RPCM_ack.ar_results.proc = (xdrproc_t) xdr_void;
+		if (svc_sendreply(req) >= XPRT_DIED) {
+			__warnx(TIRPC_DEBUG_FLAG_ERROR,
+				"%s() svc_sendreply failed",
+				__func__);
+		}
 
 		/* We acquired a reference on gd with authgss_ctx_hash_get
 		 * call.  Time to release the reference as we don't need
@@ -640,23 +635,14 @@ _svcauth_gss(struct svc_req *req, struct rpc_msg *msg,
 }
 
 static bool
-svcauth_gss_release(SVCAUTH *auth, struct svc_req *req)
+svcauth_gss_release(struct svc_req *req)
 {
 	struct svc_rpc_gss_data *gd;
-	caddr_t last_oa_base;
 
-	gd = SVCAUTH_PRIVATE(auth);
+	gd = SVCAUTH_PRIVATE(req->rq_auth);
 	if (gd)
 		unref_svc_rpc_gss_data(gd, SVC_RPC_GSS_FLAG_NONE);
 	req->rq_auth = NULL;
-
-	if ((last_oa_base = req->rq_verf.oa_base)) {
-		/* XXX wrapper conflict: mem_free vs. gssalloc_free ? */
-		/* ... but this only matters for win32 | kernel */
-		req->rq_verf.oa_base = 0;
-		mem_free(last_oa_base, req->rq_verf.oa_length);
-	}
-
 	return (true);
 }
 
@@ -686,42 +672,158 @@ svcauth_gss_destroy(SVCAUTH *auth)
 	return (true);
 }
 
-bool
-svcauth_gss_wrap(SVCAUTH *auth, struct svc_req *req, XDR *xdrs,
-		 xdrproc_t xdr_func, caddr_t xdr_ptr)
+static bool
+svcauth_gss_wrap(struct svc_req *req, XDR *xdrs)
 {
-	bool result;
 	struct svc_rpc_gss_data *gd = SVCAUTH_PRIVATE(req->rq_auth);
 	u_int gc_seq = (u_int) (uintptr_t) req->rq_ap1;
+	struct rpc_gss_cred *gc = (struct rpc_gss_cred *)
+					req->rq_msg.rq_cred_body;
+	bool result;
 
-	if (!gd->established || gd->sec.svc == RPCSEC_GSS_SVC_NONE)
-		return ((*xdr_func) (xdrs, xdr_ptr));
+	if (!gd->established || gc->gc_svc == RPCSEC_GSS_SVC_NONE)
+		return (svc_auth_none.svc_ah_ops->svc_ah_wrap(req, xdrs));
 
 	mutex_lock(&gd->lock);
-	result = xdr_rpc_gss_data(xdrs, xdr_func, xdr_ptr, gd->ctx,
-				  gd->sec.qop, gd->sec.svc, gc_seq);
+	result = xdr_rpc_gss_wrap(xdrs, req->rq_msg.RPCM_ack.ar_results.proc,
+				  req->rq_msg.RPCM_ack.ar_results.where,
+				  gd->ctx, gd->sec.qop, gc->gc_svc, gc_seq);
 	mutex_unlock(&gd->lock);
 	return (result);
 }
 
-bool
-svcauth_gss_unwrap(SVCAUTH *auth, struct svc_req *req, XDR *xdrs,
-		   xdrproc_t xdr_func, caddr_t xdr_ptr)
+static bool
+svcauth_gss_unwrap(struct svc_req *req)
 {
-	bool result;
 	struct svc_rpc_gss_data *gd = SVCAUTH_PRIVATE(req->rq_auth);
 	u_int gc_seq = (u_int) (uintptr_t) req->rq_ap1;
+	bool result;
 
 	if (!gd->established || gd->sec.svc == RPCSEC_GSS_SVC_NONE)
-		return ((*xdr_func) (xdrs, xdr_ptr));
+		return (svc_auth_none.svc_ah_ops->svc_ah_unwrap(req));
 
 	mutex_lock(&gd->lock);
-	result = xdr_rpc_gss_data(xdrs, xdr_func, xdr_ptr, gd->ctx,
-				  gd->sec.qop, gd->sec.svc, gc_seq);
+	result = xdr_rpc_gss_unwrap(req->rq_xdrs, req->rq_msg.rm_xdr.proc,
+				    req->rq_msg.rm_xdr.where, gd->ctx,
+				    gd->sec.qop, gd->sec.svc, gc_seq);
 	mutex_unlock(&gd->lock);
 	return (result);
 }
 
+static inline bool
+xdr_rpc_gss_checksum(struct svc_req *req, gss_ctx_id_t ctx, gss_qop_t qop,
+		     rpc_gss_svc_t svc, u_int seq)
+{
+	XDR *xdrs = req->rq_xdrs;
+	XDR tmpxdrs;
+	gss_buffer_desc databuf, wrapbuf;
+	OM_uint32 maj_stat, min_stat;
+	u_int seq_num, qop_state;
+	int conf_state;
+	bool xdr_stat;
+
+	if (req->rq_msg.rm_xdr.proc == (xdrproc_t) xdr_void
+	 || req->rq_msg.rm_xdr.where == NULL)
+		return (TRUE);
+
+	memset(&databuf, 0, sizeof(databuf));
+	memset(&wrapbuf, 0, sizeof(wrapbuf));
+
+	if (svc == RPCSEC_GSS_SVC_INTEGRITY) {
+		/* Decode databody_integ. */
+		if (!xdr_rpc_gss_decode(xdrs, &databuf)) {
+			__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS,
+				"%s() xdr_rpc_gss_decode databody_integ failed",
+				__func__);
+			return (FALSE);
+		}
+		/* Decode checksum. */
+		if (!xdr_rpc_gss_decode(xdrs, &wrapbuf)) {
+			gss_release_buffer(&min_stat, &databuf);
+			__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS,
+				"%s() xdr_rpc_gss_decode checksum failed",
+				__func__);
+			return (FALSE);
+		}
+		/* Verify checksum and QOP. */
+		maj_stat =
+		    gss_verify_mic(&min_stat, ctx, &databuf, &wrapbuf,
+				   &qop_state);
+		gss_release_buffer(&min_stat, &wrapbuf);
+
+		if (maj_stat != GSS_S_COMPLETE || qop_state != qop) {
+			gss_release_buffer(&min_stat, &databuf);
+			gss_log_status("gss_verify_mic", maj_stat, min_stat);
+			return (FALSE);
+		}
+	} else if (svc == RPCSEC_GSS_SVC_PRIVACY) {
+		/* Decode databody_priv. */
+		if (!xdr_rpc_gss_decode(xdrs, &wrapbuf)) {
+			__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS,
+				"%s() xdr_rpc_gss_decode databody_priv failed",
+				__func__);
+			return (FALSE);
+		}
+		/* Decrypt databody. */
+		maj_stat =
+		    gss_unwrap(&min_stat, ctx, &wrapbuf, &databuf, &conf_state,
+			       &qop_state);
+
+		gss_release_buffer(&min_stat, &wrapbuf);
+
+		/* Verify encryption and QOP. */
+		if (maj_stat != GSS_S_COMPLETE || qop_state != qop
+		    || conf_state != TRUE) {
+			gss_release_buffer(&min_stat, &databuf);
+			gss_log_status("gss_unwrap", maj_stat, min_stat);
+			return (FALSE);
+		}
+	}
+	/* Decode rpc_gss_data_t (sequence number + arguments). */
+	xdrmem_create(&tmpxdrs, databuf.value, databuf.length, XDR_DECODE);
+	SVC_CHECKSUM(req, databuf.value, databuf.length);
+	xdr_stat = (xdr_u_int(&tmpxdrs, &seq_num)
+		    && (*req->rq_msg.rm_xdr.proc)
+			(&tmpxdrs, req->rq_msg.rm_xdr.where));
+	XDR_DESTROY(&tmpxdrs);
+	gss_release_buffer(&min_stat, &databuf);
+
+	/* Verify sequence number. */
+	if (xdr_stat == TRUE && seq_num != seq) {
+		__warnx(TIRPC_DEBUG_FLAG_RPCSEC_GSS,
+			"%s() wrong sequence number in databody",
+			__func__);
+		return (FALSE);
+	}
+	return (xdr_stat);
+}
+
+static bool
+svcauth_gss_checksum(struct svc_req *req)
+{
+	struct svc_rpc_gss_data *gd = SVCAUTH_PRIVATE(req->rq_auth);
+	u_int gc_seq = (u_int) (uintptr_t) req->rq_ap1;
+	bool result;
+
+	if (!gd->established || gd->sec.svc == RPCSEC_GSS_SVC_NONE) {
+		return (svc_auth_none.svc_ah_ops->svc_ah_checksum(req));
+	}
+
+	mutex_lock(&gd->lock);
+	result = xdr_rpc_gss_checksum(req, gd->ctx, gd->sec.qop, gd->sec.svc,
+				      gc_seq);
+	mutex_unlock(&gd->lock);
+	return (result);
+}
+
+static struct svc_auth_ops svc_auth_gss_ops = {
+	svcauth_gss_wrap,
+	svcauth_gss_unwrap,
+	svcauth_gss_checksum,
+	svcauth_gss_release,
+	svcauth_gss_destroy
+};
+
 char *
 svcauth_gss_get_principal(SVCAUTH *auth)
 {
