@@ -26,16 +26,20 @@
 #ifndef TIRPC_SVC_INTERNAL_H
 #define TIRPC_SVC_INTERNAL_H
 
+#include <sys/socket.h>
+#include <netinet/in.h>
 #include <misc/os_epoll.h>
+#include <rpc/rpc_msg.h>
+
+#include "rpc_dplx_internal.h"
 
 extern int __svc_maxiov;
 extern int __svc_maxrec;
 
 /* threading fdsets around is annoying */
 struct svc_params {
-	bool initialized;
 	mutex_t mtx;
-	u_long flags;
+	bool initialized;
 
 	/* package global event handling--may be overridden using the
 	 * svc_rqst interface */
@@ -50,9 +54,8 @@ struct svc_params {
 		} fd;
 	} ev_u;
 
-	int32_t idle_timeout;
-	u_int max_connections;
-	u_int svc_ioq_maxbuf;
+	svc_xprt_fun_t disconnect_cb;
+	svc_xprt_xdr_fun_t request_cb;
 
 	union {
 		struct {
@@ -69,8 +72,13 @@ struct svc_params {
 	} gss;
 
 	struct {
+		u_int send_max;
 		u_int thrd_max;
 	} ioq;
+
+	u_long flags;
+	u_int max_connections;
+	int32_t idle_timeout;
 };
 
 extern struct svc_params __svc_params[1];
@@ -112,76 +120,44 @@ svc_vc_new_conn_ok(void)
 		mutex_unlock(&(__svc_params->xprt_u.vc.mtx)); \
 	} while (0)
 
-struct __svc_ops {
-	bool (*svc_clean_idle) (fd_set *fds, int timeout, bool cleanblock);
-	void (*svc_run) (void);
-	void (*svc_getreq) (int rdfds);	/* XXX */
-	void (*svc_getreqset) (fd_set *readfds); /* XXX */
-	void (*svc_exit) (void);
-};
-
-extern struct __svc_ops *svc_ops;
-
-#define	su_data(xprt)	((struct svc_dg_data *)(xprt->xp_p2))
-
-/*  The CACHING COMPONENT */
 
 /*
- * Could have been a separate file, but some part of it depends upon the
- * private structure of the client handle.
+ * The following union is defined just to use SVC_CMSG_SIZE macro for an array
+ * length. _GNU_SOURCE must be defined to get in6_pktinfo declaration!
+ */
+union pktinfo_u {
+	struct in_pktinfo x;
+	struct in6_pktinfo y;
+};
+#define SVC_CMSG_SIZE CMSG_SPACE(sizeof(union pktinfo_u))
+
+/**
+ * \struct svc_dg_xprt
+ * DG transport instance
  *
- * Fifo cache for cl server
- * Copies pointers to reply buffers into fifo cache
- * Buffers are sent again if retransmissions are detected.
+ * Replaces old struct svc_dg_data by locally wrapping struct rpc_dplx_rec,
+ * which wraps struct svc_xprt indexed by fd.
  */
-
-#define	SPARSENESS 4		/* 75% sparse */
-
-/*
- * An entry in the cache
- */
-typedef struct cache_node *cache_ptr;
-struct cache_node {
-	/*
-	 * Index into cache is xid, proc, vers, prog and address
-	 */
-	u_int32_t cache_xid;
-	rpcproc_t cache_proc;
-	rpcvers_t cache_vers;
-	rpcprog_t cache_prog;
-	struct netbuf cache_addr;
-	/*
-	 * The cached reply and length
-	 */
-	char *cache_reply;
-	size_t cache_replylen;
-	/*
-	 * Next node on the list, if there is a collision
-	 */
-	cache_ptr cache_next;
+struct svc_dg_xprt {
+	struct rpc_dplx_rec su_dr;	/* SVCXPRT indexed by fd */
+	struct msghdr su_msghdr;	/* msghdr received from clnt */
+	unsigned char su_cmsg[SVC_CMSG_SIZE];	/* cmsghdr received from clnt */
 };
+#define DG_DR(p) (opr_containerof((p), struct svc_dg_xprt, su_dr))
+#define su_data(xprt) (DG_DR(REC_XPRT(xprt)))
 
-/*
- * the hashing function
+/**
+ * \struct svc_vc_xprt
+ * VC transport instance
+ *
+ * Replaces old struct x_vc_data by locally wrapping struct rpc_dplx_rec,
+ * which wraps struct svc_xprt indexed by fd.
  */
-#define	CACHE_LOC(transp, xid)	\
-	(xid % (SPARSENESS * ((struct cl_cache *) \
-		su_data(transp)->su_cache)->uc_size))
-
-extern mutex_t dupreq_lock;
-
-/*
- * The entire cache
- */
-struct cl_cache {
-	u_int uc_size;		/* size of cache */
-	cache_ptr *uc_entries;	/* hash table of entries in cache */
-	cache_ptr *uc_fifo;	/* fifo list of entries in cache */
-	u_int uc_nextvictim;	/* points to next victim in fifo list */
-	rpcprog_t uc_prog;	/* saved program number */
-	rpcvers_t uc_vers;	/* saved version number */
-	rpcproc_t uc_proc;	/* saved procedure number */
+struct svc_vc_xprt {
+	struct rpc_dplx_rec sx_dr;	/* SVCXPRT indexed by fd */
+	int32_t sx_fbtbc;		/* fragment bytes to be consumed */
 };
+#define VC_DR(p) (opr_containerof((p), struct svc_vc_xprt, sx_dr))
 
 /* Epoll interface change */
 #ifndef EPOLL_CLOEXEC
@@ -203,6 +179,21 @@ epoll_create_wr(size_t size, int flags)
 extern void __rpc_set_blkin_endpoint(SVCXPRT *xprt, const char *tag);
 #endif
 
-void svc_rqst_shutdown(void);
+enum xprt_stat svc_rendezvous_stat(SVCXPRT *);
+
+static inline void
+svc_override_ops(struct xp_ops *ops, SVCXPRT *rendezvous)
+{
+	if (rendezvous) {
+		if (!ops->xp_free_user_data)
+			ops->xp_free_user_data =
+				rendezvous->xp_ops->xp_free_user_data;
+	}
+}
+
+/* in svc_rqst.c */
+int svc_rqst_rearm_events(SVCXPRT *);
+int svc_rqst_xprt_register(SVCXPRT *, SVCXPRT *);
+void svc_rqst_xprt_unregister(SVCXPRT *);
 
 #endif				/* TIRPC_SVC_INTERNAL_H */
