@@ -1,6 +1,6 @@
-
 /*
  * Copyright (c) 2009, Sun Microsystems, Inc.
+ * Copyright (c) 2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,11 +40,14 @@
 #include <config.h>
 #include <pthread.h>
 #include <reentrant.h>
-#include <rpc/rpc.h>
 #include <sys/types.h>
-#include <rpc/raw.h>
 #include <stdlib.h>
 
+#include <rpc/rpc.h>
+#include <rpc/svc_rqst.h>
+#include "rpc_raw.h"
+#include "svc_internal.h"
+
 #ifndef UDPMSGSIZE
 #define UDPMSGSIZE 8800
 #endif
@@ -52,47 +55,69 @@
 /*
  * This is the "network" that we will be moving data over
  */
-static struct svc_raw_private {
-	char *raw_buf;		/* should be shared with the cl handle */
-	SVCXPRT server;
-	XDR xdr_stream;
-	char verf_body[MAX_AUTH_BYTES];
-} *svc_raw_private;
+static struct rpc_raw_xprt *svc_raw_private;
 
 extern mutex_t svcraw_lock;
 
 static void svc_raw_ops(SVCXPRT *);
 
-char *__rpc_rawcombuf = NULL;
+/*static*/ void
+svc_raw_xprt_free(struct rpc_raw_xprt *srp)
+{
+	XDR_DESTROY(srp->raw_dr.ioq.xdrs);
+	rpc_dplx_rec_destroy(&srp->raw_dr);
+	mem_free(srp, sizeof(struct rpc_raw_xprt) + srp->raw_dr.maxrec);
+}
+
+static struct rpc_raw_xprt *
+svc_raw_xprt_zalloc(size_t sz)
+{
+	struct rpc_raw_xprt *srp = mem_zalloc(sizeof(struct rpc_raw_xprt) + sz);
+
+	/* Init SVCXPRT locks, etc */
+	rpc_dplx_rec_init(&srp->raw_dr);
+	xdr_ioq_setup(&srp->raw_dr.ioq);
+	return (srp);
+}
 
 SVCXPRT *
 svc_raw_ncreate(void)
 {
-	struct svc_raw_private *srp;
+	SVCXPRT *xprt;
+	struct rpc_raw_xprt *srp;
 
 	/* VARIABLES PROTECTED BY svcraw_lock: svc_raw_private, srp */
 	mutex_lock(&svcraw_lock);
 	srp = svc_raw_private;
 	if (srp == NULL) {
-		srp = (struct svc_raw_private *)mem_alloc(sizeof(*srp));
-		if (__rpc_rawcombuf == NULL)
-			__rpc_rawcombuf = mem_alloc(UDPMSGSIZE * sizeof(char));
-		srp->raw_buf = __rpc_rawcombuf;	/* Share it with the client */
+		srp = svc_raw_xprt_zalloc(UDPMSGSIZE);
+		srp->raw_dr.xprt.xp_fd = FD_SETSIZE;
 		svc_raw_private = srp;
 	}
-	srp->server.xp_fd = FD_SETSIZE;
-	srp->server.xp_port = 0;
-	srp->server.xp_p3 = NULL;
-	svc_raw_ops(&srp->server);
+	xprt = &srp->raw_dr.xprt;
+
+	svc_raw_ops(xprt);
 /* XXX check and or fixme */
 #if 0
 	srp->server.xp_verf.oa_base = srp->verf_body;
 #endif
-	xdrmem_create(&srp->xdr_stream, srp->raw_buf, UDPMSGSIZE, XDR_DECODE);
-	xprt_register(&srp->server);
-	mutex_unlock(&svcraw_lock);
+	srp->raw_dr.sendsz =
+	srp->raw_dr.recvsz =
+	srp->raw_dr.maxrec = UDPMSGSIZE;
 
-	return (&srp->server);
+	xdrmem_create(srp->raw_dr.ioq.xdrs, srp->raw_buf, UDPMSGSIZE,
+		      XDR_DECODE);
+
+	svc_rqst_evchan_reg(__svc_params->ev_u.evchan.id, xprt,
+			    SVC_RQST_FLAG_CHAN_AFFINITY);
+	mutex_unlock(&svcraw_lock);
+	XPRT_TRACE(xprt, __func__, __func__, __LINE__);
+
+#if defined(HAVE_BLKIN)
+	__rpc_set_blkin_endpoint(xprt, "svc_raw");
+#endif
+
+	return (xprt);
 }
 
  /*ARGSUSED*/
@@ -103,92 +128,60 @@ svc_raw_stat(SVCXPRT *xprt)
 }
 
  /*ARGSUSED*/
-static bool
-svc_raw_recv(SVCXPRT *xprt, struct svc_req *req)
+static enum xprt_stat
+svc_raw_recv(SVCXPRT *xprt)
 {
-	struct rpc_msg *msg = req->rq_msg;
-	struct svc_raw_private *srp;
-	XDR *xdrs;
+	struct rpc_raw_xprt *srp;
 
 	mutex_lock(&svcraw_lock);
 	srp = svc_raw_private;
 	if (srp == NULL) {
 		mutex_unlock(&svcraw_lock);
-		return (false);
+		return (XPRT_DIED);
 	}
 	mutex_unlock(&svcraw_lock);
 
-	xdrs = &srp->xdr_stream;
+	return (__svc_params->request_cb(xprt, srp->raw_dr.ioq.xdrs));
+}
+
+static enum xprt_stat
+svc_raw_decode(struct svc_req *req)
+{
+	XDR *xdrs = req->rq_xdrs;
+
 	xdrs->x_op = XDR_DECODE;
 	(void)XDR_SETPOS(xdrs, 0);
-	if (!xdr_callmsg(xdrs, msg))
-		return (false);
+	rpc_msg_init(&req->rq_msg);
 
-	return (true);
+	if (!xdr_callmsg(xdrs, &req->rq_msg))
+		return (XPRT_DIED);
+
+	return (req->rq_xprt->xp_dispatch.process_cb(req));
 }
 
  /*ARGSUSED*/
-static bool
-svc_raw_reply(SVCXPRT *xprt, struct svc_req *req, struct rpc_msg *msg)
+static enum xprt_stat
+svc_raw_reply(struct svc_req *req)
 {
-	struct svc_raw_private *srp;
+	struct rpc_raw_xprt *srp;
 	XDR *xdrs;
 
 	mutex_lock(&svcraw_lock);
 	srp = svc_raw_private;
 	if (srp == NULL) {
 		mutex_unlock(&svcraw_lock);
-		return (false);
+		return (XPRT_DIED);
 	}
 	mutex_unlock(&svcraw_lock);
 
-	xdrs = &srp->xdr_stream;
+	xdrs = srp->raw_dr.ioq.xdrs;
 	xdrs->x_op = XDR_ENCODE;
 	(void)XDR_SETPOS(xdrs, 0);
-	if (!xdr_replymsg(xdrs, msg))
-		return (false);
+	if (!xdr_replymsg(xdrs, &req->rq_msg))
+		return (XPRT_DIED);
 	(void)XDR_GETPOS(xdrs);	/* called just for overhead */
 
-	return (true);
-}
-
- /*ARGSUSED*/
-static bool
-svc_raw_freeargs(SVCXPRT *xprt, struct svc_req *req, xdrproc_t xdr_args,
-		 void *args_ptr)
-{
-	struct svc_raw_private *srp;
-	XDR *xdrs;
-
-	mutex_lock(&svcraw_lock);
-	srp = svc_raw_private;
-	if (srp == NULL) {
-		mutex_unlock(&svcraw_lock);
-		return (false);
-	}
-	mutex_unlock(&svcraw_lock);
-
-	xdrs = &srp->xdr_stream;
-	xdrs->x_op = XDR_FREE;
-
-	return (*xdr_args) (xdrs, args_ptr);
-}
-
- /*ARGSUSED*/
-static bool
-svc_raw_getargs(SVCXPRT *xprt, struct svc_req *req, xdrproc_t xdr_args,
-		void *args_ptr, void *u_data)
-{
-	struct svc_raw_private *srp;
-
-	mutex_lock(&svcraw_lock);
-	srp = svc_raw_private;
-	if (srp == NULL) {
-		mutex_unlock(&svcraw_lock);
-		return (false);
-	}
-	mutex_unlock(&svcraw_lock);
-	return (*xdr_args) (&srp->xdr_stream, args_ptr);
+	return (XPRT_IDLE);
 }
 
  /*ARGSUSED*/
@@ -216,16 +209,11 @@ svc_raw_ops(SVCXPRT *xprt)
 	if (ops.xp_recv == NULL) {
 		ops.xp_recv = svc_raw_recv;
 		ops.xp_stat = svc_raw_stat;
-		ops.xp_getargs = svc_raw_getargs;
+		ops.xp_decode = svc_raw_decode;
 		ops.xp_reply = svc_raw_reply;
-		ops.xp_freeargs = svc_raw_freeargs;
+		ops.xp_checksum = NULL;		/* optional */
 		ops.xp_destroy = svc_raw_destroy;
 		ops.xp_control = svc_raw_control;
-		ops.xp_lock = NULL;	/* no default */
-		ops.xp_unlock = NULL;	/* no default */
-		ops.xp_getreq = svc_getreq_default;
-		ops.xp_dispatch = svc_dispatch_default;
-		ops.xp_recv_user_data = NULL;	/* no default */
 		ops.xp_free_user_data = NULL;	/* no default */
 	}
 	xprt->xp_ops = &ops;
