@@ -54,9 +54,7 @@
 
 #include "rpc_com.h"
 #include "misc/city.h"
-#include "rpc/xdr_inrec.h"
 #include "svc_internal.h"
-#include "clnt_internal.h"
 #include "svc_xprt.h"
 #include "rpc_rdma.h"
 #include <rpc/svc_rqst.h>
@@ -76,68 +74,67 @@ struct svc_rdma_xdr {
 
 #define	sm_data(xprt)	((struct svc_rdma_xdr *)(xprt->xp_p2))
 
-extern struct svc_params __svc_params[1];
-
 static void svc_rdma_ops(SVCXPRT *);
 
 /*
- * svc_rdma_ncreate: waits for connection request and returns transport
+ * svc_rdma_rendezvous: waits for connection request
  */
-SVCXPRT *
-svc_rdma_ncreate(void *arg, const u_int sendsize, const u_int recvsize,
-		 const u_int flags)
+enum xprt_stat
+svc_rdma_rendezvous(SVCXPRT *s_xprt)
 {
 	struct svc_rdma_xdr *sm;
 	struct sockaddr_storage *ss;
-	RDMAXPRT *l_xprt = arg;
-	RDMAXPRT *xprt = rpc_rdma_accept_wait(l_xprt, l_xprt->xa->timeout);
+	RDMAXPRT *l_xprt = RDMA_DR(REC_XPRT(s_xprt));
+	RDMAXPRT *r_xprt = rpc_rdma_accept_wait(l_xprt,
+						__svc_params->idle_timeout);
 
-	if (!xprt) {
+	if (!r_xprt) {
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s:%u ERROR (return)",
 			__func__, __LINE__);
-		return (NULL);
+		return (XPRT_DIED);
 	}
 
 	sm = mem_zalloc(sizeof (*sm));
-	if (sm == NULL)
-		goto freedata;
 
-	sm->sm_xdrs.x_lib[1] = xprt;
-	xprt->xprt.xp_p2 = sm;
+	sm->sm_xdrs.x_lib[1] = r_xprt;
+	r_xprt->sm_dr.xprt.xp_p2 = sm;
 
-	xprt->xprt.xp_flags = flags;
+	r_xprt->sm_dr.xprt.xp_flags = SVC_XPRT_FLAG_CLOSE
+				    | SVC_XPRT_FLAG_INITIAL
+				    | SVC_XPRT_FLAG_INITIALIZED;
 	/* fixme: put something here, but make it not work on fd operations. */
-	xprt->xprt.xp_fd = -1;
+	r_xprt->sm_dr.xprt.xp_fd = -1;
 
-	ss = (struct sockaddr_storage *)rdma_get_local_addr(xprt->cm_id);
-	__rpc_set_address(&xprt->xprt.xp_local, ss, 0);
+	ss = (struct sockaddr_storage *)rdma_get_local_addr(r_xprt->cm_id);
+	__rpc_address_setup(&r_xprt->sm_dr.xprt.xp_local);
+	memcpy(&r_xprt->sm_dr.xprt.xp_local.nb.buf, ss,
+		r_xprt->sm_dr.xprt.xp_local.nb.len);
 
-	ss = (struct sockaddr_storage *)rdma_get_peer_addr(xprt->cm_id);
-	__rpc_set_address(&xprt->xprt.xp_remote, ss, 0);
+	ss = (struct sockaddr_storage *)rdma_get_peer_addr(r_xprt->cm_id);
+	__rpc_address_setup(&r_xprt->sm_dr.xprt.xp_remote);
+	memcpy(&r_xprt->sm_dr.xprt.xp_remote.nb.buf, ss,
+		r_xprt->sm_dr.xprt.xp_remote.nb.len);
 
-	svc_rdma_ops(&xprt->xprt);
+	svc_rdma_ops(&r_xprt->sm_dr.xprt);
 
-	if (xdr_rdma_create(&sm->sm_xdrs, xprt, sendsize, recvsize, flags)) {
+	if (xdr_rdma_create(&sm->sm_xdrs, r_xprt)) {
 		goto freedata;
 	}
 
-	if (rpc_rdma_accept_finalize(xprt)) {
+	if (rpc_rdma_accept_finalize(r_xprt)) {
 		goto freedata;
 	}
 
-	return (&xprt->xprt);
+	SVC_REF(s_xprt, SVC_REF_FLAG_NONE);
+	r_xprt->sm_dr.xprt.xp_parent = s_xprt;
+	return (s_xprt->xp_dispatch.rendezvous_cb(&r_xprt->sm_dr.xprt));
 
 freedata:
-	__warnx(TIRPC_DEBUG_FLAG_ERROR,
-		"%s() out of memory",
-		__func__);
-	if (sm) {
-		(void) mem_free(sm, sizeof (*sm));
-		xprt->xprt.xp_p2 = NULL;
-	}
-	xprt_unregister(&xprt->xprt);
-	return (NULL);
+	mem_free(sm, sizeof (*sm));
+	r_xprt->sm_dr.xprt.xp_p2 = NULL;
+	svc_rqst_xprt_unregister(&r_xprt->sm_dr.xprt);
+	return (XPRT_DIED);
 }
 
 /*ARGSUSED*/
@@ -145,7 +142,7 @@ static enum xprt_stat
 svc_rdma_stat(SVCXPRT *xprt)
 {
 	/* note: RDMAXPRT is this xprt! */
-	switch(((RDMAXPRT *)xprt)->state) {
+	switch((RDMA_DR(REC_XPRT(xprt)))->state) {
 		case RDMAXS_LISTENING:
 		case RDMAXS_CONNECTED:
 			return (XPRT_IDLE);
@@ -155,130 +152,91 @@ svc_rdma_stat(SVCXPRT *xprt)
 	}
 }
 
-static bool
-svc_rdma_recv(SVCXPRT *xprt, struct svc_req *req)
+static enum xprt_stat
+svc_rdma_decode(struct svc_req *req)
 {
-	struct rpc_rdma_cbc *cbc = req->rq_context;
-	XDR *xdrs = cbc->holdq.xdrs;
+	XDR *xdrs = req->rq_xdrs;
+	struct xdr_ioq *holdq = XIOQ(xdrs);
+	struct rpc_rdma_cbc *cbc =
+		opr_containerof(holdq, struct rpc_rdma_cbc, holdq);
 
 	__warnx(TIRPC_DEBUG_FLAG_SVC_RDMA,
-		"%s() xprt %p cbc %p incoming xdr %p\n",
-		__func__, xprt, cbc, xdrs);
-
-	req->rq_msg = alloc_rpc_msg();
+		"%s() xprt %p req %p cbc %p incoming xdr %p\n",
+		__func__, req->rq_xprt, req, cbc, xdrs);
 
 	if (!xdr_rdma_svc_recv(cbc, 0)){
-		__warnx(TIRPC_DEBUG_FLAG_SVC_RDMA,
-			"%s: xdr_rdma_svc_recv failed (will set dead)",
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: xdr_rdma_svc_recv failed",
 			__func__);
-		return (FALSE);
+		return (XPRT_DIED);
 	}
-	xdrs->x_op = XDR_DECODE;
 
+	xdrs->x_op = XDR_DECODE;
 	/* No need, already positioned to beginning ...
 	XDR_SETPOS(xdrs, 0);
 	 */
-	if (!xdr_dplx_decode(xdrs, req->rq_msg)) {
-		__warnx(TIRPC_DEBUG_FLAG_SVC_RDMA,
-			"%s: xdr_dplx_decode failed (will set dead)",
-			__func__);
-		return (FALSE);
-	}
+	rpc_msg_init(&req->rq_msg);
 
-	req->rq_xprt = xprt;
-	req->rq_prog = req->rq_msg->rm_call.cb_prog;
-	req->rq_vers = req->rq_msg->rm_call.cb_vers;
-	req->rq_proc = req->rq_msg->rm_call.cb_proc;
-	req->rq_xid = req->rq_msg->rm_xid;
-	req->rq_clntcred = req->rq_msg->rq_cred_body;
+	if (!xdr_dplx_decode(xdrs, &req->rq_msg)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: xdr_dplx_decode failed",
+			__func__);
+		return (XPRT_DIED);
+	}
 
 	/* the checksum */
 	req->rq_cksum = 0;
 
-	return (TRUE);
+	return (req->rq_xprt->xp_dispatch.process_cb(req));
 }
 
-static bool
-svc_rdma_reply(SVCXPRT *xprt, struct svc_req *req, struct rpc_msg *msg)
+static enum xprt_stat
+svc_rdma_reply(struct svc_req *req)
 {
-	struct rpc_rdma_cbc *cbc = req->rq_context;
-	XDR *xdrs = cbc->holdq.xdrs;
-	xdrproc_t proc;
-	void *where;
-	bool has_args;
+	XDR *xdrs = req->rq_xdrs;
+	struct xdr_ioq *holdq = XIOQ(xdrs);
+	struct rpc_rdma_cbc *cbc =
+		opr_containerof(holdq, struct rpc_rdma_cbc, holdq);
 
 	__warnx(TIRPC_DEBUG_FLAG_SVC_RDMA,
-		"%s() xprt %p cbc %p outgoing xdr %p\n",
-		__func__, xprt, cbc, xdrs);
-
-	if (msg->rm_reply.rp_stat == MSG_ACCEPTED
-	 && msg->rm_reply.rp_acpt.ar_stat == SUCCESS) {
-	    has_args = TRUE;
-	    proc = msg->acpted_rply.ar_results.proc;
-	    where = msg->acpted_rply.ar_results.where;
-	    msg->acpted_rply.ar_results.proc = (xdrproc_t)xdr_void;
-	    msg->acpted_rply.ar_results.where = NULL;
-	} else {
-	    has_args = FALSE;
-	    proc = NULL;
-	    where = NULL;
-	}
+		"%s() xprt %p req %p cbc %p outgoing xdr %p\n",
+		__func__, req->rq_xprt, req, cbc, xdrs);
 
 	if (!xdr_rdma_svc_reply(cbc, 0)){
-		__warnx(TIRPC_DEBUG_FLAG_SVC_RDMA,
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s: xdr_rdma_svc_reply failed (will set dead)",
 			__func__);
-		return (FALSE);
+		return (XPRT_DIED);
 	}
 	xdrs->x_op = XDR_ENCODE;
 
-	if (!xdr_reply_encode(xdrs, msg)) {
-		__warnx(TIRPC_DEBUG_FLAG_SVC_RDMA,
+	if (!xdr_reply_encode(xdrs, &req->rq_msg)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s: xdr_reply_encode failed (will set dead)",
 			__func__);
-		return (FALSE);
+		return (XPRT_DIED);
 	}
 	xdr_tail_update(xdrs);
 
-	if (has_args && req->rq_auth
-	  && !SVCAUTH_WRAP(req->rq_auth, req, xdrs, proc, where)) {
-		__warnx(TIRPC_DEBUG_FLAG_SVC_RDMA,
+	if (req->rq_msg.rm_reply.rp_stat == MSG_ACCEPTED
+	 && req->rq_msg.rm_reply.rp_acpt.ar_stat == SUCCESS
+	 && req->rq_auth
+	 && !SVCAUTH_WRAP(req, xdrs)) {
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s: SVCAUTH_WRAP failed (will set dead)",
 			__func__);
-		return (FALSE);
+		return (XPRT_DIED);
 	}
 	xdr_tail_update(xdrs);
 
-	return xdr_rdma_svc_flushout(cbc);
-}
+	if (!xdr_rdma_svc_flushout(cbc)){
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
+			"%s: flushout failed (will set dead)",
+			__func__);
+		return (XPRT_DIED);
+	}
 
-static bool
-svc_rdma_freeargs(SVCXPRT *xprt, struct svc_req *req, xdrproc_t xdr_args,
-		  void *args_ptr)
-{
-	struct rpc_rdma_cbc *cbc = req->rq_context;
-	XDR *xdrs = cbc->holdq.xdrs;
-
-	xdrs->x_op = XDR_FREE;
-	return (*xdr_args)(xdrs, args_ptr);
-}
-
-static bool
-svc_rdma_getargs(SVCXPRT *xprt, struct svc_req *req, xdrproc_t xdr_args,
-		 void *args_ptr, void *u_data)
-{
-	struct rpc_rdma_cbc *cbc = req->rq_context;
-	XDR *xdrs = cbc->holdq.xdrs;
-	bool rslt;
-
-	/* threads u_data for advanced decoders*/
-	xdrs->x_public = u_data;
-
-	rslt = SVCAUTH_UNWRAP(req->rq_auth, req, xdrs, xdr_args, args_ptr);
-	if (!rslt)
-		svc_rdma_freeargs(xprt, req, xdr_args, args_ptr);
-
-	return (rslt);
+	return (XPRT_IDLE);
 }
 
 static void
@@ -292,29 +250,19 @@ svc_rdma_destroy(SVCXPRT *xprt, u_int flags, const char *tag, const int line)
 		__func__, xprt, xprt->xp_refs, tag, line);
 
 	xdr_rdma_destroy(&(sm->sm_xdrs));
-	(void) mem_free(sm, sizeof (*sm));
+	mem_free(sm, sizeof (*sm));
 
 	if (xprt->xp_ops->xp_free_user_data) {
 		/* call free hook */
 		xprt->xp_ops->xp_free_user_data(xprt);
 	}
-
+	if (xprt->xp_parent)
+		SVC_RELEASE(xprt->xp_parent, SVC_RELEASE_FLAG_NONE);
 	rpc_rdma_destroy(xprt);
 }
 
 extern mutex_t ops_lock;
 
-static void
-svc_rdma_lock(SVCXPRT *xprt, uint32_t flags, const char *file, int line)
-{
-/* pretend we lock for now */
-}
-
-static void
-svc_rdma_unlock(SVCXPRT *xprt, uint32_t flags, const char *file, int line)
-{
-}
-
 static bool
 /*ARGSUSED*/
 svc_rdma_control(SVCXPRT *xprt, const u_int rq, void *in)
@@ -326,44 +274,14 @@ svc_rdma_control(SVCXPRT *xprt, const u_int rq, void *in)
 	case SVCSET_XP_FLAGS:
 	    xprt->xp_flags = *(u_int *)in;
 	    break;
-	case SVCGET_XP_RECV:
-	    mutex_lock(&ops_lock);
-	    *(xp_recv_t *)in = xprt->xp_ops->xp_recv;
-	    mutex_unlock(&ops_lock);
-	    break;
-	case SVCSET_XP_RECV:
-	    mutex_lock(&ops_lock);
-	    xprt->xp_ops->xp_recv = *(xp_recv_t)in;
-	    mutex_unlock(&ops_lock);
-	    break;
-	case SVCGET_XP_GETREQ:
-	    mutex_lock(&ops_lock);
-	    *(xp_getreq_t *)in = xprt->xp_ops->xp_getreq;
-	    mutex_unlock(&ops_lock);
-	    break;
-	case SVCSET_XP_GETREQ:
-	    mutex_lock(&ops_lock);
-	    xprt->xp_ops->xp_getreq = *(xp_getreq_t)in;
-	    mutex_unlock(&ops_lock);
-	    break;
-	case SVCGET_XP_DISPATCH:
-	    mutex_lock(&ops_lock);
-	    *(xp_dispatch_t *)in = xprt->xp_ops->xp_dispatch;
-	    mutex_unlock(&ops_lock);
-	    break;
-	case SVCSET_XP_DISPATCH:
-	    mutex_lock(&ops_lock);
-	    xprt->xp_ops->xp_dispatch = *(xp_dispatch_t)in;
-	    mutex_unlock(&ops_lock);
-	    break;
 	case SVCGET_XP_FREE_USER_DATA:
 	    mutex_lock(&ops_lock);
-	    *(xp_free_user_data_t *)in = xprt->xp_ops->xp_free_user_data;
+	    *(svc_xprt_fun_t *)in = xprt->xp_ops->xp_free_user_data;
 	    mutex_unlock(&ops_lock);
 	    break;
 	case SVCSET_XP_FREE_USER_DATA:
 	    mutex_lock(&ops_lock);
-	    xprt->xp_ops->xp_free_user_data = *(xp_free_user_data_t)in;
+	    xprt->xp_ops->xp_free_user_data = *(svc_xprt_fun_t)in;
 	    mutex_unlock(&ops_lock);
 	    break;
 	default:
@@ -385,18 +303,13 @@ svc_rdma_ops(SVCXPRT *xprt)
 	xprt->xp_type = XPRT_RDMA;
 
 	if (ops.xp_recv == NULL) {
-		ops.xp_recv = svc_rdma_recv;
+		ops.xp_recv = NULL;
 		ops.xp_stat = svc_rdma_stat;
-		ops.xp_getargs = svc_rdma_getargs;
+		ops.xp_decode = svc_rdma_decode;
 		ops.xp_reply = svc_rdma_reply;
-		ops.xp_freeargs = svc_rdma_freeargs;
+		ops.xp_checksum = NULL;		/* not used */
 		ops.xp_destroy = svc_rdma_destroy,
 		ops.xp_control = svc_rdma_control;
-		ops.xp_lock = svc_rdma_lock;
-		ops.xp_unlock = svc_rdma_unlock;
-		ops.xp_getreq = svc_getreq_default;
-		ops.xp_dispatch = svc_dispatch_default;
-		ops.xp_recv_user_data = NULL;	/* no default */
 		ops.xp_free_user_data = NULL;	/* no default */
 	}
 	xprt->xp_ops = &ops;
