@@ -48,6 +48,7 @@
 #include <sys/types.h>
 #include <rpc/rpc.h>
 #include <rpc/nettype.h>
+#include <rpc/svc_auth.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -55,7 +56,7 @@
 
 #include "rpc_com.h"
 
-static void universal(struct svc_req *, SVCXPRT *);
+static void universal(struct svc_req *);
 
 static struct proglst {
 	char *(*p_progname) (char *);
@@ -98,7 +99,7 @@ rpc_reg(rpcprog_t prognum, rpcvers_t versnum, rpcproc_t procnum,
 	extern mutex_t proglst_lock;
 
 	if (procnum == NULLPROC) {
-		__warnx(TIRPC_DEBUG_FLAG_SVC,
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s can't reassign procedure number %u", rpc_reg_msg,
 			NULLPROC);
 		return (-1);
@@ -108,7 +109,7 @@ rpc_reg(rpcprog_t prognum, rpcvers_t versnum, rpcproc_t procnum,
 		nettype = "netpath";	/* The default behavior */
 	handle = __rpc_setconf(nettype);
 	if (!handle) {
-		__warnx(TIRPC_DEBUG_FLAG_SVC, rpc_reg_err, rpc_reg_msg,
+		__warnx(TIRPC_DEBUG_FLAG_ERROR, rpc_reg_err, rpc_reg_msg,
 			__reg_err1);
 		return (-1);
 	}
@@ -143,14 +144,14 @@ rpc_reg(rpcprog_t prognum, rpcvers_t versnum, rpcproc_t procnum,
 			if (svcxprt == NULL)
 				continue;
 			if (!__rpc_fd2sockinfo(svcxprt->xp_fd, &si)) {
-				__warnx(TIRPC_DEBUG_FLAG_SVC, rpc_reg_err,
+				__warnx(TIRPC_DEBUG_FLAG_ERROR, rpc_reg_err,
 					rpc_reg_msg, __reg_err2);
 				SVC_DESTROY(svcxprt);
 				continue;
 			}
 			recvsz = __rpc_get_t_size(si.si_af, si.si_proto, 0);
 			if (recvsz == 0) {
-				__warnx(TIRPC_DEBUG_FLAG_SVC, rpc_reg_err,
+				__warnx(TIRPC_DEBUG_FLAG_ERROR, rpc_reg_err,
 					rpc_reg_msg, __reg_err3);
 				SVC_DESTROY(svcxprt);
 				continue;
@@ -176,7 +177,7 @@ rpc_reg(rpcprog_t prognum, rpcvers_t versnum, rpcproc_t procnum,
 		}
 
 		if (!svc_reg(svcxprt, prognum, versnum, universal, nconf)) {
-			__warnx(TIRPC_DEBUG_FLAG_SVC,
+			__warnx(TIRPC_DEBUG_FLAG_ERROR,
 				"%s couldn't register prog %u vers %u for %s",
 				rpc_reg_msg, (unsigned)prognum,
 				(unsigned)versnum, netid);
@@ -207,7 +208,7 @@ rpc_reg(rpcprog_t prognum, rpcvers_t versnum, rpcproc_t procnum,
 	mutex_unlock(&proglst_lock);
 
 	if (done == false) {
-		__warnx(TIRPC_DEBUG_FLAG_SVC,
+		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s cant find suitable transport for %s", rpc_reg_msg,
 			nettype);
 		return (-1);
@@ -216,12 +217,12 @@ rpc_reg(rpcprog_t prognum, rpcvers_t versnum, rpcproc_t procnum,
 }
 
 /*
- * The universal handler for the services registered using registerrpc.
+ * The universal handler for the services registered using svc_reg.
  * It handles both the connectionless and the connection oriented cases.
  */
 
 static void
-universal(struct svc_req *req, SVCXPRT *transp)
+universal(struct svc_req *req)
 {
 	rpcprog_t prog;
 	rpcvers_t vers;
@@ -234,21 +235,22 @@ universal(struct svc_req *req, SVCXPRT *transp)
 	/*
 	 * enforce "procnum 0 is echo" convention
 	 */
-	if (req->rq_proc == NULLPROC) {
-		if (svc_sendreply(transp, req, (xdrproc_t) xdr_void, NULL) ==
-		    false) {
-			__warnx(TIRPC_DEBUG_FLAG_SVC, "svc_sendreply failed");
+	if (req->rq_msg.cb_proc == NULLPROC) {
+		req->rq_msg.RPCM_ack.ar_results.where = NULL;
+		req->rq_msg.RPCM_ack.ar_results.proc = (xdrproc_t) xdr_void;
+		if (svc_sendreply(req) >= XPRT_DIED) {
+			__warnx(TIRPC_DEBUG_FLAG_ERROR, "svc_sendreply failed");
 		}
 		return;
 	}
-	prog = req->rq_prog;
-	vers = req->rq_vers;
-	proc = req->rq_proc;
+	prog = req->rq_msg.cb_prog;
+	vers = req->rq_msg.cb_vers;
+	proc = req->rq_msg.cb_proc;
 	mutex_lock(&proglst_lock);
 	for (pl = proglst; pl; pl = pl->p_nxt)
 		if (pl->p_prognum == prog && pl->p_procnum == proc
 		    && pl->p_versnum == vers
-		    && (strcmp(pl->p_netid, transp->xp_netid) == 0)) {
+		    && (strcmp(pl->p_netid, req->rq_xprt->xp_netid) == 0)) {
 			/* decode arguments into a CLEAN buffer */
 			xdrbuf = pl->p_xdrbuf;
 			/* Zero the arguments: reqd ! */
@@ -258,35 +260,42 @@ universal(struct svc_req *req, SVCXPRT *transp)
 			 * for the arguments; if not then the program
 			 * may bomb. BEWARE!
 			 */
-			if (!svc_getargs
-			    (transp, req, pl->p_inproc, xdrbuf, NULL)) {
-				svcerr_decode(transp, req);
+			req->rq_msg.rm_xdr.where = xdrbuf;
+			req->rq_msg.rm_xdr.proc = pl->p_inproc;
+			if (!SVCAUTH_CHECKSUM(req)) {
+				__warnx(TIRPC_DEBUG_FLAG_ERROR,
+					"rpc: SVCAUTH_CHECKSUM failed prog %u vers %u",
+					(unsigned)prog, (unsigned)vers);
+				xdr_free(pl->p_inproc, xdrbuf);
+				svcerr_decode(req);
 				mutex_unlock(&proglst_lock);
 				return;
 			}
+
 			outdata = (*(pl->p_progname)) (xdrbuf);
 			if (outdata == NULL
 			    && pl->p_outproc != (xdrproc_t) xdr_void) {
 				/* there was an error */
+				xdr_free(pl->p_inproc, xdrbuf);
 				mutex_unlock(&proglst_lock);
 				return;
 			}
-			if (!svc_sendreply(transp, req, pl->p_outproc,
-					   outdata)) {
-				__warnx(TIRPC_DEBUG_FLAG_SVC,
-					"rpc: rpc_reg trouble replying to prog %u vers %u",
+
+			req->rq_msg.RPCM_ack.ar_results.where = outdata;
+			req->rq_msg.RPCM_ack.ar_results.proc = pl->p_outproc;
+			if (svc_sendreply(req) >= XPRT_DIED) {
+				__warnx(TIRPC_DEBUG_FLAG_ERROR,
+					"rpc: svc_sendreply failed prog %u vers %u",
 					(unsigned)prog, (unsigned)vers);
-				mutex_unlock(&proglst_lock);
-				return;
 			}
 			/* free the decoded arguments */
-			(void)svc_freeargs(transp, req, pl->p_inproc, xdrbuf);
+			xdr_free(pl->p_inproc, xdrbuf);
 			mutex_unlock(&proglst_lock);
 			return;
 		}
 	mutex_unlock(&proglst_lock);
 	/* This should never happen */
-	__warnx(TIRPC_DEBUG_FLAG_SVC,
+	__warnx(TIRPC_DEBUG_FLAG_ERROR,
 		"rpc: rpc_reg: never registered prog %u vers %u",
 		(unsigned)prog, (unsigned)vers);
 	return;
