@@ -1,5 +1,7 @@
 /*
  * Copyright (c) 2012 Linux Box Corporation.
+ * Copyright (c) 2013-2015 CohortFS, LLC.
+ * Copyright (c) 2017 Red Hat, Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -36,18 +38,29 @@
 #include <rpc/types.h>
 #include <misc/portable.h>
 #include <rpc/rpc.h>
-#ifdef PORTMAP
-#include <rpc/pmap_clnt.h>
-#endif				/* PORTMAP */
 #include "rpc_com.h"
-#include <rpc/svc.h>
-#include <misc/rbtree_x.h>
-#include <reentrant.h>
-#include "clnt_internal.h"
 #include "rpc_dplx_internal.h"
 #include "svc_xprt.h"
 
-#define SVC_XPRT_PARTITIONS 7
+/**
+ * @file svc_xprt.c
+ * @contributeur William Allen Simpson <bill@cohortfs.com>
+ * @brief Service transports package
+ *
+ * @section DESCRIPTION
+ *
+ * Maintains a tree of all extant transports by fd.
+ *
+ * Each SVCXPRT has its own instance, however, so operations to
+ * close and delete (for example) given an existing xprt handle
+ * are O(1) without any ordered or hashed representation.
+ *
+ * @note currently static sizes
+ *	partitions should be largish prime, relative to connections.
+ *	no cache slots, as rpc_dplx_rec has fd_node for direct access.
+ */
+
+#define SVC_XPRT_PARTITIONS 193
 
 static bool initialized;
 
@@ -59,7 +72,7 @@ struct svc_xprt_fd {
 static struct svc_xprt_fd svc_xprt_fd = {
 	MUTEX_INITIALIZER /* svc_xprt_lock */ ,
 	{
-	 0,			/* npart */
+	 SVC_XPRT_PARTITIONS,	/* npart */
 	 RBT_X_FLAG_NONE,	/* flags */
 	 0,			/* cachesz */
 	 NULL			/* tree */
@@ -70,21 +83,22 @@ static inline int
 svc_xprt_fd_cmpf(const struct opr_rbtree_node *lhs,
 		 const struct opr_rbtree_node *rhs)
 {
-	SVCXPRT *lk, *rk;
+	struct rpc_dplx_rec *lk, *rk;
 
-	lk = opr_containerof(lhs, struct rpc_svcxprt, xp_fd_node);
-	rk = opr_containerof(rhs, struct rpc_svcxprt, xp_fd_node);
+	lk = opr_containerof(lhs, struct rpc_dplx_rec, fd_node);
+	rk = opr_containerof(rhs, struct rpc_dplx_rec, fd_node);
 
-	if (lk->xp_fd < rk->xp_fd)
+	if (lk->xprt.xp_fd < rk->xprt.xp_fd)
 		return (-1);
 
-	if (lk->xp_fd == rk->xp_fd)
+	if (lk->xprt.xp_fd == rk->xprt.xp_fd)
 		return (0);
 
 	return (1);
 }
 
-void svc_xprt_init()
+int
+svc_xprt_init(void)
 {
 	int code = 0;
 
@@ -106,116 +120,133 @@ void svc_xprt_init()
 
  unlock:
 	mutex_unlock(&svc_xprt_fd.lock);
+	return (code);
 }
 
-#define cond_init_svc_xprt() { \
-		do { \
-			if (!initialized) \
-				svc_xprt_init(); \
-		} while (0); \
-	}
-
-SVCXPRT *
-svc_xprt_get(int fd)
+static inline bool
+svc_xprt_init_failure(void)
 {
-	struct rpc_svcxprt sk;
+	if (initialized)
+		return (false);
+	return (svc_xprt_init() != 0);
+}
+
+/*
+ * On success, returns with RPC_DPLX_FLAG_LOCKED
+ */
+SVCXPRT *
+svc_xprt_lookup(int fd, svc_xprt_setup_t setup)
+{
+	struct rpc_dplx_rec sk;
+	struct rpc_dplx_rec *rec;
 	struct rbtree_x_part *t;
 	struct opr_rbtree_node *nv;
-	SVCXPRT *srec = NULL;
+	SVCXPRT *xprt = NULL;
+	uint16_t xp_flags;
 
-	cond_init_svc_xprt();
+	if (svc_xprt_init_failure())
+		return (NULL);
 
-	sk.xp_fd = fd;
+	sk.xprt.xp_fd = fd;
 	t = rbtx_partition_of_scalar(&svc_xprt_fd.xt, fd);
 
 	rwlock_rdlock(&t->lock);
-	nv = opr_rbtree_lookup(&t->t, &sk.xp_fd_node);
+	nv = opr_rbtree_lookup(&t->t, &sk.fd_node);
+	if (!nv) {
+		rwlock_unlock(&t->lock);
+		if (!setup)
+			return (NULL);
+
+		rwlock_wrlock(&t->lock);
+		nv = opr_rbtree_lookup(&t->t, &sk.fd_node);
+		if (!nv) {
+			(*setup)(&xprt); /* zalloc, xp_refs = 1 */
+			xprt->xp_fd = fd;
+			xprt->xp_flags = SVC_XPRT_FLAG_INITIAL;
+
+			rec = REC_XPRT(xprt);
+			rpc_dplx_rli(rec);
+			if (opr_rbtree_insert(&t->t, &rec->fd_node)) {
+				/* cant happen */
+				rpc_dplx_rui(rec);
+				__warnx(TIRPC_DEBUG_FLAG_LOCK,
+					"%s: collision inserting in locked rbtree partition",
+					__func__);
+				(*setup)(&xprt);	/* free, sets NULL */
+			}
+			rwlock_unlock(&t->lock);
+			return (xprt);
+		}
+		/* raced, fallthru */
+	}
+	rec = opr_containerof(nv, struct rpc_dplx_rec, fd_node);
+	xprt = &rec->xprt;
+
+	/* lookup reference before unlock ensures shutdown cannot release */
+	SVC_REF(xprt, SVC_REF_FLAG_NONE);
 	rwlock_unlock(&t->lock);
 
-	/* XXX safe, even if tree is reorganizing */
-	if (nv)
-		srec = opr_containerof(nv, struct rpc_svcxprt, xp_fd_node);
-
-	return (srec);
-}
-
-SVCXPRT *
-svc_xprt_set(SVCXPRT *xprt, uint32_t flags)
-{
-	struct rbtree_x_part *t;
-	struct opr_rbtree_node *nv;
-	SVCXPRT *srec = NULL;
-
-	cond_init_svc_xprt();
-
-	if (!(flags & SVC_XPRT_FLAG_LOCKED))
-		mutex_lock(&xprt->xp_lock);
-
-	t = rbtx_partition_of_scalar(&svc_xprt_fd.xt, xprt->xp_fd);
-
-	rwlock_wrlock(&t->lock);
-
-	nv = opr_rbtree_insert(&t->t, &xprt->xp_fd_node);
-	if (nv) {
-		srec = opr_containerof(nv, struct rpc_svcxprt, xp_fd_node);
-		__warnx(TIRPC_DEBUG_FLAG_SVC_XPRT,
-			"%s: %p xp_refs %" PRIu32
-			" insert collision with %p xp_refs %" PRIu32,
-			__func__, srec, srec->xp_refs, xprt, xprt->xp_refs);
+	/* unlocked window here permits shutdown to destroy without release;
+	 * then duplex lock is required to match allocation return,
+	 * ensuring SVC_XPRT_FLAG_INITIAL cleared in this thread only
+	 * (obviating extra atomic fetch).
+	 */
+	rpc_dplx_rli(rec);
+	xp_flags = atomic_clear_uint16_t_bits(&xprt->xp_flags,
+					      SVC_XPRT_FLAG_INITIAL);
+	if (!(xp_flags & SVC_XPRT_FLAG_DESTROYED)) {
+		/* do not return destroyed xprts */
+		return (xprt);
 	}
 
-	rwlock_unlock(&t->lock);
-
-	if (flags & SVC_XPRT_FLAG_UNLOCK)
-		mutex_unlock(&xprt->xp_lock);
-
-	return (srec);
+	/* unlock before release permits releasing here after destroy */
+	rpc_dplx_rui(rec);
+	SVC_RELEASE(xprt, SVC_RELEASE_FLAG_NONE);
+	return (NULL);
 }
 
 /**
  * Clear an xprt
  *
  * @note Locking
- * - xprt is locked, unless SVC_XPRT_FLAG_LOCKED is passed
- * - xprt is unlocked if SVC_XPRT_FLAG_UNLOCK is passed, otherwise it is
+ * - xprt is locked
  *   returned locked
  */
 void
-svc_xprt_clear(SVCXPRT *xprt, uint32_t flags)
+svc_xprt_clear(SVCXPRT *xprt)
 {
 	struct rbtree_x_part *t;
 
-	cond_init_svc_xprt();
+	if (svc_xprt_init_failure())
+		return;
 
-	if (!(flags & SVC_XPRT_FLAG_LOCKED))
-		mutex_lock(&xprt->xp_lock);
-
-	if (opr_rbtree_node_valid(&xprt->xp_fd_node)) {
+	/* xprt lock ensures only one active thread here */
+	if (opr_rbtree_node_valid(&REC_XPRT(xprt)->fd_node)) {
 		t = rbtx_partition_of_scalar(&svc_xprt_fd.xt, xprt->xp_fd);
 
+		/* if another thread passes test during svc_xprt_shutdown(),
+		 * this lock (and generation test) prevents repeats.
+		 */
 		rwlock_wrlock(&t->lock);
-		opr_rbtree_remove(&t->t, &xprt->xp_fd_node);
+		opr_rbtree_remove(&t->t, &REC_XPRT(xprt)->fd_node);
 		rwlock_unlock(&t->lock);
 	}
-
-	if (flags & SVC_XPRT_FLAG_UNLOCK)
-		mutex_unlock(&xprt->xp_lock);
 }
 
 int
 svc_xprt_foreach(svc_xprt_each_func_t each_f, void *arg)
 {
-	struct rpc_svcxprt sk;
+	struct rpc_dplx_rec sk;
+	struct rpc_dplx_rec *rec;
 	struct rbtree_x_part *t;
 	struct opr_rbtree_node *n;
-	SVCXPRT *xprt;
 	uint64_t tgen;
-	uint32_t rflag;
 	int p_ix;
 	int x_ix;
 	int restarts;
 
-	cond_init_svc_xprt();
+	if (svc_xprt_init_failure())
+		return (-1);
 
 	/* concurrent, restartable iteration over t */
 	p_ix = 0;
@@ -238,22 +269,21 @@ svc_xprt_foreach(svc_xprt_each_func_t each_f, void *arg)
 		while (n != NULL) {
 			++x_ix;	/* diagnostic, index into logical srec
 				 * sequence */
-			xprt = opr_containerof(n, struct rpc_svcxprt, xp_fd_node);
-			sk.xp_fd = xprt->xp_fd;
+			rec = opr_containerof(n, struct rpc_dplx_rec, fd_node);
+			sk.xprt.xp_fd = rec->xprt.xp_fd;
 
 			/* call each_func with t !LOCKED */
 			rwlock_unlock(&t->lock);
 
 			/* restart if each_f disposed xprt */
-			rflag = each_f(xprt, arg);
-			if (rflag == SVC_XPRT_FOREACH_CLEAR)
+			if (each_f(&rec->xprt, arg))
 				goto restart;
 
 			/* validate */
 			rwlock_rdlock(&t->lock);
 
 			if (tgen != t->t.gen) {
-				n = opr_rbtree_lookup(&t->t, &sk.xp_fd_node);
+				n = opr_rbtree_lookup(&t->t, &sk.fd_node);
 				if (!n) {
 					/* invalidated, try harder */
 					rwlock_unlock(&t->lock);
@@ -275,7 +305,7 @@ svc_xprt_dump_xprts(const char *tag)
 {
 	struct rbtree_x_part *t = NULL;
 	struct opr_rbtree_node *n;
-	SVCXPRT *xprt;
+	struct rpc_dplx_rec *rec;
 	int p_ix;
 
 	if (!initialized)
@@ -289,10 +319,10 @@ svc_xprt_dump_xprts(const char *tag)
 			"xprts at %s: tree %d size %d", tag, p_ix, t->t.size);
 		n = opr_rbtree_first(&t->t);
 		while (n != NULL) {
-			xprt = opr_containerof(n, struct rpc_svcxprt, xp_fd_node);
+			rec = opr_containerof(n, struct rpc_dplx_rec, fd_node);
 			__warnx(TIRPC_DEBUG_FLAG_SVC_XPRT,
 				"xprts at %s: %p xp_fd %d",
-				tag, xprt, xprt->xp_fd);
+				tag, &rec->xprt, rec->xprt.xp_fd);
 			n = opr_rbtree_next(n);
 		}		/* curr partition */
 		rwlock_unlock(&t->lock);	/* t !LOCKED */
@@ -307,7 +337,7 @@ svc_xprt_shutdown()
 {
 	struct rbtree_x_part *t;
 	struct opr_rbtree_node *n;
-	SVCXPRT *xprt;
+	struct rpc_dplx_rec *rec;
 	int p_ix;
 
 	if (!initialized)
@@ -318,17 +348,18 @@ svc_xprt_shutdown()
 		t = &svc_xprt_fd.xt.tree[p_ix];
 
 		rwlock_wrlock(&t->lock);	/* t WLOCKED */
-		n = opr_rbtree_first(&t->t);
-		while (n != NULL) {
-			xprt = opr_containerof(n, struct rpc_svcxprt, xp_fd_node);
-			n = opr_rbtree_next(n);
+		while ((n = opr_rbtree_first(&t->t))) {
+			rec = opr_containerof(n, struct rpc_dplx_rec, fd_node);
 
 			/* prevent repeats, see svc_xprt_clear() */
-			mutex_lock(&xprt->xp_lock);
-			opr_rbtree_remove(&t->t, &xprt->xp_fd_node);
-			mutex_unlock(&xprt->xp_lock);
+			opr_rbtree_remove(&t->t, &rec->fd_node);
 
-			SVC_DESTROY(xprt);
+			/* fd_node is counted by initial xp_refs = 1,
+			 * SVC_DESTROY() decrements that reference.
+			 */
+			rwlock_unlock(&t->lock);
+			SVC_DESTROY(&rec->xprt);
+			rwlock_wrlock(&t->lock);
 		}		/* curr partition */
 		rwlock_unlock(&t->lock);	/* t !LOCKED */
 		rwlock_destroy(&t->lock);
@@ -343,24 +374,11 @@ svc_xprt_shutdown()
 void
 svc_xprt_trace(SVCXPRT *xprt, const char *func, const char *tag, const int line)
 {
-	struct sockaddr_storage *ss = (struct sockaddr_storage *)
-	    &(xprt->xp_remote.ss);
-	int port;
-
-	switch (ss->ss_family) {
-	case AF_INET6:
-		port = ntohs(((struct sockaddr_in6 *)ss)->sin6_port);
-		break;
-	case AF_INET:
-		port = ntohs(((struct sockaddr_in *)ss)->sin_port);
-		break;
-	default:
-		port = -1;
-		break;
-	}
 	__warnx(TIRPC_DEBUG_FLAG_REFCNT,
-		"%s() %p xp_refs %" PRId32
-		" fd %d port %d @ %s:%d",
-		func, xprt, xprt->xp_refs,
-		xprt->xp_fd, port, tag, line);
+		"%s() %p fd %d xp_refs %" PRId32
+		" af %u port %u @ %s:%d",
+		func, xprt, xprt->xp_fd, xprt->xp_refs,
+		xprt->xp_remote.ss.ss_family,
+		__rpc_address_port(&xprt->xp_remote),
+		tag, line);
 }
