@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2013 Linux Box Corporation.
+ * Copyright (c) 2013-2017 Red Hat, Inc. and/or its affiliates.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -66,7 +67,7 @@ static uint64_t next_id;
 #define free_buffer(addr,size) mem_free((addr), size)
 
 struct xdr_ioq_uv *
-xdr_ioq_uv_create(u_int size, u_int uio_flags)
+xdr_ioq_uv_create(size_t size, u_int uio_flags)
 {
 	struct xdr_ioq_uv *uv = mem_zalloc(sizeof(struct xdr_ioq_uv));
 
@@ -135,7 +136,7 @@ xdr_ioq_uv_fetch(struct xdr_ioq *xioq, struct poolq_head *ioqh,
 			xioq->xdrs[0].x_handy = saved;
 
 			/* entry was already added directly to the queue */
-			have = TAILQ_LAST(&xioq->ioq_uv.uvqh.qh, q_head);
+			have = TAILQ_LAST(&xioq->ioq_uv.uvqh.qh, poolq_head_s);
 		}
 	}
 
@@ -212,6 +213,28 @@ xdr_ioq_uv_release(struct xdr_ioq_uv *uv)
 	}
 }
 
+/*
+ * Set current read/insert or fill position.
+ */
+static inline void
+xdr_ioq_uv_reset(struct xdr_ioq *xioq, struct xdr_ioq_uv *uv)
+{
+	xioq->xdrs[0].x_data = uv->v.vio_head;
+	xioq->xdrs[0].x_base = &uv->v;
+	xioq->xdrs[0].x_v = uv->v;
+}
+
+/*
+ * Update read/insert or fill position.
+ */
+static inline void
+xdr_ioq_uv_update(struct xdr_ioq *xioq, struct xdr_ioq_uv *uv)
+{
+	xdr_ioq_uv_reset(xioq, uv);
+	(xioq->ioq_uv.pcount)++;
+	/* xioq->ioq_uv.plength is calculated in xdr_ioq_uv_advance() */
+}
+
 /*
  * Set initial read/insert or fill position.
  *
@@ -232,9 +255,7 @@ xdr_ioq_reset(struct xdr_ioq *xioq, u_int wh_pos)
 	} else {
 		uv->v.vio_head = uv->v.vio_base + wh_pos;
 	}
-	xioq->xdrs[0].x_v = uv->v;
-	xioq->xdrs[0].x_base = &uv->v;
-	xioq->xdrs[0].x_data = uv->v.vio_head;
+	xdr_ioq_uv_reset(xioq, uv);
 
 	__warnx(TIRPC_DEBUG_FLAG_XDR,
 		"%s() xioq %p head %p wh_pos %d",
@@ -250,7 +271,7 @@ xdr_ioq_setup(struct xdr_ioq *xioq)
 	assert((void *)xdrs == (void *)xioq);
 
 	TAILQ_INIT_ENTRY(&xioq->ioq_s, q);
-	xioq->ioq_s.qflags = IOQ_FLAG_NONE;
+	xioq->ioq_s.qflags = IOQ_FLAG_SEGMENT;
 
 	poolq_head_setup(&xioq->ioq_uv.uvqh);
 	pthread_cond_init(&xioq->ioq_cond, NULL);
@@ -266,12 +287,13 @@ xdr_ioq_setup(struct xdr_ioq *xioq)
 	xioq->id = atomic_inc_uint64_t(&next_id);
 }
 
-XDR *
-xdr_ioq_create(u_int min_bsize, u_int max_bsize, u_int uio_flags)
+struct xdr_ioq *
+xdr_ioq_create(size_t min_bsize, size_t max_bsize, u_int uio_flags)
 {
 	struct xdr_ioq *xioq = mem_zalloc(sizeof(struct xdr_ioq));
 
 	xdr_ioq_setup(xioq);
+	xioq->xdrs[0].x_flags |= XDR_FLAG_FREE;
 	xioq->ioq_uv.min_bsize = min_bsize;
 	xioq->ioq_uv.max_bsize = max_bsize;
 
@@ -282,7 +304,7 @@ xdr_ioq_create(u_int min_bsize, u_int max_bsize, u_int uio_flags)
 		xdr_ioq_reset(xioq, 0);
 	}
 
-	return (xioq->xdrs);
+	return (xioq);
 }
 
 /*
@@ -294,7 +316,7 @@ xdr_ioq_create(u_int min_bsize, u_int max_bsize, u_int uio_flags)
  * reducing computations in the get/put/inline routines.
  */
 static inline struct xdr_ioq_uv *
-xdr_ioq_uv_next(struct xdr_ioq *xioq, u_int ioq_flags)
+xdr_ioq_uv_advance(struct xdr_ioq *xioq)
 {
 	struct xdr_ioq_uv *uv = IOQV(xioq->xdrs[0].x_base);
 	size_t len;
@@ -306,82 +328,74 @@ xdr_ioq_uv_next(struct xdr_ioq *xioq, u_int ioq_flags)
 	xioq->ioq_uv.plength += len;
 
 	/* next buffer, if any */
-	uv = IOQ_(TAILQ_NEXT(&uv->uvq, q));
+	return IOQ_(TAILQ_NEXT(&uv->uvq, q));
+}
 
-	/* append new segments, iif requested */
-	if ((!uv) && likely(ioq_flags & IOQ_FLAG_XTENDQ)) {
-		uv = IOQV(xioq->xdrs[0].x_base);
+/*
+ * Append at read/insert or fill position.
+ */
+static struct xdr_ioq_uv *
+xdr_ioq_uv_append(struct xdr_ioq *xioq, u_int ioq_flags)
+{
+	struct xdr_ioq_uv *uv = IOQV(xioq->xdrs[0].x_base);
 
-		if (xioq->ioq_uv.uvq_fetch) {
-			/* more of the same kind */
-			struct poolq_entry *have =
-				xioq->ioq_uv.uvq_fetch(xioq, uv->u.uio_p1,
-							"next buffer", 1,
-							IOQ_FLAG_NONE);
+	if (xioq->ioq_uv.uvq_fetch) {
+		/* more of the same kind */
+		struct poolq_entry *have =
+			xioq->ioq_uv.uvq_fetch(xioq, uv->u.uio_p1,
+						"next buffer", 1,
+						IOQ_FLAG_NONE);
 
-			/* poolq_entry is the top element of xdr_ioq_uv */
-			uv = IOQ_(have);
-			assert((void *)uv == (void *)have);
-		} else if (ioq_flags & IOQ_FLAG_BALLOC) {
-			/* XXX workaround for lack of segmented buffer
-			 * interfaces in some callers (e.g, GSS_WRAP) */
-			if (uv->u.uio_flags & UIO_FLAG_REALLOC) {
-				void *base;
-				size_t size = ioquv_size(uv);
-				size_t delta = xdr_tail_inline(xioq->xdrs);
+		/* poolq_entry is the top element of xdr_ioq_uv */
+		uv = IOQ_(have);
+		assert((void *)uv == (void *)have);
+	} else if (ioq_flags & IOQ_FLAG_BALLOC) {
+		/* XXX workaround for lack of segmented buffer
+		 * interfaces in some callers (e.g, GSS_WRAP) */
+		if (uv->u.uio_flags & UIO_FLAG_REALLOC) {
+			uint8_t *base;
+			size_t size = ioquv_size(uv);
+			size_t delta = xdr_tail_inline(xioq->xdrs);
+			size_t len = ioquv_length(uv);
 
-				/* bail if we have reached max bufsz */
-				if (size >= xioq->ioq_uv.max_bsize)
-					return (NULL);
+			/* bail if we have reached max bufsz */
+			if (size >= xioq->ioq_uv.max_bsize)
+				return (NULL);
 
-				/* backtrack */
-				xioq->ioq_uv.plength -= len;
-				assert(uv->u.uio_flags & UIO_FLAG_FREE);
+			/* backtrack */
+			xioq->ioq_uv.plength -= len;
+			assert(uv->u.uio_flags & UIO_FLAG_FREE);
 
-				base = mem_alloc(xioq->ioq_uv.max_bsize);
-				memcpy(base, uv->v.vio_head, len);
-				mem_free(uv->v.vio_base, size);
-				uv->v.vio_base =
-				uv->v.vio_head = base + 0;
-				uv->v.vio_tail = base + len;
-				uv->v.vio_wrap = base
-						+ xioq->ioq_uv.max_bsize;
-				xioq->xdrs[0].x_v = uv->v;
-				xioq->xdrs[0].x_data = uv->v.vio_tail
-							- delta;
-				return (uv);
-			}
-			uv = xdr_ioq_uv_create(xioq->ioq_uv.min_bsize,
-						   UIO_FLAG_FREE);
-		} else {
-			/* XXX empty buffer slot (not supported for now) */
-			uv = xdr_ioq_uv_create(0, UIO_FLAG_NONE);
+			base = mem_alloc(xioq->ioq_uv.max_bsize);
+			memcpy(base, uv->v.vio_head, len);
+			mem_free(uv->v.vio_base, size);
+			uv->v.vio_base =
+			uv->v.vio_head = base + 0;
+			uv->v.vio_tail = base + len;
+			uv->v.vio_wrap = base + xioq->ioq_uv.max_bsize;
+			xioq->xdrs[0].x_v = uv->v;
+			xioq->xdrs[0].x_data = uv->v.vio_tail - delta;
+			return (uv);
 		}
+		uv = xdr_ioq_uv_create(xioq->ioq_uv.min_bsize, UIO_FLAG_FREE);
+		(xioq->ioq_uv.uvqh.qcount)++;
+		TAILQ_INSERT_TAIL(&xioq->ioq_uv.uvqh.qh, &uv->uvq, q);
+	} else {
+		/* XXX empty buffer slot (not supported for now) */
+		uv = xdr_ioq_uv_create(0, UIO_FLAG_NONE);
+		(xioq->ioq_uv.uvqh.qcount)++;
+		TAILQ_INSERT_TAIL(&xioq->ioq_uv.uvqh.qh, &uv->uvq, q);
 	}
 
-	if (uv) {
-		if (!xioq->ioq_uv.uvq_fetch) {
-			/* new xdr_ioq_uv */
-			(xioq->ioq_uv.uvqh.qcount)++;
-			TAILQ_INSERT_TAIL(&xioq->ioq_uv.uvqh.qh, &uv->uvq, q);
-		}
-
-		/* advance iterator */
-		xioq->xdrs[0].x_data = uv->v.vio_head;
-		xioq->xdrs[0].x_base = &uv->v;
-		xioq->xdrs[0].x_v = uv->v;
-		(xioq->ioq_uv.pcount)++;
-		/* xioq->ioq_uv.plength is unchanged (calculated above) */
-	}
-
+	xdr_ioq_uv_update(xioq, uv);
 	return (uv);
 }
 
 static bool
-xdr_ioq_getlong(XDR *xdrs, long *lp)
+xdr_ioq_getunit(XDR *xdrs, uint32_t *p)
 {
 	struct xdr_ioq_uv *uv;
-	void *future = xdrs->x_data + sizeof(uint32_t);
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
 
 	while (future > xdrs->x_v.vio_tail) {
 		if (unlikely(xdrs->x_data != xdrs->x_v.vio_tail)) {
@@ -391,37 +405,40 @@ xdr_ioq_getlong(XDR *xdrs, long *lp)
 				__func__);
 			return (false);
 		}
-		uv = xdr_ioq_uv_next(XIOQ(xdrs), IOQ_FLAG_NONE);
+
+		uv = xdr_ioq_uv_advance(XIOQ(xdrs));
 		if (!uv) {
 			return (false);
 		}
+		xdr_ioq_uv_update(XIOQ(xdrs), uv);
 		/* fill pointer has changed */
 		future = xdrs->x_data + sizeof(uint32_t);
 	}
 
-	*lp = (long)ntohl(*((uint32_t *) (xdrs->x_data)));
+	*p = (uint32_t)ntohl(*((uint32_t *) (xdrs->x_data)));
 	xdrs->x_data = future;
 	return (true);
 }
 
 static bool
-xdr_ioq_putlong(XDR *xdrs, const long *lp)
+xdr_ioq_putunit(XDR *xdrs, const uint32_t v)
 {
 	struct xdr_ioq_uv *uv;
-	void *future = xdrs->x_data + sizeof(uint32_t);
+	uint8_t *future = xdrs->x_data + sizeof(uint32_t);
 
 	while (future > xdrs->x_v.vio_wrap) {
 		/* advance fill pointer, skipping unaligned */
-		uv = xdr_ioq_uv_next(XIOQ(xdrs),
-				IOQ_FLAG_XTENDQ | IOQ_FLAG_BALLOC);
+		uv = xdr_ioq_uv_advance(XIOQ(xdrs));
 		if (!uv) {
-			return (false);
+			uv = xdr_ioq_uv_append(XIOQ(xdrs), IOQ_FLAG_BALLOC);
+		} else {
+			xdr_ioq_uv_update(XIOQ(xdrs), uv);
 		}
 		/* fill pointer has changed */
 		future = xdrs->x_data + sizeof(uint32_t);
 	}
 
-	*((int32_t *) (xdrs->x_data)) = (int32_t) htonl((int32_t) (*lp));
+	*((uint32_t *) (xdrs->x_data)) = (uint32_t) htonl(v);
 	xdrs->x_data = future;
 	return (true);
 }
@@ -446,10 +463,11 @@ xdr_ioq_getbytes(XDR *xdrs, char *addr, u_int len)
 			delta = len;
 		} else if (unlikely(!delta)) {
 			/* advance fill pointer */
-			uv = xdr_ioq_uv_next(XIOQ(xdrs), IOQ_FLAG_NONE);
+			uv = xdr_ioq_uv_advance(XIOQ(xdrs));
 			if (!uv) {
 				return (false);
 			}
+			xdr_ioq_uv_update(XIOQ(xdrs), uv);
 			continue;
 		}
 		memcpy(addr, xdrs->x_data, delta);
@@ -476,10 +494,12 @@ xdr_ioq_putbytes(XDR *xdrs, const char *addr, u_int len)
 			delta = len;
 		} else if (!delta) {
 			/* advance fill pointer */
-			uv = xdr_ioq_uv_next(XIOQ(xdrs),
-					IOQ_FLAG_XTENDQ | IOQ_FLAG_BALLOC);
+			uv = xdr_ioq_uv_advance(XIOQ(xdrs));
 			if (!uv) {
-				return (false);
+				uv = xdr_ioq_uv_append(XIOQ(xdrs),
+							IOQ_FLAG_BALLOC);
+			} else {
+				xdr_ioq_uv_update(XIOQ(xdrs), uv);
 			}
 			continue;
 		}
@@ -525,9 +545,11 @@ xdr_ioq_getbufs(XDR *xdrs, xdr_uio *uio, u_int flags)
 		if (unlikely(delta > len)) {
 			delta = len;
 		} else if (unlikely(!delta)) {
-			uv = xdr_ioq_uv_next(XIOQ(xdrs), IOQ_FLAG_NONE);
+			uv = xdr_ioq_uv_advance(XIOQ(xdrs));
 			if (!uv)
 				return (false);
+
+			xdr_ioq_uv_update(XIOQ(xdrs), uv);
 			continue;
 		}
 		(uio->xbs_buf[ix]).xb_p1 = uv;
@@ -554,11 +576,11 @@ xdr_ioq_putbufs(XDR *xdrs, xdr_uio *uio, u_int flags)
 
 	for (ix = 0; ix < uio->uio_count; ++ix) {
 		/* advance fill pointer, do not allocate buffers, refs =1 */
-		if (!(flags & IOQ_FLAG_XTENDQ))
-			return (FALSE);
-		uv = xdr_ioq_uv_next(XIOQ(xdrs), flags);
+		uv = xdr_ioq_uv_advance(XIOQ(xdrs));
 		if (!uv)
-			return (FALSE);
+			uv = xdr_ioq_uv_append(XIOQ(xdrs), flags);
+		else
+			xdr_ioq_uv_update(XIOQ(xdrs), uv);
 
 		v = &(uio->uio_vio[ix]);
 		uv->u.uio_flags = UIO_FLAG_NONE; /* !RECLAIM */
@@ -577,9 +599,13 @@ Saved for later golden buttery results -- Matt
 	} else {
 		for (ix = 0; ix < uio->xbs_cnt; ++ix) {
 			/* advance fill pointer, do not allocate buffers */
-			*uv = xdr_ioq_uv_next(XIOQ(xdrs), IOQ_FLAG_XTENDQ);
+			*uv = xdr_ioq_uv_advance(XIOQ(xdrs));
 			if (!uv)
-				return (false);
+				uv = xdr_ioq_uv_append(XIOQ(xdrs),
+							IOQ_FLAG_NONE);
+			else
+				xdr_ioq_uv_update(XIOQ(xdrs), uv);
+
 			xbuf = &(uio->xbs_buf[ix]);
 			XIOQ(xdrs)->ioq_uv.plength += xbuf->xb_len;
 			uv->u.uio_flags = UIO_FLAG_NONE;	/* !RECLAIM */
@@ -607,8 +633,6 @@ Saved for later golden buttery results -- Matt
  *
  * Update the logical and physical offsets and lengths,
  * based upon the most recent position information.
- * All such updates are consolidated here and xdr_ioq_uv_next,
- * reducing computations in the get/put/inline routines.
  */
 static u_int
 xdr_ioq_getpos(XDR *xdrs)
@@ -626,8 +650,6 @@ xdr_ioq_getpos(XDR *xdrs)
  *
  * Update the logical and physical offsets and lengths,
  * based upon the most recent position information.
- * All such updates are consolidated here and xdr_ioq_uv_next,
- * reducing computations in the get/put/inline routines.
  */
 static bool
 xdr_ioq_setpos(XDR *xdrs, u_int pos)
@@ -663,42 +685,6 @@ xdr_ioq_setpos(XDR *xdrs, u_int pos)
 	return (false);
 }
 
-static int32_t *
-xdr_ioq_inline(XDR *xdrs, u_int len)
-{
-	/* bugfix:  return fill pointer, not head! */
-	int32_t *buf = (int32_t *)xdrs->x_data;
-	void *future = xdrs->x_data + len;
-
-	/* bugfix: do not move fill position beyond tail or wrap */
-	switch (xdrs->x_op) {
-	case XDR_ENCODE:
-		if (future <= xdrs->x_v.vio_wrap) {
-			/* bugfix:  do not move head! */
-			xdrs->x_data = future;
-			/* bugfix: do not move tail beyond pfoff or wrap! */
-			xdr_tail_update(xdrs);
-			return (buf);
-		}
-		break;
-	case XDR_DECODE:
-		/* re-consuming bytes in a stream
-		 * (after SETPOS/rewind) */
-		if (future <= xdrs->x_v.vio_tail) {
-			/* bugfix:  do not move head! */
-			xdrs->x_data = future;
-			/* bugfix:  do not move tail! */
-			return (buf);
-		}
-		break;
-	default:
-		abort();
-		break;
-	};
-
-	return (NULL);
-}
-
 void
 xdr_ioq_release(struct poolq_head *ioqh)
 {
@@ -711,7 +697,11 @@ xdr_ioq_release(struct poolq_head *ioqh)
 		TAILQ_REMOVE(&ioqh->qh, have, q);
 		(ioqh->qcount)--;
 
-		xdr_ioq_uv_release(IOQ_(have));
+		if (have->qflags & IOQ_FLAG_SEGMENT) {
+			xdr_ioq_destroy(_IOQ(have), have->qsize);
+		} else {
+			xdr_ioq_uv_release(IOQ_(have));
+		}
 		have = next;
 	}
 	assert(ioqh->qcount == 0);
@@ -720,12 +710,19 @@ xdr_ioq_release(struct poolq_head *ioqh)
 void
 xdr_ioq_destroy(struct xdr_ioq *xioq, size_t qsize)
 {
+	__warnx(TIRPC_DEBUG_FLAG_XDR,
+		"%s() xioq %p",
+		__func__, xioq);
+
 	xdr_ioq_release(&xioq->ioq_uv.uvqh);
 
 	if (xioq->ioq_pool) {
 		xdr_ioq_uv_recycle(xioq->ioq_pool, &xioq->ioq_s);
-	} else {
-		poolq_head_destroy(&xioq->ioq_uv.uvqh);
+		return;
+	}
+	poolq_head_destroy(&xioq->ioq_uv.uvqh);
+
+	if (xioq->xdrs[0].x_flags & XDR_FLAG_FREE) {
 		mem_free(xioq, qsize);
 	}
 }
@@ -768,13 +765,12 @@ xdr_ioq_noop(void)
 }
 
 const struct xdr_ops xdr_ioq_ops = {
-	xdr_ioq_getlong,
-	xdr_ioq_putlong,
+	xdr_ioq_getunit,
+	xdr_ioq_putunit,
 	xdr_ioq_getbytes,
 	xdr_ioq_putbytes,
 	xdr_ioq_getpos,
 	xdr_ioq_setpos,
-	xdr_ioq_inline,
 	xdr_ioq_destroy_internal,
 	xdr_ioq_control,
 	xdr_ioq_getbufs,
