@@ -49,6 +49,7 @@
 #include <rpc/rpc.h>
 #include "un-namespace.h"
 
+#include "svc_internal.h"
 #include "rpc_rdma.h"
 
 /* NOTA BENE: as in xdr_ioq.c, although indications of failure are returned,
@@ -197,7 +198,7 @@ xdr_rdma_chunk_in(struct poolq_entry *have, u_int k, u_int m, u_int sized)
 	IOQ_(have)->v.vio_tail = (char *)IOQ_(have)->v.vio_base + m;
 	IOQ_(have)->v.vio_wrap = (char *)IOQ_(have)->v.vio_base + sized;
 
-	while (0 < --k && NULL != (have = TAILQ_PREV(have, q_head, q))) {
+	while (0 < --k && NULL != (have = TAILQ_PREV(have, poolq_head_s, q))) {
 		/* restore defaults after previous usage */
 		IOQ_(have)->v.vio_head = IOQ_(have)->v.vio_base;
 		IOQ_(have)->v.vio_tail =
@@ -213,7 +214,7 @@ xdr_rdma_chunk_out(struct poolq_entry *have, u_int k, u_int m, u_int sized)
 	IOQ_(have)->v.vio_tail = IOQ_(have)->v.vio_base;
 	IOQ_(have)->v.vio_wrap = (char *)IOQ_(have)->v.vio_base + m;
 
-	while (0 < --k && NULL != (have = TAILQ_PREV(have, q_head, q))) {
+	while (0 < --k && NULL != (have = TAILQ_PREV(have, poolq_head_s, q))) {
 		/* restore defaults after previous usage */
 		IOQ_(have)->v.vio_head =
 		IOQ_(have)->v.vio_tail = IOQ_(have)->v.vio_base;
@@ -269,10 +270,10 @@ xdr_rdma_respond_callback(struct rpc_rdma_cbc *cbc, RDMAXPRT *xprt)
 		"%s() %p[%u] cbc %p\n",
 		__func__, xprt, xprt->state, cbc);
 
-	mutex_lock(&xprt->waitq.ioq_uv.uvqh.qmutex);
-	TAILQ_REMOVE(&xprt->waitq.ioq_uv.uvqh.qh, &cbc->workq.ioq_s, q);
-	(xprt->waitq.ioq_uv.uvqh.qcount)--;
-	mutex_unlock(&xprt->waitq.ioq_uv.uvqh.qmutex);
+	mutex_lock(&xprt->sm_dr.ioq.ioq_uv.uvqh.qmutex);
+	TAILQ_REMOVE(&xprt->sm_dr.ioq.ioq_uv.uvqh.qh, &cbc->workq.ioq_s, q);
+	(xprt->sm_dr.ioq.ioq_uv.uvqh.qcount)--;
+	mutex_unlock(&xprt->sm_dr.ioq.ioq_uv.uvqh.qmutex);
 
 	xdr_ioq_destroy(&cbc->workq, sizeof(*cbc));
 	return (0);
@@ -285,10 +286,10 @@ xdr_rdma_destroy_callback(struct rpc_rdma_cbc *cbc, RDMAXPRT *xprt)
 		"%s() %p[%u] cbc %p\n",
 		__func__, xprt, xprt->state, cbc);
 
-	mutex_lock(&xprt->waitq.ioq_uv.uvqh.qmutex);
-	TAILQ_REMOVE(&xprt->waitq.ioq_uv.uvqh.qh, &cbc->workq.ioq_s, q);
-	(xprt->waitq.ioq_uv.uvqh.qcount)--;
-	mutex_unlock(&xprt->waitq.ioq_uv.uvqh.qmutex);
+	mutex_lock(&xprt->sm_dr.ioq.ioq_uv.uvqh.qmutex);
+	TAILQ_REMOVE(&xprt->sm_dr.ioq.ioq_uv.uvqh.qh, &cbc->workq.ioq_s, q);
+	(xprt->sm_dr.ioq.ioq_uv.uvqh.qcount)--;
+	mutex_unlock(&xprt->sm_dr.ioq.ioq_uv.uvqh.qmutex);
 
 	xdr_ioq_destroy(&cbc->workq, sizeof(*cbc));
 	return (0);
@@ -335,7 +336,9 @@ xdr_rdma_warn_callback(struct rpc_rdma_cbc *cbc, RDMAXPRT *xprt)
 static int
 xdr_rdma_wrap_callback(struct rpc_rdma_cbc *cbc, RDMAXPRT *xprt)
 {
-	return (int)xprt->xa->request_cb(cbc, &xprt->xprt);
+	XDR *xdrs = cbc->holdq.xdrs;
+
+	return (int)__svc_params->request_cb(&xprt->sm_dr.xprt, xdrs);
 }
 
 /***********************************/
@@ -843,7 +846,7 @@ xdr_rdma_encode_error(struct xdr_ioq_uv *call_uv, enum rpcrdma_errcode err)
 	case RDMA_ERR_BADHEADER:
 		break;
 	}
-	call_uv->v.vio_tail = va;
+	call_uv->v.vio_tail = (uint8_t *)va;
 }
 
 #ifdef UNUSED
@@ -890,7 +893,7 @@ static void
 xdr_rdma_callq(RDMAXPRT *xprt)
 {
 	struct poolq_entry *have =
-		xdr_ioq_uv_fetch(&xprt->waitq, &xprt->cbqh,
+		xdr_ioq_uv_fetch(&xprt->sm_dr.ioq, &xprt->cbqh,
 				 "callq context", 1, IOQ_FLAG_NONE);
 	struct rpc_rdma_cbc *cbc = (struct rpc_rdma_cbc *)(_IOQ(have));
 
@@ -900,7 +903,8 @@ xdr_rdma_callq(RDMAXPRT *xprt)
 	/* input positions */
 	IOQ_(have)->v.vio_head = IOQ_(have)->v.vio_base;
 	IOQ_(have)->v.vio_tail = IOQ_(have)->v.vio_wrap;
-	IOQ_(have)->v.vio_wrap = (char *)IOQ_(have)->v.vio_base + xprt->recvsize;
+	IOQ_(have)->v.vio_wrap = (char *)IOQ_(have)->v.vio_base
+			       + xprt->sm_dr.recvsz;
 
 	cbc->workq.xdrs[0].x_lib[1] =
 	cbc->holdq.xdrs[0].x_lib[1] = xprt;
@@ -930,7 +934,7 @@ xdr_rdma_destroy(XDR *xdrs)
 		xprt->mr = NULL;
 	}
 
-	xdr_ioq_destroy_pool(&xprt->waitq.ioq_uv.uvqh);
+	xdr_ioq_destroy_pool(&xprt->sm_dr.ioq.ioq_uv.uvqh);
 
 	/* must be after queues, xdr_ioq_destroy() moves them here */
 	xdr_ioq_release(&xprt->inbufs.uvqh);
@@ -956,11 +960,9 @@ xdr_rdma_destroy(XDR *xdrs)
  * credits is the number of buffers used
  */
 int
-xdr_rdma_create(XDR *xdrs, RDMAXPRT *xprt, const u_int sendsize,
-		const u_int recvsize, const u_int flags)
+xdr_rdma_create(XDR *xdrs, RDMAXPRT *xprt)
 {
 	uint8_t *b;
-	long ps = sysconf(_SC_PAGESIZE);
 
 	if (!xprt->pd || !xprt->pd->pd) {
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
@@ -970,22 +972,17 @@ xdr_rdma_create(XDR *xdrs, RDMAXPRT *xprt, const u_int sendsize,
 		return ENODEV;
 	}
 
-	/* buffer sizes MUST be page sized */
-	xprt->sendsize = sendsize & ~(ps - 1);
-	xprt->sendsize = xprt->sendsize >= ps ? xprt->sendsize : ps;
-	xprt->recvsize = recvsize & ~(ps - 1);
-	xprt->recvsize = xprt->recvsize >= ps ? xprt->recvsize : ps;
-
 	/* pre-allocated buffer_total:
 	 * the number of credits is irrelevant here.
 	 * instead, allocate buffers to match the read/write contexts.
 	 * more than one buffer can be chained to one ioq_uv head,
 	 * but never need more ioq_uv heads than buffers.
 	 */
-	xprt->buffer_total = recvsize * xprt->xa->rq_depth
-			   + sendsize * xprt->xa->sq_depth;
+	xprt->buffer_total = xprt->sm_dr.recvsz * xprt->xa->rq_depth
+			   + xprt->sm_dr.sendsz * xprt->xa->sq_depth;
 
-	xprt->buffer_aligned = mem_aligned(ps, xprt->buffer_total);
+	xprt->buffer_aligned =
+		mem_aligned(xprt->sm_dr.pagesz, xprt->buffer_total);
 
 	__warnx(TIRPC_DEBUG_FLAG_RPC_RDMA,
 		"%s() buffer_aligned at %p",
@@ -999,12 +996,12 @@ xdr_rdma_create(XDR *xdrs, RDMAXPRT *xprt, const u_int sendsize,
 				IBV_ACCESS_REMOTE_READ);
 
 	poolq_head_setup(&xprt->inbufs.uvqh);
-	xprt->inbufs.min_bsize = ps;
-	xprt->inbufs.max_bsize = xprt->recvsize;
+	xprt->inbufs.min_bsize = xprt->sm_dr.pagesz;
+	xprt->inbufs.max_bsize = xprt->sm_dr.recvsz;
 
 	poolq_head_setup(&xprt->outbufs.uvqh);
-	xprt->outbufs.min_bsize = ps;
-	xprt->outbufs.max_bsize = xprt->sendsize;
+	xprt->outbufs.min_bsize = xprt->sm_dr.pagesz;
+	xprt->outbufs.max_bsize = xprt->sm_dr.sendsz;
 
 	/* Each pre-allocated buffer has a corresponding xdr_ioq_uv,
 	 * stored on the pool queues.
@@ -1019,12 +1016,12 @@ xdr_rdma_create(XDR *xdrs, RDMAXPRT *xprt, const u_int sendsize,
 		data->v.vio_base =
 		data->v.vio_head =
 		data->v.vio_tail = b;
-		data->v.vio_wrap = (char *)b + xprt->recvsize;
+		data->v.vio_wrap = (char *)b + xprt->sm_dr.recvsz;
 		data->u.uio_p1 = &xprt->inbufs.uvqh;
 		data->u.uio_p2 = xprt->mr;
 		TAILQ_INSERT_TAIL(&xprt->inbufs.uvqh.qh, &data->uvq, q);
 
-		b += xprt->recvsize;
+		b += xprt->sm_dr.recvsz;
 	}
 
 	for (xprt->outbufs.uvqh.qcount = 0;
@@ -1035,16 +1032,16 @@ xdr_rdma_create(XDR *xdrs, RDMAXPRT *xprt, const u_int sendsize,
 		data->v.vio_base =
 		data->v.vio_head =
 		data->v.vio_tail = b;
-		data->v.vio_wrap = (char *)b + xprt->sendsize;
+		data->v.vio_wrap = (char *)b + xprt->sm_dr.sendsz;
 		data->u.uio_p1 = &xprt->outbufs.uvqh;
 		data->u.uio_p2 = xprt->mr;
 		TAILQ_INSERT_TAIL(&xprt->outbufs.uvqh.qh, &data->uvq, q);
 
-		b += xprt->sendsize;
+		b += xprt->sm_dr.sendsz;
 	}
 
-	xdr_ioq_setup(&xprt->waitq);
-	while (xprt->waitq.ioq_uv.uvqh.qcount < CALLQ_SIZE) {
+	xdr_ioq_setup(&xprt->sm_dr.ioq);
+	while (xprt->sm_dr.ioq.ioq_uv.uvqh.qcount < CALLQ_SIZE) {
 		xdr_rdma_callq(xprt);
 	}
 	return 0;
@@ -1163,7 +1160,7 @@ bool
 xdr_rdma_svc_recv(struct rpc_rdma_cbc *cbc, u_int32_t xid)
 {
 	RDMAXPRT *xprt;
-	struct rdma_msg *rmsg;
+	struct rdma_msg *cmsg;
 	uint32_t k;
 	uint32_t l;
 
@@ -1181,30 +1178,31 @@ xdr_rdma_svc_recv(struct rpc_rdma_cbc *cbc, u_int32_t xid)
 
 	cbc->call_uv = IOQ_(TAILQ_FIRST(&cbc->workq.ioq_uv.uvqh.qh));
 	(cbc->call_uv->u.uio_references)++;
-	rmsg = m_(cbc->call_uv->v.vio_head);
-	rpcrdma_dump_msg(cbc->call_uv, "call", rmsg->rdma_xid);
+	cbc->call_head = cbc->call_uv->v.vio_head;
+	cmsg = m_(cbc->call_head);
+	rpcrdma_dump_msg(cbc->call_uv, "call", cmsg->rdma_xid);
 
-	switch (ntohl(rmsg->rdma_vers)) {
+	switch (ntohl(cmsg->rdma_vers)) {
 	case RPCRDMA_VERSION:
 		break;
 	default:
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s() rdma_vers %" PRIu32 "?",
-			__func__, ntohl(rmsg->rdma_vers));
+			__func__, ntohl(cmsg->rdma_vers));
 		xdr_rdma_encode_error(cbc->call_uv, RDMA_ERR_VERS);
 		xdr_rdma_post_send_cb(xprt, cbc, 1);
 		xdr_ioq_uv_release(cbc->call_uv);
 		return (false);
 	}
 
-	switch (ntohl(rmsg->rdma_type)) {
+	switch (ntohl(cmsg->rdma_type)) {
 	case RDMA_MSG:
 	case RDMA_NOMSG:
 		break;
 	default:
 		__warnx(TIRPC_DEBUG_FLAG_ERROR,
 			"%s() rdma_type %" PRIu32 "?",
-			__func__, ntohl(rmsg->rdma_type));
+			__func__, ntohl(cmsg->rdma_type));
 		xdr_rdma_encode_error(cbc->call_uv, RDMA_ERR_BADHEADER);
 		xdr_rdma_post_send_cb(xprt, cbc, 1);
 		xdr_ioq_uv_release(cbc->call_uv);
@@ -1212,7 +1210,7 @@ xdr_rdma_svc_recv(struct rpc_rdma_cbc *cbc, u_int32_t xid)
 	}
 
 	/* locate NFS/RDMA (RFC-5666) chunk positions */
-	cbc->read_chunk = xdr_rdma_get_read_list(rmsg);
+	cbc->read_chunk = xdr_rdma_get_read_list(cmsg);
 	cbc->write_chunk = (wl_t *)cbc->read_chunk;
 	xdr_rdma_skip_read_list((uint32_t **)&cbc->write_chunk);
 	cbc->reply_chunk = cbc->write_chunk;
@@ -1227,19 +1225,20 @@ xdr_rdma_svc_recv(struct rpc_rdma_cbc *cbc, u_int32_t xid)
 
 	/* skip past the header for the calling buffer */
 	xdr_ioq_reset(&cbc->holdq, ((uintptr_t)cbc->call_data
-				  - (uintptr_t)rmsg));
+				  - (uintptr_t)cmsg));
 
 	while (rl(cbc->read_chunk)->present != 0
 	    && rl(cbc->read_chunk)->position == 0) {
 		l = ntohl(rl(cbc->read_chunk)->target.length);
 		k = xdr_rdma_chunk_fetch(&cbc->workq, &xprt->inbufs.uvqh,
-					 "call chunk", l, xprt->recvsize,
+					 "call chunk", l,
+					 xprt->sm_dr.recvsz,
 					 xprt->xa->max_recv_sge,
 					 xdr_rdma_chunk_in);
 
 		xdr_rdma_wait_read_cb(xprt, cbc, k, &rl(cbc->read_chunk)->target);
 		rpcrdma_dump_msg(IOQ_(TAILQ_FIRST(&cbc->workq.ioq_uv.uvqh.qh)),
-				 "call chunk", rmsg->rdma_xid);
+				 "call chunk", cmsg->rdma_xid);
 
 		/* concatenate any additional buffers after the calling message,
 		 * faking there is more call data in the calling buffer.
@@ -1320,7 +1319,8 @@ xdr_rdma_svc_reply(struct rpc_rdma_cbc *cbc, u_int32_t xid)
 		for (i = 0; i < n; i++) {
 			l = ntohl(reply_array->entry[i].target.length);
 			xdr_rdma_chunk_fetch(&cbc->holdq, &xprt->outbufs.uvqh,
-					     "sreply chunk", l, xprt->sendsize,
+					     "sreply chunk", l,
+					     xprt->sm_dr.sendsz,
 					     xprt->xa->max_send_sge,
 					     xdr_rdma_chunk_out);
 		}
@@ -1478,6 +1478,14 @@ xdr_rdma_svc_flushout(struct rpc_rdma_cbc *cbc)
 			__func__, ntohl(msg->rm_direction));
 		return (false);
 	}
+	cmsg = m_(cbc->call_head);
+
+	if (cmsg->rdma_xid != msg->rm_xid) {
+		__warnx(TIRPC_DEBUG_FLAG_XDR,
+			"%s() xid (%u) not equal RPC (%u)",
+			__func__, ntohl(cmsg->rdma_xid), ntohl(msg->rm_xid));
+		return (false);
+	}
 
 	/* usurp the holdq for the head, move to workq later */
 	head_uv = IOQ_(xdr_ioq_uv_fetch(&cbc->holdq, &xprt->outbufs.uvqh,
@@ -1486,14 +1494,11 @@ xdr_rdma_svc_flushout(struct rpc_rdma_cbc *cbc)
 	/* entry was already added directly to the queue */
 	head_uv->v.vio_head = head_uv->v.vio_base;
 	/* tail adjusted below */
-	head_uv->v.vio_wrap = (char *)head_uv->v.vio_base + xprt->sendsize;
+	head_uv->v.vio_wrap = (char *)head_uv->v.vio_base + xprt->sm_dr.sendsz;
 
 	/* build the header that goes with the data */
-	cmsg = m_(cbc->call_uv->v.vio_head);
 	rmsg = m_(head_uv->v.vio_head);
 	rmsg->rdma_xid = cmsg->rdma_xid;
-	/* TODO: check it matches msg->rm_xid */
-
 	rmsg->rdma_vers = cmsg->rdma_vers;
 	rmsg->rdma_credit = htonl(xprt->xa->credits);
 
@@ -1529,20 +1534,21 @@ xdr_rdma_svc_flushout(struct rpc_rdma_cbc *cbc)
 			struct xdr_rdma_segment *w_seg =
 				&w_array->entry[i++].target;
 			uint32_t length = ntohl(c_seg->length);
-			uint32_t k = length / xprt->sendsize;
-			uint32_t m = length % xprt->sendsize;
-		
+			uint32_t k = length / xprt->sm_dr.sendsz;
+			uint32_t m = length % xprt->sm_dr.sendsz;
+
 			if (m) {
 				/* need fractional buffer */
 				k++;
 			}
-		
+
 			/* ensure never asking for more buffers than allowed */
 			if (k > xprt->xa->max_send_sge) {
 				__warnx(TIRPC_DEBUG_FLAG_XDR,
 					"%s() requested chunk %" PRIu32
 					" is too long (%" PRIu32 ">%" PRIu32 ")",
-					__func__, length, k, xprt->xa->max_send_sge);
+					__func__, length, k,
+					xprt->xa->max_send_sge);
 				k = xprt->xa->max_send_sge;
 			}
 
